/* common.h -- common include for direct library and FMM code */

/* M. Holm 2011-12-12 */

#ifndef __common_h
#define __common_h

// get rid of c99 "restrict" keyword that aids cache optimization.
#ifdef __cplusplus
#define restrict __restrict
#endif



#ifdef __cplusplus 
#include <complex>
// from FMM2Dgpumake
using namespace std;
typedef complex<double> dcmplx;
// here we can rely on the dcmplx class // not (yet?) used
#define COMPLEXADD(z,x,y)    ((z) += dcmplx(x,y))
#define COMPLEXSUB(z,x,y)    ((z) -= dcmplx(x,y))
#define COMPLEXASSIGN(z,x,y) ((z) = dcmplx(x,y))
#define creal(z) ((z).real())
#define cimag(z) ((z).imag())
// redirect C99-syntax
#define clog(x) (log(x))
#define cabs(x) (abs(x))
const dcmplx I = dcmplx(0.0,1.0);
#else // IF NOT C++
#ifdef __CUDACC__
typedef struct { double re,im; } dcmplx;
#define COMPLEXADD(z,x,y)    {(z).re += x; (z).im += y;}
#define COMPLEXSUB(z,x,y)    {(z).re -= x; (z).im -= y;}
#define COMPLEXASSIGN(z,x,y) {(z).re = x;  (z).im = y;}
#define creal(z) ((z).re)
#define cimag(z) ((z).im)

#else // IF NOT C++ AND NOT CUDA
//#define bool int
typedef double complex dcmplx;

#define COMPLEXADD(z,x,y)    {(z).re += x; (z).im += y;}
#define COMPLEXSUB(z,x,y)    {(z).re -= x; (z).im -= y;}
#define COMPLEXASSIGN(z,x,y) {(z) = (x) + (y)*1.fi;}// {(z).re = x;  (z).im = y;}
#endif // CUDACC
#endif // cplusplus

/* Multipole expansion structure. The coefficients are to be
   understood as
      mpexp(z) = coeff1[0]*log(z-z0)+
      coeff1[1]/(z-z0)+coeff1[2]/(z-z0)^2+coeff1[3]/(z-z0)^3+...
   or, depending on the stage of the algorithm,
      mpexp(z) = coeff2[0]+coeff2[1]*(z-z0)+coeff2[2]*(z-z0)^2+...
*/
typedef struct {
  dcmplx z0,d0;           // midpoint and radius (z0 to upper right corner)
  dcmplx *coeff1,*coeff2; // multipole coefficients
} mpexp;



/* Sparse connectivity matrix (special CCS format integer matrix): the
   row indices in ir of column j are found in [jcptr[j],jcptr[j+1]),
   the near connections (think of these as the value 1) in
   [jcptr[j],kcptr[j]) and the far connections (i.e. value -1) in
   [kcptr[j],jcptr[j+1]). The row indices of the near part is sorted
   in ascending- and the far part in descending order,
   respectively. Only the strictly lower triangular part is stored.

   For far connections between two levels (value -2), a negative index
   will occur in ir, pointing to a box at the previous levels. Then
   three positive indices follow, pointing to children of the previous
   negative index. Those indices are ignored in the FMM assembly and are
   a by-product of the algorithm.

   At the finest level a special syntax is used in which the near part
   is grouped into a 'truly' near part (value 1) in [jptr[j],kcptr[j+N])
   and a 'less strongly' connected part (value 2) in
   [kcptr[j+N],kcptr[j]). In this case the near part is not sorted and
   also, the rowindices are signed (positive for cluster to particle,
   negative for particle to cluster connection). */
typedef struct {
  int *jcptr,*kcptr; /* column pointer and pointer to the beginning of
			the 'far'-part */
  int *ir;           // row indices
} mpSparse;

/* Multipole expansion tree in vectorized form. First comes the root,
   then the 4 children of the root and so on */
typedef struct {
  double xmin,xmax,
    ymin,ymax;       // bounding box
  int pcoeff;        /* number of coefficients offset by one; coeff[0]
			through coeff[pcoeff] */
  int nlevel;        /* number of levels offset by one; level =
			0..nlevel */
  int *lptr;         /* pointer to levels in tree; level l can be
			found in root[i] where lptr[l] <= i <
			lptr[l+1] and lptr[nlevel+1] is the total
			number of boxes. */
  mpexp *root;       // box 0, the root of the tree
  mpSparse *connect; // connectivity information for each level
  int *ix,*ixptr;    // permutation of potentials
  int *jx,*jxptr;    // permutation of evaluation points
  int jsize;
} MPexp;
const static double gamma_ = 0.5772156649015328606065121;

typedef enum {SCULLY = 0,RANKINE = 1,OSEEN = 2,DIRAC = 3,
	      ERR_SMOOTHER = -1} SMOOTHER;

#ifndef M_PI
#define M_PI 3.1415926535897932384626434
#endif

// helper function
void MPexp_smooth_(int pot,SMOOTHER smooth,double xopt,bool cont,
                   double *cutoff,double *shape,double *scale);

#endif // COMMON_H
