/* cudaeval.cu -- CUDA host code for direct evaluation. 
 * Provides similar functionality as direct.c. */

/* M. Holm 2011-12-20, from S. Engblom and A. Goude 2011-10-19 */

#include "cuda.h"
#include "cudaeval.h"

//#define TIMEGPU 
#ifdef TIMEGPU
#include <time.h>
timespec gpu_t_start;
timespec gpu_t_stop;
timespec gpu_c_start;
timespec gpu_c_stop;
#define GPU_T_START clock_gettime(CLOCK_REALTIME,&gpu_t_start); 
#define GPU_T_STOP clock_gettime(CLOCK_REALTIME,&gpu_t_stop);
#define GPU_C_START clock_gettime(CLOCK_REALTIME,&gpu_c_start);
#define GPU_C_STOP clock_gettime(CLOCK_REALTIME,&gpu_c_stop);
#define GPU_PRINTTIME fprintf(stderr,"data time = %f, compute time = %f, total=%f \n", \
		(double)(gpu_t_stop.tv_sec-gpu_t_start.tv_sec) + \
        (double)(gpu_t_stop.tv_nsec-gpu_t_start.tv_nsec)/1000000000, \
        (double)(gpu_c_stop.tv_sec-gpu_c_start.tv_sec) + \
        (double)(gpu_c_stop.tv_nsec-gpu_c_start.tv_nsec)/1000000000, \
        (double)(gpu_c_stop.tv_sec-gpu_t_start.tv_sec) + \
        (double)(gpu_c_stop.tv_nsec-gpu_t_start.tv_nsec)/1000000000);
#else

#define GPU_T_START ; 
#define GPU_T_STOP ;
#define GPU_C_START  ;
#define GPU_C_STOP  ;
#define GPU_PRINTTIME ;

#endif


//textures have to be global
texture<float> ter;
texture<float> tei;
texture<float> tzr;
texture<float> tzi;
texture<float> tmr;
texture<float> tmi;
texture<int> tixptr;
texture<int> tjxptr;

/*------------------------------------------------------------------------*/
int cudaalloccount; // remembers the number of cuda allocations


#define MAXBLOCKS 65535
#define MAXTHREADSEVALUATION 64 //max threads for fmm direct evaluation and multipole evaluation
#define DIRECTMAXTHREADS 256 //max threads for tol=0, i.e. direct summation
#define REORDERMAXTHREADS 256 //max threads for CUDA reordering

//cuda kernels has to be declared in same file as they are called
#include "cudaevalkernels.h"



//Allocates GPU memory and copies data to the GPU.
void MPexp_setup_cuda(const MPexp *This,
                     const double *er,const double *ei,
                     const double *zr,const double *zi,
                     const double *mr,const double *mi,
                     const double *pr,const double *qr,
                     cudavariables *GPUvars,int N,int NE,
                     int pot,SMOOTHER smooth,
                     double xopt,double cutoff,bool cont)
{
  // number of boxes on finest level, total number of boxes
  const int Nf = This->lptr[This->nlevel+1]-This->lptr[This->nlevel];
  mpSparse *C = This->connect;

  const int *restrict ix = This->ix;
  const int *restrict jx = This->jx;

  checkcudaerror("Check eval_cuda start");
            
  //start by allocating variables on GPU. 
  if(pr!=NULL) {
    cudasafeMalloc((void**)&GPUvars->pr,N*sizeof(double));
    if(pot!=0||mi!=NULL)
      cudasafeMalloc((void**)&GPUvars->pi,N*sizeof(double));
  }
  if(qr!=NULL) {
    cudasafeMalloc((void**)&GPUvars->er,NE*sizeof(double));
    cudasafeMalloc((void**)&GPUvars->ei,NE*sizeof(double));
    cudasafeMalloc((void**)&GPUvars->qr,NE*sizeof(double));
    if(pot!=0||mi!=NULL)
      cudasafeMalloc((void**)&GPUvars->qi,NE*sizeof(double));
  }
  else {
    GPUvars->er=NULL;
    GPUvars->ei=NULL;
  }


  cudasafeMalloc((void**)&GPUvars->zr,N*sizeof(double));
  cudasafeMalloc((void**)&GPUvars->zi,N*sizeof(double));
  cudasafeMalloc((void**)&GPUvars->mr,N*sizeof(double));
    
  if(mi!=NULL) {
    cudasafeMalloc((void**)&GPUvars->mi,N*sizeof(double));
  }
  else
    GPUvars->mi=NULL;
    
  // Allocate and copy connectivity to GPU
  // copy all levels of connectivity information to the GPU
  int size =  This->jsize;//C[This->nlevel].jcptr[Nf];
  
  
  cudasafe(cudaMallocDebug((void **)&(GPUvars->connect->jcptr),
                           size*sizeof(int)),"cudaMalloc GPUvars->connect");
  //GPUvars->connect->jcptr = (int*) cudaMalloc2(size*sizeof(int));

  cudaSafeMemcpy(GPUvars->connect->jcptr,C[This->nlevel].jcptr,
                 size*sizeof(int),cudaMemcpyHostToDevice,
                 "cudaMemcpy GPUvars->connect" );

  GPUvars->connect->kcptr = &GPUvars->connect->jcptr[Nf+1];
  GPUvars->connect->ir = &GPUvars->connect->kcptr[2*Nf];
  
  //reorder the elements on the GPU according to ix/jx to avoid random accesses to device memory. 
           
  double* mrtmp,*mitmp;

  double* zrtmp,*zitmp;
  int* cudaix,*cudajx;
  if(qr!=NULL) { //reorder the evaluation points
    if(pot==0&&mi==NULL)
      cudasafeMalloc((void**)&GPUvars->qi, NE*sizeof(double));
    cudasafeMalloc((void**)&cudajx, NE*sizeof(int));
    cudasafe( cudaMemcpy(cudajx, jx, NE*sizeof(int), cudaMemcpyHostToDevice), "cudaMemcpy jx" );
    cudasafe( cudaMemcpy(GPUvars->qr, er, NE*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy ertmp" );
    cudasafe( cudaMemcpy(GPUvars->qi, ei, NE*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy eitmp" );
        
    reorder_cuda_2<<<imin(MAXBLOCKS, (NE+REORDERMAXTHREADS-1)/REORDERMAXTHREADS), REORDERMAXTHREADS>>>(GPUvars->er, GPUvars->ei, GPUvars->qr, GPUvars->qi, cudajx, NE);
    if(pot==0&&mi==NULL)
      cudaFreeDebug(GPUvars->qi); //used as temporary variable, if not used anymore, cleanup
    GPUvars->jx=cudajx;
  }

  //reorder the potential points
  if(pr==NULL) { //the case no evaluation should happen on the potential points, in this case, do a proper cleanup afterwards
    cudasafeMalloc((void**)&zrtmp,N*sizeof(double));
    cudasafeMalloc((void**)&zitmp,N*sizeof(double));
    cudasafeMalloc((void**)&mrtmp,N*sizeof(double));
    cudasafeMalloc((void**)&cudaix,N*sizeof(int));
        
    if(mi!=NULL) {
      cudasafeMalloc((void**)&mitmp,N*sizeof(double));
      cudasafe( cudaMemcpy(mitmp, mi, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy mitmp" );
    }
    cudasafe( cudaMemcpy(zrtmp, zr, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy zrtmp" );
    cudasafe( cudaMemcpy(zitmp, zi, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy zitmp" );
    cudasafe( cudaMemcpy(mrtmp, mr, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy mrtmp" );
    cudasafe( cudaMemcpy(cudaix, ix, N*sizeof(int), cudaMemcpyHostToDevice), "cudaMemcpy ix" );
        
    if(mi!=NULL)
      reorder_cuda_4<<<imin(MAXBLOCKS,(N+REORDERMAXTHREADS-1)/REORDERMAXTHREADS),REORDERMAXTHREADS>>>(GPUvars->zr,GPUvars->zi,GPUvars->mr,GPUvars->mi,zrtmp,zitmp,mrtmp,mitmp,cudaix,N);
    else {
      reorder_cuda_3<<<imin(MAXBLOCKS,(N+REORDERMAXTHREADS-1)/REORDERMAXTHREADS),REORDERMAXTHREADS>>>(GPUvars->zr,GPUvars->zi,GPUvars->mr,zrtmp,zitmp,mrtmp,cudaix,N);
    }
    cudaFreeDebug(zrtmp);
    cudaFreeDebug(zitmp);
    cudaFreeDebug(mrtmp);
        
    if(mi!=NULL)
      cudaFreeDebug(mitmp);
    cudaFreeDebug(cudaix);
  }
  else { //evaluation will happen at the potential points. Now, the arrays can be allocated and kept on the GPU for future use
    if(pot==0&&mi==NULL)
      cudasafeMalloc((void**)&GPUvars->pi,N*sizeof(double));
    cudasafeMalloc((void**)&mrtmp,N*sizeof(double));
    cudasafeMalloc((void**)&cudaix,N*sizeof(int));
    if(mi!=NULL) {
      cudasafeMalloc((void**)&mitmp,N*sizeof(double));
      cudasafe( cudaMemcpy(mitmp, mi, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy mitmp" );
    }
    // temporarily copy zr, mr to the wrong place on GPU so reorder_cuda can put them in right order in the right place 
    cudasafe( cudaMemcpy(GPUvars->pr, zr, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy zrtmp" );
    cudasafe( cudaMemcpy(GPUvars->pi, zi, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy zitmp" );
    cudasafe( cudaMemcpy(mrtmp, mr, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy mrtmp" );
    cudasafe( cudaMemcpy(cudaix, ix, N*sizeof(int), cudaMemcpyHostToDevice), "cudaMemcpy ix" );
    if(mi!=NULL){
      reorder_cuda_4<<<imin(MAXBLOCKS,(N+REORDERMAXTHREADS-1)/REORDERMAXTHREADS),REORDERMAXTHREADS>>>(GPUvars->zr,GPUvars->zi,GPUvars->mr,GPUvars->mi,GPUvars->pr,GPUvars->pi,mrtmp,mitmp,cudaix,N);
    }else{
      reorder_cuda_3<<<imin(MAXBLOCKS,(N+REORDERMAXTHREADS-1)/REORDERMAXTHREADS),REORDERMAXTHREADS>>>(GPUvars->zr,GPUvars->zi,GPUvars->mr,GPUvars->pr,GPUvars->pi,mrtmp,cudaix,N);
    }
    cudaFreeDebug(mrtmp);
    if(pot==0&&mi==NULL)
      cudaFreeDebug(GPUvars->pi);
    if(mi!=NULL)
      cudaFreeDebug(mitmp);
    GPUvars->ix=cudaix;
  }
  
  GPUvars->Nf=Nf;
  
  //Allocate output
  if(pr!=NULL) {
    cudasafe( cudaMemset(GPUvars->pr, 0, N*sizeof(double)), "cudaMemset pr" );
    if(pot!=0||mi!=NULL)
      cudasafe( cudaMemset(GPUvars->pi, 0, N*sizeof(double)), "cudaMemset pi" );
  }  
  if(qr!=NULL) {
    cudasafe( cudaMemset(GPUvars->qr, 0, NE*sizeof(double)), "cudaMemset qr" );
    if(pot!=0||mi!=NULL)
      cudasafe( cudaMemset(GPUvars->qi, 0, NE*sizeof(double)), "cudaMemset qi" );
  }
  
  //use textures for faster memory access
  if(mi!=NULL) {
    cudasafe(cudaBindTexture(NULL,tmi,GPUvars->mi,N*sizeof(double)),"cudaBindTexture tmi");
  }
  cudasafe(cudaBindTexture(NULL,tzr,GPUvars->zr,N*sizeof(double)),"cudaBindTexture tzr");
  cudasafe(cudaBindTexture(NULL,tzi,GPUvars->zi,N*sizeof(double)),"cudaBindTexture tzi");
  cudasafe(cudaBindTexture(NULL,tmr,GPUvars->mr,N*sizeof(double)),"cudaBindTexture tmr");
  
  //copy the list of points in each box
  cudasafeMalloc((void**)&GPUvars->ixptr,(Nf+1)*sizeof(int));
  cudasafe( cudaMemcpy(GPUvars->ixptr, This->ixptr, (Nf+1)*sizeof(int), cudaMemcpyHostToDevice), "cudaMemcpy ixptr" );
  if(qr!=NULL) {
    cudasafeMalloc((void**)&GPUvars->jxptr,(Nf+1)*sizeof(int));
    cudasafe( cudaMemcpy(GPUvars->jxptr, This->jxptr, (Nf+1)*sizeof(int), cudaMemcpyHostToDevice), "cudaMemcpy jxptr" );
  }
  cudasafe(cudaBindTexture(NULL,tixptr,GPUvars->ixptr,(Nf+1)*sizeof(int)),"cudaBindTexture ixptr");
}

/*------------------------------------------------------------------------*/
/*------------------------------------------------------------------------*/
void cuda_evals_direct(const MPexp *This, int N, int NE,
                double *pr,double *pi,
                double *qr,double *qi,
                const double *er,const double *ei,
                const double *zr,const double *zi,
                const double *mr,const double *mi,
                int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Evaluation of the potential. Function is critical to performance. */
{
  // Cuda stuff:
  cudavariables GPUvars;
  
  GPUvars.connect = (mpSparse *)malloc(sizeof(mpSparse));
  
  double shape, scale;
  const int Nf = This->lptr[This->nlevel+1]-This->lptr[This->nlevel];
  MPexp_smooth_(pot, smooth, xopt, cont, &cutoff, &shape, &scale);

  GPU_T_START
  MPexp_setup_cuda(This,er,ei,zr,zi,mr,mi,pr,qr,&GPUvars,
                   N,NE,pot,smooth,xopt,cutoff,cont);
  GPU_T_STOP

  if(pot==0&&mi==NULL) { //mark in the GPUvars structure if evaluations should be done or not
    if(qr!=NULL)
      GPUvars.qi=NULL;
    if(pr!=NULL)
      GPUvars.pi=NULL;
    if(qr!=NULL)
      GPUvars.qi=NULL;
  }

  //Allocations and memory copies to cuda done. Start the algorithm
  //launch kernel for direct evaluation

  GPU_C_START

  checkcudaerror("Check before direct interact");
  //one case for evaluation in potential points and one for evaluation points
  if(pr!=NULL) {
    cudasafe(cudaBindTexture(NULL,ter,GPUvars.zr,N*sizeof(double)),"cudaBindTexture tzr");
    cudasafe(cudaBindTexture(NULL,tei,GPUvars.zi,N*sizeof(double)),"cudaBindTexture tzi");
    cudasafe(cudaBindTexture(NULL,tjxptr,GPUvars.ixptr,(Nf+1)*sizeof(int)),"cudaBindTexture jxptr");
    launchdirectkernel(This,pot,mi,&GPUvars,GPUvars.pr,GPUvars.pi,smooth,shape,scale,cutoff);
  }
  if(qr!=NULL) {
    //move the textures to the evaluation points before calling the kernel again
    if(pr!=NULL) {
      cudaUnbindTexture(ter);
      cudaUnbindTexture(tei);
      cudaUnbindTexture(tjxptr);
    }
    cudasafe(cudaBindTexture(NULL,ter,GPUvars.er,NE*sizeof(double)),"cudaBindTexture ter");
    cudasafe(cudaBindTexture(NULL,tei,GPUvars.ei,NE*sizeof(double)),"cudaBindTexture tei");
    cudasafe(cudaBindTexture(NULL,tjxptr,GPUvars.jxptr,(Nf+1)*sizeof(int)),"cudaBindTexture jxptr");
    launchdirectkernel(This,pot,mi,&GPUvars,GPUvars.qr,GPUvars.qi,smooth,shape,scale,cutoff);
  }

  checkcudaerror("Kernel launch error direct interact");
  
  cudaDeviceSynchronize(); // wait for kernels to finish 

  checkcudaerror("Threads are done. Check.");
// get results back and clean up
  MPexp_eval_cuda_result(pr,pi,qr,qi,This->ix,This->jx,N,NE,pot,&GPUvars,This);

  cudaDeviceReset();
  
  GPU_C_STOP

  GPU_PRINTTIME

}
/*------------------------------------------------------------------------*/
//chooses the correct direct evaluation kernel and launches GPU calculation.
void launchdirectkernel(const MPexp *This,int pot,const double *mi,cudavariables *GPUvars,
                        double *GPUqr,double *GPUqi,
                        SMOOTHER smooth,double shape,double scale,double cutoff)
{
  if(pot==1) {
    if(mi==NULL) {
      switch(smooth) { //one kernel for each type. This is to reduce the number of local variables for each kernel
      case DIRAC:
        pot1_nomi_dirac_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUqi,GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir);
        break;
      case RANKINE:
        pot1_nomi_rankine_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUqi, GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir, shape);
        break;
      case SCULLY:
        pot1_nomi_scully_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUqi, GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir, shape, scale, cutoff);
        break;
      case OSEEN:
        pot1_nomi_oseen_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUqi, GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir, shape, scale, cutoff);
        break;
      }
    }
    else {
      switch(smooth) { //one kernel for each type. This is to reduce the number of local variables for each kernel
      case DIRAC:
        pot1_mi_dirac_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUqi,GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir );
        break;
      case RANKINE:
        pot1_mi_rankine_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUqi, GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir, shape );
        break;
      case SCULLY:
        pot1_mi_scully_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUqi, GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir, shape, scale, cutoff );
        break;
      case OSEEN:
        pot1_mi_oseen_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUqi, GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir, shape, scale, cutoff );
        break;
      }
    }
  }
  else {
    if(mi==NULL) {
      switch(smooth) { //one kernel for each type. This is to reduce the number of local variables for each kernel
      case DIRAC:
        pot0_nomi_dirac_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr,GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir );
        break;
      case RANKINE:
        pot0_nomi_rankine_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir, shape, cutoff );
        break;
      case SCULLY:
        pot0_nomi_scully_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir, shape, scale, cutoff );
        break;
      case OSEEN:
        pot0_nomi_oseen_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir, shape, scale, cutoff );
        break;
      }
    }
    else {
      switch(smooth) { //one kernel for each type. This is to reduce the number of local variables for each kernel
      case DIRAC:
        pot0_mi_dirac_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUqi,GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir );
        break;
      case RANKINE:
        pot0_mi_rankine_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUqi, GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir, shape, cutoff );
        break;
      case SCULLY:
        pot0_mi_scully_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUqi, GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir, shape, scale, cutoff );
        break;
      case OSEEN:
        pot0_mi_oseen_synchronized<<<imin(MAXBLOCKS, GPUvars->Nf), MAXTHREADSEVALUATION>>>(GPUqr, GPUqi, GPUvars->Nf,GPUvars->connect->jcptr,&GPUvars->connect->kcptr[GPUvars->Nf],GPUvars->connect->ir, shape, scale, cutoff );
        break;
      }
    }
  }
  checkcudaerror("Kernel launch error direct interact");
}

/*------------------------------------------------------------------------*/
//function to collect results and performes cleanup, should be called at the end of the algorithm

void MPexp_eval_cuda_result(double *pr,double *pi,
                            double *qr,double *qi,
                            int *ix,int *jx,int N,int NE,
                            int pot,cudavariables *GPUvars,
                            const MPexp *This)
{
  //copy results back to normal memory and combine with CPU results
 double *fqr,*fqi,*fpr,*fpi;
  if(pr!=NULL) {
    fpr=(double*)malloc(N*sizeof(double));
    cudasafe( cudaMemcpy(fpr, GPUvars->pr, N*sizeof(double), cudaMemcpyDeviceToHost), "cudaMemcpy pr result" );
    if(pi!=NULL) {
      fpi=(double*)malloc(N*sizeof(double));
      cudasafe( cudaMemcpy(fpi, GPUvars->pi, N*sizeof(double), cudaMemcpyDeviceToHost), "cudaMemcpy pi result" );
    }
  }
  if(qr!=NULL) {
    fqr=(double*)malloc(NE*sizeof(double));
    cudasafe( cudaMemcpy(fqr, GPUvars->qr, NE*sizeof(double), cudaMemcpyDeviceToHost), "cudaMemcpy qr result" );
    if(qi!=NULL) {
      fqi=(double*)malloc(NE*sizeof(double));
      cudasafe( cudaMemcpy(fqi, GPUvars->qi, NE*sizeof(double), cudaMemcpyDeviceToHost), "cudaMemcpy qi result" );
    }
  }

  if(pr!=NULL) {
    if(pi!=NULL) {
      for(int i=0;i<N;i++) {
        pr[ix[i]]+=(double)fpr[i];
        pi[ix[i]]+=(double)fpi[i];
      }
      free(fpi);
    }
    else {
      for(int i=0;i<N;i++) {
        pr[ix[i]]+=(double)fpr[i];
      }
    }
    free(fpr);
  }
  if(qr!=NULL) {
    if(qi!=NULL) {
      for(int i=0;i<NE;i++) {
        qr[jx[i]]+=(double)fqr[i];
        qi[jx[i]]+=(double)fqi[i];
      }
      free(fqi);
    }
    else {
      for(int i=0;i<NE;i++) {
        qr[jx[i]]+=(double)fqr[i];
      }
    }
    free(fqr);
  }
    
            
  //unbind all textures and deallocate all GPU variables.
  cudaUnbindTexture(tzr);
  cudaUnbindTexture(tzi);
  cudaUnbindTexture(tmr);
  if(GPUvars->mi!=NULL){
    cudaUnbindTexture(tmi);
    cudaFreeDebug(GPUvars->mi);
  }
  cudaFreeDebug(GPUvars->zr);
  cudaFreeDebug(GPUvars->zi);
  if(pr!=NULL) {
    cudaFreeDebug(GPUvars->pr);
    if(pi!=NULL)
      cudaFreeDebug(GPUvars->pi);
  }
  if(qr!=NULL) {
    cudaFreeDebug(GPUvars->qr);
    if(qi!=NULL)
      cudaFreeDebug(GPUvars->qi);
  }

  cudaFreeDebug(GPUvars->mr);
            
  cudaUnbindTexture(tixptr);
  cudaFreeDebug(GPUvars->ixptr);
  cudaUnbindTexture(tjxptr);
  if(GPUvars->er!=NULL) {
    cudaFreeDebug(GPUvars->jxptr);
  }

  cudaUnbindTexture(ter);
  cudaUnbindTexture(tei);
  if(GPUvars->er!=NULL) {        
    cudaFreeDebug(GPUvars->er);
    cudaFreeDebug(GPUvars->ei);
  }
  if(qr!=NULL)
    cudaFreeDebug(GPUvars->jx);
  if(pr!=NULL)
    cudaFreeDebug(GPUvars->ix);

  // Device jcptr is allocated in MPexp_setup_cuda
  cudaFreeDebug(GPUvars->connect->jcptr); 
  // GPUvars->connect allocated in cuda_evals_direct
  free(GPUvars->connect);  
  checkcudaerror("Check after final cuda call");
}
/*------------------------------------------------------------------------*/

//direct evaluation (tol==0). This function is independent of the others and
//performs its own memory allocation, cuda calls cleanup etc.
void allpair_direct_cuda(int N,
                      const double *zr,const double *zi,
                      const double *mr,const double *mi,
                      int NE,
                      const double *er,const double *ei,
                      double *pr,double *pi,
                      double *qr,double *qi,int pot,
                      SMOOTHER smooth,double xopt,double cutoff, bool cont)
/* Evaluation of the potential at the points (er,ei). Function is
   critical to performance. */
{
  cudavariables ctmp;
  cudavariables *GPUvars=&ctmp; //simply to always use GPUvars as a pointer
  double shape, scale;
  if(N==0) /*no potential points, code will give Unspecified Driver Error otherwise*/
    return;
  checkcudaerror("Check eval_cuda start");
            
  //start by making necessary allocations
  if(qr!=NULL) {
    cudasafeMalloc((void**)&GPUvars->er,NE*sizeof(double));
    cudasafeMalloc((void**)&GPUvars->ei,NE*sizeof(double));
    cudasafeMalloc((void**)&GPUvars->qr,NE*sizeof(double));
    //         cudasafe(cudaBindTexture(NULL,ter,GPUvars->er,NE*sizeof(double)),"cudaBindTexture ter");
    //         cudasafe(cudaBindTexture(NULL,tei,GPUvars->ei,NE*sizeof(double)),"cudaBindTexture tei");
    cudasafe( cudaMemcpy(GPUvars->er, er, NE*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy er" );
    cudasafe( cudaMemcpy(GPUvars->ei, ei, NE*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy ei" );
    if(pot!=0||mi!=NULL)
      cudasafeMalloc((void**)&GPUvars->qi,NE*sizeof(double));
  }
  else {
    GPUvars->er=NULL;
    GPUvars->ei=NULL;
  }
  if(pr!=NULL) {
    cudasafeMalloc((void**)&GPUvars->pr,N*sizeof(double));
    if(pot!=0||mi!=NULL) {
      cudasafeMalloc((void**)&GPUvars->pi,N*sizeof(double));
    }
  }
  cudasafeMalloc((void**)&GPUvars->zr,N*sizeof(double));
  cudasafeMalloc((void**)&GPUvars->zi,N*sizeof(double));
  cudasafeMalloc((void**)&GPUvars->mr,N*sizeof(double));
    
  if(mi!=NULL) {
    cudasafeMalloc((void**)&GPUvars->mi,N*sizeof(double));
    cudasafe(cudaBindTexture(NULL,tmi,GPUvars->mi,N*sizeof(double)),"cudaBindTexture tmi");
    cudasafe( cudaMemcpy(GPUvars->mi, mi, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy mi" );
  }
  else
    GPUvars->mi=NULL;

  MPexp_smooth_(pot, smooth, xopt, cont, &cutoff, &shape, &scale);


  cudasafe( cudaMemcpy(GPUvars->zr, zr, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy zr" );
  cudasafe( cudaMemcpy(GPUvars->zi, zi, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy zi" );
  cudasafe( cudaMemcpy(GPUvars->mr, mr, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy mr" );

  cudasafe(cudaBindTexture(NULL,tzr,GPUvars->zr,N*sizeof(double)),"cudaBindTexture tzr");
  cudasafe(cudaBindTexture(NULL,tzi,GPUvars->zi,N*sizeof(double)),"cudaBindTexture tzi");
  cudasafe(cudaBindTexture(NULL,tmr,GPUvars->mr,N*sizeof(double)),"cudaBindTexture tmr");


  //move textures to make it possible to use same code in both cases.
  if(pr!=NULL) {
    cudasafe(cudaBindTexture(NULL,ter,GPUvars->zr,N*sizeof(double)),"cudaBindTexture ter");
    cudasafe(cudaBindTexture(NULL,tei,GPUvars->zi,N*sizeof(double)),"cudaBindTexture tei");
    allpairdirectkernellaunch(mi,GPUvars->pr,GPUvars->pi,N,N,pot,smooth,cutoff,shape,scale);
  }
  if(qr!=NULL) {
    if(pr!=NULL) {
      cudaUnbindTexture(ter);
      cudaUnbindTexture(tei);
    }
    cudasafe(cudaBindTexture(NULL,ter,GPUvars->er,NE*sizeof(double)),"cudaBindTexture ter");
    cudasafe(cudaBindTexture(NULL,tei,GPUvars->ei,NE*sizeof(double)),"cudaBindTexture tei");
    allpairdirectkernellaunch(mi,GPUvars->qr,GPUvars->qi,N,NE,pot,smooth,cutoff,shape,scale);
  }

  //start the cleanup    
  if(qr!=NULL) {
    cudasafe( cudaMemcpy(qr,GPUvars->qr, NE*sizeof(double), cudaMemcpyDeviceToHost), "cudaMemcpy qr" );
    //         cudaUnbindTexture(ter);
    //         cudaUnbindTexture(tei);
    cudaFreeDebug(GPUvars->er);
    cudaFreeDebug(GPUvars->ei);
    cudaFreeDebug(GPUvars->qr);
    if(pot!=0||mi!=NULL) {
      cudasafe( cudaMemcpy(qi,GPUvars->qi, NE*sizeof(double), cudaMemcpyDeviceToHost), "cudaMemcpy qi" );
      cudaFreeDebug(GPUvars->qi);
    }
  }
  if(pr!=NULL) {
    cudasafe( cudaMemcpy(pr,GPUvars->pr, N*sizeof(double), cudaMemcpyDeviceToHost), "cudaMemcpy pr" );
    cudaFreeDebug(GPUvars->pr);
    if(pot!=0||mi!=NULL) {
      cudasafe( cudaMemcpy(pi,GPUvars->pi, N*sizeof(double), cudaMemcpyDeviceToHost), "cudaMemcpy pr" );
      cudaFreeDebug(GPUvars->pi);
    }
  }
  cudaUnbindTexture(tzr);
  cudaUnbindTexture(tzi);
  cudaUnbindTexture(tmr);
  cudaUnbindTexture(ter);
  cudaUnbindTexture(tei);
    
  cudaFreeDebug(GPUvars->zr);
  cudaFreeDebug(GPUvars->zi);
  cudaFreeDebug(GPUvars->mr);
    
  if(mi!=NULL) {
    cudaUnbindTexture(tmi);
    cudaFreeDebug(GPUvars->mi);
  }  
}
/*------------------------------------------------------------------------*/
//selects correct direct summation kernel and starts it
void allpairdirectkernellaunch(const double *mi,double *GPUqr,double *GPUqi,
                           int N,int NE,int pot,SMOOTHER smooth,double cutoff,
                           double shape,double scale)
{
//printf("%d %d=(%d+%d-1)/%d %d\n",MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS, NE,DIRECTMAXTHREADS,DIRECTMAXTHREADS,DIRECTMAXTHREADS);
  if(pot==1) {
    if(mi==NULL) {
      switch(smooth) { //one kernel for each type. This is to reduce the number of local variables for each kernel
      case DIRAC:
        direct_pot1_nomi_dirac_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr, GPUqi,N,NE);
        break;
      case RANKINE:
        direct_pot1_nomi_rankine_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr, GPUqi,N,NE, shape);
        break;
      case SCULLY:
        direct_pot1_nomi_scully_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr, GPUqi,N,NE, shape, scale, cutoff);
        break;
      case OSEEN:
        direct_pot1_nomi_oseen_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr, GPUqi,N,NE, shape, scale, cutoff);
        break;
      }
    }
    else {
      switch(smooth) { //one kernel for each type. This is to reduce the number of local variables for each kernel
      case DIRAC:
        direct_pot1_mi_dirac_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr, GPUqi,N,NE);
        break;
      case RANKINE:

        direct_pot1_mi_rankine_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr, GPUqi,N,NE, shape);
        break;
      case SCULLY:
        direct_pot1_mi_scully_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr, GPUqi,N,NE, shape, scale, cutoff);
        break;
      case OSEEN:
        direct_pot1_mi_oseen_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr, GPUqi,N,NE, shape, scale, cutoff);
        break;
      }
    }
  }
  else {
    if(mi==NULL) {
      switch(smooth) { //one kernel for each type. This is to reduce the number of local variables for each kernel
      case DIRAC:
        direct_pot0_nomi_dirac_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr,N,NE);
        break;
      case RANKINE:
        direct_pot0_nomi_rankine_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr,N,NE, shape, cutoff);
        break;
      case SCULLY:
        direct_pot0_nomi_scully_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr,N,NE, shape, scale, cutoff);
        break;
      case OSEEN:
        direct_pot0_nomi_oseen_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr,N,NE, shape, scale, cutoff);
        break;
      }
    }
    else {
      switch(smooth) { //one kernel for each type. This is to reduce the number of local variables for each kernel
      case DIRAC:
        direct_pot0_mi_dirac_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr, GPUqi,N,NE);
        break;
      case RANKINE:
        direct_pot0_mi_rankine_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr, GPUqi,N,NE, shape, cutoff);
        break;
      case SCULLY:
        direct_pot0_mi_scully_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr, GPUqi,N,NE, shape, scale, cutoff);
        break;
      case OSEEN:
        direct_pot0_mi_oseen_synchronized<<<imin(MAXBLOCKS, (NE+DIRECTMAXTHREADS-1)/DIRECTMAXTHREADS), DIRECTMAXTHREADS>>>(GPUqr, GPUqi,N,NE, shape, scale, cutoff);
        break;
      }
    }
  }
  checkcudaerror("Kernel launch error direct interact");
}
/*------------------------------------------------------------------------*/
void cudaSafeMemcpy(void *dst,const void *src,size_t count,
                    enum cudaMemcpyKind kind,const char *str)
// wrapper
{
  cudasafe(cudaMemcpy(dst,src,count,kind),str);
}
/*------------------------------------------------------------------------*/
void cudaSafeMemset(void *dst,int value,size_t count,const char *str)
// wrapper
{
  cudasafe(cudaMemset(dst,value,count),str);
}
/*------------------------------------------------------------------------*/
cudaError_t cudaGetLastErrorDummy()
// wrapper
{
  return cudaGetLastError();
}
/*------------------------------------------------------------------------*/
cudaError_t cudaMallocDebug(void **addr,size_t size) 
// GPU memory allocation debug, can track memory leaks
{
  cudaalloccount++;
  cudaError_t ret = cudaMalloc(addr,size);
  if (*addr == NULL) cudaalloccount--;
  return ret;
}
/*------------------------------------------------------------------------*/
cudaError_t cudaFreeDebug(void* addr)
// corresponding to CudaMallocDebug, all allocations should be
// performed through these calls
{
  if (addr != NULL) cudaalloccount--;
  return cudaFree(addr);
}
/*------------------------------------------------------------------------*/
void printcudaalloccount(char *message)
// for tracking memory leaks
{
  printf("%s %d memory allocations on GPU.\n",message,cudaalloccount);
}
/*------------------------------------------------------------------------*/
int getcudaalloccount()
// since the variable currently only is declared in this file.
{
  return cudaalloccount;
}
/*------------------------------------------------------------------------*/
void cudasafe(cudaError_t error,const char* message)
// checks if a cuda function succeded, if not, print error message and
// abort (Cuda functions fail a lot more often than normal cpu
// functions. it is important to check the return values)
{
  char errormessage[1000];
  if (error != cudaSuccess) {
    snprintf(errormessage,999,"%s : %s (%i) \n",message,
             cudaGetErrorString(error),error);
    cudaGetLastError(); // remove the error message
    fprintf(stderr,errormessage); 
  }
}
/*------------------------------------------------------------------------*/
void cudasafeMalloc(void** addr,size_t size)
// safe memory allocation, cudasafe(cudaMalloc(...),"message") can
// also be used
{
  char errormessage[1000];
  cudaError_t error = cudaMallocDebug(addr,size);
  if(error != cudaSuccess) {
    sprintf(errormessage,"cudaMalloc (%u bytes): %s (%i) \n",
            (unsigned)size,cudaGetErrorString(error),error);
    cudaGetLastError(); //remove the error message
    fprintf(stderr,errormessage); 
  }
}
/*------------------------------------------------------------------------*/
void checkcudaerror(const char *message)
{
  //only check for error and abort if error is found
  char errormessage[1000];
  cudaError_t error=cudaGetLastError();
  if(error!=cudaSuccess) {
    sprintf(errormessage,"%s : %s (%i) \n",message,
            cudaGetErrorString(error),error);
    fprintf(stderr,errormessage); 
  }
}
/*------------------------------------------------------------------------*/
void *cudaMalloc2(size_t len)
{
    void *p;
    if (cudaMalloc(&p, len) == cudaSuccess) return p;
    return 0;
}
/*------------------------------------------------------------------------*/

