/* cudaeval.h -- CUDA host code for direct evaluation. */

/* S. Engblom and A. Goude 2011-10-19 */

#ifndef __cudaeval_h
#define __cudaeval_h

#include "common.h"

#include "cuda.h"
#include "driver_types.h"
#include <stdio.h>


#define TEXTUREFETCH(dest,texref,index)                   \
  ((float *)&(dest))[0] = tex1Dfetch(texref,(index)*2);   \
  ((float *)&(dest))[1] = tex1Dfetch(texref,(index)*2+1);


// struct collecting all data to send to the GPU
typedef struct cudavariables_struct {
  int Nf;
  double *er,*ei,*zr,*zi,*mr,*mi,*qr,*qi,*pr,*pi;
  int *ixptr,*jxptr;
  cudaEvent_t start,stop;
  cudaEvent_t mpstart,mpstop;
  cudaEvent_t initstart,initstop;
  cudaEvent_t stepm2mstart,stepm2mstop;
  cudaEvent_t stepp2pstart,stepp2pstop;
  cudaEvent_t stepm2psstart,stepm2psstop;
  cudaEvent_t fulltimestart,fulltimestop;
  cudaEvent_t reorder1start,reorder1stop;
  cudaEvent_t reorder2start,reorder2stop;
  cudaEvent_t sortstart,sortstop;
  cudaEvent_t checkstart,checkstop;

  double *z0;
  double *d0;
  int *ix,*jx;
  // allocated on host memory, but internal variables on GPU memory
  mpSparse* connect;
  double *dabs;
} cudavariables;

//Keeps track of all memory allocations making it easy to spot memory leaks
//Make all allocations through these functions
cudaError_t cudaMallocDebug(void **addr,size_t size);
cudaError_t cudaFreeDebug(void *addr);
void printcudaalloccount(char *message);
int getcudaalloccount();

//Checks the error message of a cura call and aborts the code if a cuda call fails
void cudasafe(cudaError_t error,const char *message);
void cudasafeMalloc(void **addr,size_t size);
void *cudaMalloc2(size_t len);
void checkcudaerror(const char *message);

// CudaMemcpy and CudaMemset with automatic check of the return value. Aborts if call fails
void cudaSafeMemcpy(void *dst,const void *src,size_t count,
                    enum cudaMemcpyKind kind,const char *str);
void cudaSafeMemset(void *dst,int value,size_t count,const char* str);
cudaError_t cudaGetLastErrorDummy();

// Prototypes for functions called by the GPU initialization of multipole coefficients
void MPexp_init_cuda(const MPexp *This,cudavariables *GPUvars,
                     int pot,int complexpoint);
void cuda_check_init(MPexp *This,int Nt,cudavariables* GPUvars);

#ifdef __CUDACC__
__host__ __device__
#endif
inline int imin(int x,int y)
{
  return x < y ? x : y;
}
#ifdef  __CUDACC__
__host__ __device__
#endif
inline int imax(int x,int y)
{
  return x < y ? y : x;
}

void cuda_evals_direct(const MPexp *This, int N, int NE,
                double *pr,double *pi,
                double *qr,double *qi,
                const double *er,const double *ei,
                const double *zr,const double *zi,
                const double *mr,const double *mi,
                int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont);

void allpair_direct_cuda(int N,
		      const double *zr,const double *zi,
		      const double *mr,const double *mi,
		      int NE,
		      const double *er,const double *ei,
		      double *pr,double *pi,
		      double *qr,double *qi,int pot,
		      SMOOTHER smooth,double xopt,double cutoff,
                      bool cont);
                      
void allpairdirectkernellaunch(const double *mi,double *GPUqr,double *GPUqi,
                           int N,int NE,int pot,SMOOTHER smooth,double cutoff,
                           double shape,double scale);

void MPexp_setup_cuda(const MPexp *This,
                     const double *er,const double *ei,
                     const double *zr,const double *zi,
                     const double *mr,const double *mi,
                     const double *pr,cudavariables *GPUvars,int N,int NE,
                     int pot,SMOOTHER smooth,
                     double xopt,double cutoff,bool cont);
                    
void launchdirectkernel(const MPexp *This,int pot,const double *mi,
                        cudavariables *GPUvars,
                        double *GPUqr,double *GPUqi,
                        SMOOTHER smooth,double shape,
                        double scale,double cutoff);

void MPexp_eval_cuda_result(double *pr,double *pi,double *qr,double *qi,
                            int *ix,int *jx,int N,int NE,int pot,
                            cudavariables *GPUvars,
                            const MPexp *This);
                            
cudaError_t cudaMallocDebug(void **addr,size_t size) ;

cudaError_t cudaFreeDebug(void* addr);

void printcudaalloccount(char *message);

int getcudaalloccount();

void cudasafe(cudaError_t error,const char* message);

void cudasafeMalloc(void** addr,size_t size);

void checkcudaerror(const char *message);  

#endif /* __cudaeval_h */
