/* direct.c */

/* M. Holm 2012-10-10 Applied SuperGlue taskification */
/* S. Engblom 2010-02-18 */

#include <cstdlib> 
#include <stdio.h>

#include "expint.h"
#include "direct.h"

#include "tl.h"

#if DIRECTTHREADS == 1
#undef TASKLOG
#endif


// code responds to #define FMM_INLINECOMPLEX_DIRECT (see fmm.c)

void rlog2(int N,
	   const double *zr,const double *zi,
	   const double *mr,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont);
void zlog2(int N,
	   const double *zr,const double *zi,
	   const double *mr,const double *mi,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont);
void rlog(int N,
	  const double *zr,const double *zi,
	  const double *mr,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont);
void zlog(int N,
	  const double *zr,const double *zi,
	  const double *mr,const double *mi,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont);

void rinv2(int N,
	   const double *zr,const double *zi,
	   const double *mr,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont);
void zinv2(int N,
	   const double *zr,const double *zi,
	   const double *mr,const double *mi,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont);
void rinv(int N,
	  const double *zr,const double *zi,
	  const double *mr,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont);
void zinv(int N,
	  const double *zr,const double *zi,
	  const double *mr,const double *mi,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont);
	  


// Task parallelism routines

//
typedef struct {
  int index, Nf, pot, jstart, jend;
  const MPexp *This;
  double *pr, *pi, *qr, *qi;
  const double *zr, *er;
  const double *zi, *ei;
  const double *mr;
  const double *mi;
  SMOOTHER smooth;
  double xopt;
  double cutoff, shape, scale;
  bool cont;
} task_eval_args;

void task_direct_eval(tl_thread_info_t tinfo, void *data, void **ptr); 
void addTask_direct_eval(task_eval_args *args);
void task_direct_eval2(tl_thread_info_t tinfo, void *data, void **ptr); 	  
void addTask_direct_eval2(task_eval_args *args);

/*------------------------------------------------------------------------*/
void MPexp_smooth_(int pot,SMOOTHER smooth,double xopt,bool cont,
                   double *cutoff,double *shape,double *scale)
/* Precomputes the three parameters (cutoff,shape,scale) given
   (pot,smooth,xopt,cutoff,cont). This yields a slicker code for
   mpexp_directInteract(). */
{
  *scale = 1.0;

  if (pot == 0)
    switch (smooth) {
    case DIRAC:
      break;

    case RANKINE:
      *cutoff = xopt*xopt;
      *shape = 0.5*log(*cutoff)-0.5;
      break;

    case SCULLY:
      *cutoff = *cutoff**cutoff;
      *shape = xopt*xopt;
      if (cont)
        *scale = log(*cutoff)/log(*shape+*cutoff);
      break;

    case OSEEN:
    *cutoff = *cutoff**cutoff;
    *shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      *scale = log(*cutoff)/(log(*cutoff)+expint(*shape**cutoff));
      break;
    }
  else
    switch (smooth) {
    case DIRAC:
      break;

    case RANKINE:
      *cutoff = xopt*xopt;
      *shape = 1.0/(*cutoff);
      break;

    case SCULLY:
      *cutoff = *cutoff**cutoff;
      *shape = xopt*xopt;
      if (cont)
        *scale = 1.0+*shape/(*cutoff);
      break;

    case OSEEN:
      *cutoff = *cutoff**cutoff;
      *shape = 1.2564312086261696770/(xopt*xopt);
      if (cont)
        *scale = -1.0/expm1(-*shape**cutoff);
      break;
    }
}
/*------------------------------------------------------------------------*/	  
void MPexp_evals_direct(const MPexp *This, int N, int NE,
                double *pr,double *pi,
                double *qr,double *qi,
                const double *er,const double *ei,
                const double *zr,const double *zi,
                const double *mr,const double *mi,
                int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont)
{
   if (pr != NULL) MPexp_eval2_direct(This,pr,pi,zr,zi,mr,mi,
                              pot,smooth,xopt,cutoff,cont);

  // evaluation at coordinates (er,ei)
  if (qr != NULL) MPexp_eval_direct(This,qr,qi,er,ei,zr,zi,mr,mi,
                             pot,smooth,xopt,cutoff,cont);
}



/*------------------------------------------------------------------------*/
void task_direct_eval(tl_thread_info_t tinfo, void *data, void **ptr)
{
  // load data from struct
  task_eval_args *args = (task_eval_args *) data;
  const MPexp *This = args->This;
  double *pr = args->pr;
  double *pi = args->pi;
  double *qr = args->qr;
  double *qi = args->qi;
  const double *zr = args->zr;
  const double *zi = args->zi;
  const double *er = args->er;
  const double *ei = args->ei;
  const double *mr = args->mr;
  const double *mi = args->mi;
  int pot = args->pot;
  SMOOTHER smooth = args->smooth;
  double xopt = args->xopt;
  double cutoff = args->cutoff;
  bool cont = args->cont;
    
  // get start and end indexes
  int jstart = args->jstart;
  int jend = args->jend;
  
  // hopefully, some optimizing compiler can use this:
  const int Nf = args->Nf;
  const mpexp *restrict offset = &This->root[This->lptr[This->nlevel]];
  const int *restrict ix = This->ix,*restrict ixptr = This->ixptr;
  const mpSparse *C = &This->connect[This->nlevel];
  const int *restrict jx = This->jx,*restrict jxptr = This->jxptr;
  const int *restrict jcptr = C->jcptr,*restrict j2cptr = &C->kcptr[Nf],
    *restrict kcptr = C->kcptr,*restrict ir = C->ir;
  double shape = args->shape;
  double scale = args->scale;

  for (int j = jstart,jj; j < jend; j++) {    
    // near-field: symmetric interaction within list of near connections
    for (jj = jcptr[j]; jj < j2cptr[j]; jj++) {
      const int k = ir[jj];
      mpexp_directInteract(qr,qi,er,ei,
                           jx,jxptr[j],jxptr[j+1],
                           zr,zi,mr,mi,
                           ix,ixptr[k],ixptr[k+1],
                           pot,smooth,cutoff,shape,scale);

    }    
  } 
} 
/*------------------------------------------------------------------------*/
void addTask_direct_eval(task_eval_args *args)
{
/* no dependencies */
  char taskname[40];
  sprintf(taskname, "P2P task %d", args->index);
  tl_add_task(task_direct_eval, args, sizeof(*args), taskname);
}
/*------------------------------------------------------------------------*/
void task_direct_eval2(tl_thread_info_t tinfo, void *data, void **ptr)
{
  // load data from struct
  task_eval_args *args = (task_eval_args *) data;
  const MPexp *This = args->This;
  double *pr = args->pr;
  double *pi = args->pi;
  const double *zr = args->zr;
  const double *zi = args->zi;
  const double *mr = args->mr;
  const double *mi = args->mi;
  int pot = args->pot;
  SMOOTHER smooth = args->smooth;
  double xopt = args->xopt;
  double cutoff = args->cutoff;
  bool cont = args->cont;
    
  // get start and end indexes
  int jstart = args->jstart;
  int jend = args->jend;
  
  // hopefully, some optimizing compiler can use this:
  const int Nf = args->Nf;
  const mpexp *restrict offset = &This->root[This->lptr[This->nlevel]];
  const int *restrict ix = This->ix,*restrict ixptr = This->ixptr;
  const mpSparse *C = &This->connect[This->nlevel];
  const int *restrict jcptr = C->jcptr,*restrict j2cptr = &C->kcptr[Nf],
    *restrict kcptr = C->kcptr,*restrict ir = C->ir;
  double shape = args->shape;
  double scale = args->scale;

  for (int j = jstart,jj; j < jend; j++) {
    // near-field: asymmetric interaction within list of near connections
    for (jj = jcptr[j]; jj < j2cptr[j]; jj++) {
      const int k = ir[jj];
      if (k != j) {  
        mpexp_directInteract2(pr,pi,zr,zi,mr,mi,ix,
                        ixptr[j],ixptr[j+1],
                        ixptr[k],ixptr[k+1],
                        pot,smooth,cutoff,shape,scale);                                                
      }
    }

    // near-field: direct interaction within box (done here instead of above  b/c k!=j)
    mpexp_directInteract2(pr,pi,zr,zi,mr,mi,ix,ixptr[j],ixptr[j+1],
                          -1,ixptr[j+1],pot,smooth,cutoff,shape,scale);
                          
  } /* end j-loop */
} 	
/*------------------------------------------------------------------------*/  
void addTask_direct_eval2(task_eval_args *args)
{
/* no dependencies */
  char taskname[40];
  sprintf(taskname, "P2P task %d", args->index);
  tl_add_task(task_direct_eval2, args, sizeof(*args), taskname);
}
/*------------------------------------------------------------------------*/

void MPexp_eval2_direct(const MPexp *This,
                 double *pr,double *pi,
                 const double *zr,const double *zi,
                 const double *mr,const double *mi,
                 int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Evaluation of the potential at the potentials themselves. Function
   is critical to performance. */
{
  // hopefully, some optimizing compiler can use this:
  const int Nf = This->lptr[This->nlevel+1]-This->lptr[This->nlevel];
  const mpexp *restrict offset = &This->root[This->lptr[This->nlevel]];
  const int *restrict ix = This->ix,*restrict ixptr = This->ixptr;
  const mpSparse *C = &This->connect[This->nlevel];
  const int *restrict jcptr = C->jcptr,*restrict j2cptr = &C->kcptr[Nf],
    *restrict kcptr = C->kcptr,*restrict ir = C->ir;
  double shape,scale;

  MPexp_smooth_(pot,smooth,xopt,cont,&cutoff,&shape,&scale); 
  
  if(DIRECTTHREADS > 1)
  {
     int numtasks = (DIRECTTHREADS*16>Nf)?DIRECTTHREADS*4:DIRECTTHREADS*16;
     numtasks = (numtasks > Nf)?Nf:numtasks;
     task_eval_args args[numtasks];
     // loop to create tasks
     for (int i = 0; i < numtasks; i++)
     {
        args[i].index = i;
        args[i].This = This; args[i].pr = pr; args[i].pi = pi;
        args[i].zr = zr; args[i].zi = zi; args[i].mr = mr; args[i].mi = mi;
        args[i].pot = pot; args[i].smooth = smooth; args[i].xopt = xopt;
        args[i].cutoff = cutoff; args[i].cont = cont;
        args[i].Nf = Nf; 
        args[i].shape = shape; args[i].scale = scale;
        args[i].jstart = (Nf/(numtasks-1))*i;
        args[i].jend = min((Nf/(numtasks-1))*(i+1),Nf); // the last task may be very small
        addTask_direct_eval2(&args[i]);
     }
     tl_barrier();
  }else{
    for (int j = 0,jj; j < Nf; j++) {
      // near-field: asymmetric interaction within list of near connections
      for (jj = jcptr[j]; jj < j2cptr[j]; jj++) {
        const int k = ir[jj];
        if (k != j) {  
          mpexp_directInteract2(pr,pi,zr,zi,mr,mi,ix,
                        ixptr[j],ixptr[j+1],
                        ixptr[k],ixptr[k+1],
                        pot,smooth,cutoff,shape,scale);                                                
        }
      }

      // near-field: direct interaction within box (done here instead of above  b/c k!=j)
      mpexp_directInteract2(pr,pi,zr,zi,mr,mi,ix,ixptr[j],ixptr[j+1],
                          -1,ixptr[j+1],pot,smooth,cutoff,shape,scale);
                          
    } // end j-loop 
  } // endif DIRECTTHREADS > 1 
}
/*------------------------------------------------------------------------*/
void MPexp_eval_direct(const MPexp *This,
                double *qr,double *qi,
                const double *er,const double *ei,
                const double *zr,const double *zi,
                const double *mr,const double *mi,
                int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Evaluation of the potential at the points (er,ei). Function is
   critical to performance. */
{
  const int Nf = This->lptr[This->nlevel+1]-This->lptr[This->nlevel];
  const mpexp *restrict offset = &This->root[This->lptr[This->nlevel]];
  const int *restrict ix = This->ix,*restrict ixptr = This->ixptr;
  const int *restrict jx = This->jx,*restrict jxptr = This->jxptr;
  const mpSparse *C = &This->connect[This->nlevel];
  const int *restrict jcptr = C->jcptr,*restrict j2cptr = &C->kcptr[Nf],
    *restrict kcptr = C->kcptr,*restrict ir = C->ir;
  double shape,scale;

  MPexp_smooth_(pot,smooth,xopt,cont,&cutoff,&shape,&scale);
  if(DIRECTTHREADS > 1)
  {
     int numtasks = (DIRECTTHREADS*16>Nf)?DIRECTTHREADS*4:DIRECTTHREADS*16;
     numtasks = (numtasks > Nf)?Nf:numtasks;
 
     task_eval_args args[numtasks];
     // loop to create tasks
     for (int i = 0; i < numtasks; i++)
     {
        args[i].index = i;
        args[i].This = This;
        args[i].qr = qr; args[i].qi = qi;
        args[i].er = er; args[i].ei = ei;
        args[i].zr = zr; args[i].zi = zi; args[i].mr = mr; args[i].mi = mi;
        args[i].pot = pot; args[i].smooth = smooth; args[i].xopt = xopt;
        args[i].cutoff = cutoff; args[i].cont = cont;
        args[i].Nf = Nf; 
        args[i].shape = shape; args[i].scale = scale;
        args[i].jstart = (Nf/(numtasks-1))*i;
        args[i].jend = min((Nf/(numtasks-1))*(i+1),Nf); // the last task may be very small
        addTask_direct_eval(&args[i]);
     }
     tl_barrier();
  }else{
    for (int j = 0,jj; j < Nf; j++) {    
      // near-field: symmetric interaction within list of near connections
      for (jj = jcptr[j]; jj < j2cptr[j]; jj++) {
        const int k = ir[jj];
        mpexp_directInteract(qr,qi,er,ei,
                           jx,jxptr[j],jxptr[j+1],
                           zr,zi,mr,mi,
                           ix,ixptr[k],ixptr[k+1],
                           pot,smooth,cutoff,shape,scale);

      }    
    }   
  } // endif DIRECTTHREADS > 1 
}
/*------------------------------------------------------------------------*/
#define BEGIN(begin,i) (begin == -1 ? i+1 : begin)

void mpexp_directInteract2(double *pr,double *pi,
                           const double *zr,const double *zi,
                           const double *mr,const double *mi,
                           const int *ix,
                           int begin1,int end1,int begin2,int end2,
                           int pot,SMOOTHER smooth,
                           double cutoff,double shape,double scale)
/* Direct interaction between potentials inside a box (begin2 == -1)
   or between two boxes. */
{
  if (pot == 0) { // logaritmic potential
    if (mi == NULL) { // rlog2()
      switch (smooth) {
      case DIRAC:
        for (int i = begin1; i < end1; i++)
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double P = 0.5*log(re*re+im*im);
            pr[ix[i]] -= mr[ix[j]]*P;
            if(begin2 == -1)
                    pr[ix[j]] -= mr[ix[i]]*P;
          }
        break;

      case RANKINE:
        for (int i = begin1; i < end1; i++) {
          if (begin2 == -1)
            pr[ix[i]] -= mr[ix[i]]*shape;
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 0.5*log(ad);
            else
              P = 0.5*(ad/cutoff)+shape;
            pr[ix[i]] -= mr[ix[j]]*P;
                if(begin2 == -1)
                    pr[ix[j]] -= mr[ix[i]]*P;
          }
        }
        break;

      case SCULLY:
        for (int i = begin1; i < end1; i++) {
          if (begin2 == -1)
            pr[ix[i]] -= mr[ix[i]]*0.5*log(shape)*scale;
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 0.5*log(ad);
            else
              P = 0.5*log(shape+ad)*scale;
            pr[ix[i]] -= mr[ix[j]]*P;
            if(begin2 == -1)
                    pr[ix[j]] -= mr[ix[i]]*P;
          }
        }
        break;

      case OSEEN:
        for (int i = begin1; i < end1; i++) {
          if (begin2 == -1)
            pr[ix[i]] += mr[ix[i]]*0.5*(log(shape)+gamma_)*scale;
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 0.5*log(ad);
            else if (ad != 0.0)
              P = 0.5*(log(ad)+expint(shape*ad))*scale;
            else
              P = -0.5*(log(shape)+gamma_)*scale;
            pr[ix[i]] -= mr[ix[j]]*P;
            if(begin2 == -1)
                    pr[ix[j]] -= mr[ix[i]]*P;
          }
        }
        break;
      }
    }
    else { // zlog2()
      switch (smooth) {
      case DIRAC:
        for (int i = begin1; i < end1; i++)
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double P = 0.5*log(re*re+im*im);
            pr[ix[i]] -= mr[ix[j]]*P;
            pi[ix[i]] -= mi[ix[j]]*P;
            if(begin2 == -1){
                    pr[ix[j]] -= mr[ix[i]]*P;
                pi[ix[j]] -= mi[ix[i]]*P;
            }
          }
        break;

      case RANKINE:
        for (int i = begin1; i < end1; i++) {
          if (begin2 == -1) {
            pr[ix[i]] -= mr[ix[i]]*shape;
            pi[ix[i]] -= mi[ix[i]]*shape;
          }
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 0.5*log(ad);
            else
              P = 0.5*(ad/cutoff)+shape;
            pr[ix[i]] -= mr[ix[j]]*P;
            pi[ix[i]] -= mi[ix[j]]*P;
            if(begin2 == -1){
                    pr[ix[j]] -= mr[ix[i]]*P;
                        pi[ix[j]] -= mi[ix[i]]*P;
            }
          }
        }
        break;

      case SCULLY:
        for (int i = begin1; i < end1; i++) {
          if (begin2 == -1) {
            pr[ix[i]] -= mr[ix[i]]*0.5*log(shape)*scale;
            pi[ix[i]] -= mi[ix[i]]*0.5*log(shape)*scale;
          }
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 0.5*log(ad);
            else
              P = 0.5*log(shape+ad)*scale;
            pr[ix[i]] -= mr[ix[j]]*P;
            pi[ix[i]] -= mi[ix[j]]*P;
            if(begin2 == -1){
                    pr[ix[j]] -= mr[ix[i]]*P;
                        pi[ix[j]] -= mi[ix[i]]*P;
            }
          }
        }
        break;

      case OSEEN:
        for (int i = begin1; i < end1; i++) {
          if (begin2 == -1) {
            pr[ix[i]] += mr[ix[i]]*0.5*(log(shape)+gamma_)*scale;
            pi[ix[i]] += mi[ix[i]]*0.5*(log(shape)+gamma_)*scale;
          }
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 0.5*log(ad);
            else if (ad != 0.0)
              P = 0.5*(log(ad)+expint(shape*ad))*scale;
            else
              P = -0.5*(log(shape)+gamma_)*scale;
            pr[ix[i]] -= mr[ix[j]]*P;
            pi[ix[i]] -= mi[ix[j]]*P;
            if(begin2 == -1){
                    pr[ix[j]] -= mr[ix[i]]*P;
                        pi[ix[j]] -= mi[ix[i]]*P;
            }
          }
        }
        break;
      }
    }
  }
  else { // harmonic potential
    if (mi == NULL) { // rinv2()
      switch (smooth) {
      case DIRAC:
        for (int i = begin1; i < end1; i++)
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double P = 1.0/(re*re+im*im);
            pr[ix[i]] -= mr[ix[j]]*P*re;
            pi[ix[i]] += mr[ix[j]]*P*im;
            if(begin2 == -1){
                    pr[ix[j]] += mr[ix[i]]*P*re;
                        pi[ix[j]] -= mr[ix[i]]*P*im;
            }
          }
        break;

      case RANKINE:
        for (int i = begin1; i < end1; i++)
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 1.0/ad;
            else
              P = shape;
            pr[ix[i]] -= mr[ix[j]]*P*re;
            pi[ix[i]] += mr[ix[j]]*P*im;
            if(begin2 == -1){
                    pr[ix[j]] += mr[ix[i]]*P*re;
                        pi[ix[j]] -= mr[ix[i]]*P*im;
            }
          }
        break;

      case SCULLY:
        for (int i = begin1; i < end1; i++)
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 1.0/ad;
            else
              P = 1.0/(shape+ad)*scale;
            pr[ix[i]] -= mr[ix[j]]*P*re;
            pi[ix[i]] += mr[ix[j]]*P*im;
            if(begin2 == -1){
                    pr[ix[j]] += mr[ix[i]]*P*re;
                        pi[ix[j]] -= mr[ix[i]]*P*im;
            }
          }
        break;

      case OSEEN:
        for (int i = begin1; i < end1; i++)
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad != 0.0) {
              if (ad >= cutoff)
                        P = 1.0/ad;
              else
                        P = -expm1(-shape*ad)/ad*scale;
              pr[ix[i]] -= mr[ix[j]]*P*re;
              pi[ix[i]] += mr[ix[j]]*P*im;
              if(begin2 == -1){
                    pr[ix[j]] += mr[ix[i]]*P*re;
                        pi[ix[j]] -= mr[ix[i]]*P*im;
              }
            }
          }
        break;
      }
    }
    else { // zinv2()
      switch (smooth) {
      case DIRAC:
        for (int i = begin1; i < end1; i++)
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double P = 1.0/(re*re+im*im);
            pr[ix[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
            pi[ix[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
            if(begin2 == -1){
                    pr[ix[j]] += (mr[ix[i]]*re+mi[ix[i]]*im)*P;
                        pi[ix[j]] -= (mr[ix[i]]*im-mi[ix[i]]*re)*P;
            }
          }
        break;

      case RANKINE:
        for (int i = begin1; i < end1; i++)
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 1.0/ad;
            else
              P = shape;
            pr[ix[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
            pi[ix[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
            if(begin2 == -1){
                    pr[ix[j]] += (mr[ix[i]]*re+mi[ix[i]]*im)*P;
                        pi[ix[j]] -= (mr[ix[i]]*im-mi[ix[i]]*re)*P;
            }
          }
        break;

      case SCULLY:
        for (int i = begin1; i < end1; i++)
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 1.0/ad;
            else
              P = 1.0/(shape+ad)*scale;
            pr[ix[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
            pi[ix[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
            if(begin2 == -1){
                    pr[ix[j]] += (mr[ix[i]]*re+mi[ix[i]]*im)*P;
                        pi[ix[j]] -= (mr[ix[i]]*im-mi[ix[i]]*re)*P;
            }
          }
        break;

      case OSEEN:
        for (int i = begin1; i < end1; i++)
          for (int j = BEGIN(begin2,i); j < end2; j++) {
            const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad != 0.0) {
              if (ad >= cutoff)
                        P = 1.0/ad;
              else
                        P = -expm1(-shape*ad)/ad*scale;
              pr[ix[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
              pi[ix[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P; 
              if(begin2 == -1){
                    pr[ix[j]] += (mr[ix[i]]*re+mi[ix[i]]*im)*P;
                        pi[ix[j]] -= (mr[ix[i]]*im-mi[ix[i]]*re)*P;
              }
            }
          }
        break;
      }
    }
  }
}
/*------------------------------------------------------------------------*/
void mpexp_directInteract(double *qr,double *qi,
                          const double *er,const double *ei,
                          const int *jx,int begin1,int end1,
                          const double *zr,const double *zi,
                          const double *mr,const double *mi,
                          const int *ix,int begin2,int end2,
                          int pot,SMOOTHER smooth,
                          double cutoff,double shape,double scale)
/* Direct evaluation of field from potentials inside one box computed
   at points (er,ei) inside another box. */
{
  if (pot == 0) { // logarithmic potential
    if (mi == NULL) { // rlog()
      switch (smooth) {
      case DIRAC:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++)
            if (er[jx[i]] != zr[ix[j]] || ei[jx[i]] != zi[ix[j]]) {
              const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
              const double P = 0.5*log(re*re+im*im);
              qr[jx[i]] -= mr[ix[j]]*P;
            }
        break;

      case RANKINE:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++) {
            const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 0.5*log(ad);
            else
              P = 0.5*(ad/cutoff)+shape;
            qr[jx[i]] -= mr[ix[j]]*P;
          }
        break;

      case SCULLY:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++) {
            const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 0.5*log(ad);
            else
              P = 0.5*log(shape+ad)*scale;
            qr[jx[i]] -= mr[ix[j]]*P;
          }
        break;

      case OSEEN:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++) {
            const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 0.5*log(ad);
            else if (ad != 0.0)
              P = 0.5*(log(ad)+expint(shape*ad))*scale;
            else
              P = -0.5*(log(shape)+gamma_)*scale;
            qr[jx[i]] -= mr[ix[j]]*P;
          }
        break;
      }
    }
    else { // zlog()
      switch (smooth) {
      case DIRAC:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++)
            if (er[jx[i]] != zr[ix[j]] || ei[jx[i]] != zi[ix[j]]) {
              const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
              const double P = 0.5*log(re*re+im*im);
              qr[jx[i]] -= mr[ix[j]]*P;
              qi[jx[i]] -= mi[ix[j]]*P;
            }
        break;

      case RANKINE:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++) {
            const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 0.5*log(ad);
            else
              P = 0.5*(ad/cutoff)+shape;
            qr[jx[i]] -= mr[ix[j]]*P;
            qi[jx[i]] -= mi[ix[j]]*P;
          }
        break;

      case SCULLY:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++) {
            const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 0.5*log(ad);
            else
              P = 0.5*log(shape+ad)*scale;
            qr[jx[i]] -= mr[ix[j]]*P;
            qi[jx[i]] -= mi[ix[j]]*P;
          }
        break;

      case OSEEN:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++) {
            const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 0.5*log(ad);
            else if (ad != 0.0)
              P = 0.5*(log(ad)+expint(shape*ad))*scale;
            else
              P = -0.5*(log(shape)+gamma_)*scale;
            qr[jx[i]] -= mr[ix[j]]*P;
            qi[jx[i]] -= mi[ix[j]]*P;
          }
        break;
      }
    }
  }
  else { // harmonic potential
    if (mi == NULL) { // rinv()
      switch (smooth) {
      case DIRAC:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++)
            if (er[jx[i]] != zr[ix[j]] || ei[jx[i]] != zi[ix[j]]) {
              const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
              const double P = 1.0/(re*re+im*im);
              qr[jx[i]] -= mr[ix[j]]*P*re;
              qi[jx[i]] += mr[ix[j]]*P*im;
            }
        break;

      case RANKINE:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++) {
            const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 1.0/ad;
            else
              P = shape;
            qr[jx[i]] -= mr[ix[j]]*P*re;
            qi[jx[i]] += mr[ix[j]]*P*im;
          }
        break;

      case SCULLY:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++) {
            const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 1.0/ad;
            else
              P = 1.0/(shape+ad)*scale;
            qr[jx[i]] -= mr[ix[j]]*P*re;
            qi[jx[i]] += mr[ix[j]]*P*im;
          }
        break;

      case OSEEN:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++) {
            const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad != 0.0) {
              if (ad >= cutoff)
                P = 1.0/ad;
              else
                P = -expm1(-shape*ad)/ad*scale;
              qr[jx[i]] -= mr[ix[j]]*P*re;
              qi[jx[i]] += mr[ix[j]]*P*im;
            }
          }
        break;
      }
    }
    else { // zinv()
      switch (smooth) {
      case DIRAC:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++)
            if (er[jx[i]] != zr[ix[j]] || ei[jx[i]] != zi[ix[j]]) {
              const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
              const double P = 1.0/(re*re+im*im);
              qr[jx[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
              qi[jx[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
            }
        break;

      case RANKINE:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++) {
            const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 1.0/ad;
            else
              P = shape;
            qr[jx[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
            qi[jx[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
          }
        break;

      case SCULLY:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++) {
            const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad >= cutoff)
              P = 1.0/ad;
            else
              P = 1.0/(shape+ad)*scale;
            qr[jx[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
            qi[jx[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
          }
        break;

      case OSEEN:
        for (int i = begin1; i < end1; i++)
          for (int j = begin2; j < end2; j++) {
            const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
            const double ad = re*re+im*im;
            double P;
            if (ad != 0.0) {
              if (ad >= cutoff)
                P = 1.0/ad;
              else
                P = -expm1(-shape*ad)/ad*scale;
              qr[jx[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
              qi[jx[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
            }
          }
        break;
      }
    }
  }
}

/*------------------------------------------------------------------------*/
void directInteract(int N,
                    const double *zr,const double *zi,
		    const double *mr,const double *mi,
		    int NE,
		    const double *er,const double *ei,
                    double *pr,double *pi,
		    double *qr,double *qi,int pot,
		    SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Direct evaluation of the potential from N pointmasses (mr,mi) (with
   mi possibly NULL) at the positions (zr,zi). The evaluation is done
   at the points (zr,zi) if (pr,pi) is not NULL and/or at the points
   (er,ei) if (qr,qi) is not NULL. The result is stored in the vectors
   (pr,pi) and (qr,qi), respectively, which must be allocated and
   cleared prior to call. Logarithmic potential for pot == 0, harmonic
   potential otherwise. The smoothing is done using smoother smooth
   with parameters (xopt,cutoff,cont). */
{
  // evaluation at (zr,zi)
  if (pr != NULL) {
    // mass is real
    if (mi == NULL)
      pot ? 
	rinv2(N,zr,zi,mr,pr,pi,smooth,xopt,cutoff,cont) : 
	rlog2(N,zr,zi,mr,pr,pi,smooth,xopt,cutoff,cont);
    // mass is complex
    else
      pot ? 
	zinv2(N,zr,zi,mr,mi,pr,pi,smooth,xopt,cutoff,cont) : 
	zlog2(N,zr,zi,mr,mi,pr,pi,smooth,xopt,cutoff,cont);
  }

  // evaluation at (er,ei)
  if (qr != NULL) {
    // mass is real
    if (mi == NULL)
      pot ? 
	rinv(N,zr,zi,mr,NE,er,ei,qr,qi,smooth,xopt,cutoff,cont) : 
	rlog(N,zr,zi,mr,NE,er,ei,qr,qi,smooth,xopt,cutoff,cont);
    // mass is complex
    else
      pot ? 
	zinv(N,zr,zi,mr,mi,NE,er,ei,qr,qi,smooth,xopt,cutoff,cont) :
	zlog(N,zr,zi,mr,mi,NE,er,ei,qr,qi,smooth,xopt,cutoff,cont);
  }
}
/*------------------------------------------------------------------------*/
/* Four routines for the logarithmic potential. */
void rlog2(int N,
	   const double *zr,const double *zi,
	   const double *mr,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Logarithmic potential for N potentials located at (zr,zi) with real
   mass mr. The contribution is added to (pr,pi). */
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double P = 0.5*log(re*re+im*im);
	pr[i] -= mr[j]*P;
	pr[j] -= mr[i]*P;
      }
    break;

  case RANKINE:
    cutoff = xopt*xopt;
    shape = 0.5*log(cutoff)-0.5;
    for (int i = 0; i < N; i++) {
      pr[i] -= mr[i]*shape;
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*(ad/cutoff)+shape;
	pr[i] -= mr[j]*P;
	pr[j] -= mr[i]*P;
      }
    }
    break;

  case SCULLY:
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = log(cutoff)/log(shape+cutoff);
    for (int i = 0; i < N; i++) {
      pr[i] -= mr[i]*0.5*log(shape)*scale;
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*log(shape+ad)*scale;
	pr[i] -= mr[j]*P;
	pr[j] -= mr[i]*P;
      }
    }
    break;

  case OSEEN:
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = log(cutoff)/(log(cutoff)+expint(shape*cutoff));
    for (int i = 0; i < N; i++) {
      pr[i] += mr[i]*0.5*(log(shape)+gamma_)*scale;
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else if (ad != 0.0)
	  P = 0.5*(log(ad)+expint(shape*ad))*scale;
	else
	  P = -0.5*(log(shape)+gamma_)*scale;
	pr[i] -= mr[j]*P;
	pr[j] -= mr[i]*P;
      }
    }
    break;
  }
}
/*------------------------------------------------------------------------*/
void zlog2(int N,
	   const double *zr,const double *zi,
	   const double *mr,const double *mi,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Same as above but with complex mass (mr,mi). */
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double P = 0.5*log(re*re+im*im);
	pr[i] -= mr[j]*P;
	pi[i] -= mi[j]*P;
	pr[j] -= mr[i]*P;
	pi[j] -= mi[i]*P;
      }
    break;

  case RANKINE:
    cutoff = xopt*xopt;
    shape = 0.5*log(cutoff)-0.5;
    for (int i = 0; i < N; i++) {
      pr[i] -= mr[i]*shape;
      pi[i] -= mi[i]*shape;
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*(ad/cutoff)+shape;
	pr[i] -= mr[j]*P;
	pi[i] -= mi[j]*P;
	pr[j] -= mr[i]*P;
	pi[j] -= mi[i]*P;
      }
    }
    break;

  case SCULLY:
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = log(cutoff)/log(shape+cutoff);
    for (int i = 0; i < N; i++) {
      pr[i] -= mr[i]*0.5*log(shape)*scale;
      pi[i] -= mi[i]*0.5*log(shape)*scale;
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*log(shape+ad)*scale;
	pr[i] -= mr[j]*P;
	pi[i] -= mi[j]*P;
	pr[j] -= mr[i]*P;
	pi[j] -= mi[i]*P;
      }
    }
    break;

  case OSEEN:
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = log(cutoff)/(log(cutoff)+expint(shape*cutoff));
    for (int i = 0; i < N; i++) {
      pr[i] += mr[i]*0.5*(log(shape)+gamma_)*scale;
      pi[i] += mi[i]*0.5*(log(shape)+gamma_)*scale;
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else if (ad != 0.0)
	  P = 0.5*(log(ad)+expint(shape*ad))*scale;
	else
	  P = -0.5*(log(shape)+gamma_)*scale;
	pr[i] -= mr[j]*P;
	pi[i] -= mi[j]*P;
	pr[j] -= mr[i]*P;
	pi[j] -= mi[i]*P;
      }
    }
    break;
  }
}
/*------------------------------------------------------------------------*/
void rlog(int N,
	  const double *zr,const double *zi,
	  const double *mr,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Logarithmic potential evaluated at NE points (er,ei) with
   potentials at the N points (zr,zi) with real mass mr. The
   contribution is added to (qr,qi). */
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++)
	if (er[i] != zr[j] || ei[i] != zi[j]) {
	  const double re = er[i]-zr[j],im = ei[i]-zi[j];
	  const double P = 0.5*log(re*re+im*im);
	  qr[i] -= mr[j]*P;
	}
    break;

  case RANKINE:
    cutoff = xopt*xopt;
    shape = 0.5*log(cutoff)-0.5;
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*(ad/cutoff)+shape;
	qr[i] -= mr[j]*P;
      }
    break;

  case SCULLY:
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = log(cutoff)/log(shape+cutoff);
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*log(shape+ad)*scale;
	qr[i] -= mr[j]*P;
      }
    break;

  case OSEEN:
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = log(cutoff)/(log(cutoff)+expint(shape*cutoff));
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else if (ad != 0.0)
	  P = 0.5*(log(ad)+expint(shape*ad))*scale;
	else
	  P = -0.5*(log(shape)+gamma_)*scale;
	qr[i] -= mr[j]*P;
      }
    break;
  }
}
/*------------------------------------------------------------------------*/
void zlog(int N,
	  const double *zr,const double *zi,
	  const double *mr,const double *mi,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Same as above but with complex mass (mr,mi). */
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++)
	if (er[i] != zr[j] || ei[i] != zi[j]) {
	  const double re = er[i]-zr[j],im = ei[i]-zi[j];
	  const double P = 0.5*log(re*re+im*im);
	  qr[i] -= mr[j]*P;
	  qi[i] -= mi[j]*P;
	}
    break;

  case RANKINE:
    cutoff = xopt*xopt;
    shape = 0.5*log(cutoff)-0.5;
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*(ad/cutoff)+shape;
	qr[i] -= mr[j]*P;
	qi[i] -= mi[j]*P;
      }
    break;

  case SCULLY:
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = log(cutoff)/log(shape+cutoff);
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*log(shape+ad)*scale;
	qr[i] -= mr[j]*P;
	qi[i] -= mi[j]*P;
      }
    break;

  case OSEEN:
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = log(cutoff)/(log(cutoff)+expint(shape*cutoff));
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else if (ad != 0.0)
	  P = 0.5*(log(ad)+expint(shape*ad))*scale;
	else
	  P = -0.5*(log(shape)+gamma_)*scale;
	qr[i] -= mr[j]*P;
	qi[i] -= mi[j]*P;
      }
    break;
  }
}
/*------------------------------------------------------------------------*/
/* Now follows the 4 corresponding functions for the Harmonic
   potential. */
void rinv2(int N,
	   const double *zr,const double *zi,
	   const double *mr,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont)
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double P = 1.0/(re*re+im*im);
	pr[i] -= mr[j]*P*re;
	pi[i] += mr[j]*P*im;
	pr[j] += mr[i]*P*re;
	pi[j] -= mr[i]*P*im;
      }
    break;

  case RANKINE:
    cutoff = xopt*xopt;
    shape = 1.0/cutoff;
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = shape;
	pr[i] -= mr[j]*P*re;
	pi[i] += mr[j]*P*im;
	pr[j] += mr[i]*P*re;
	pi[j] -= mr[i]*P*im;
      }
    break;

  case SCULLY:
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = 1.0+shape/cutoff;
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = 1.0/(shape+ad)*scale;
	pr[i] -= mr[j]*P*re;
	pi[i] += mr[j]*P*im;
	pr[j] += mr[i]*P*re;
	pi[j] -= mr[i]*P*im;
      }
    break;

  case OSEEN:
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = -1.0/expm1(-shape*cutoff);
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad != 0.0) {
	  if (ad >= cutoff)
	    P = 1.0/ad;
	  else
	    P = -expm1(-shape*ad)/ad*scale;
	  pr[i] -= mr[j]*P*re;
	  pi[i] += mr[j]*P*im;
	  pr[j] += mr[i]*P*re;
	  pi[j] -= mr[i]*P*im;
	}
      }
    break;
  }
}
/*------------------------------------------------------------------------*/
void zinv2(int N,
	   const double *zr,const double *zi,
	   const double *mr,const double *mi,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont)
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double P = 1.0/(re*re+im*im);
	pr[i] -= (mr[j]*re+mi[j]*im)*P;
	pi[i] += (mr[j]*im-mi[j]*re)*P;
	pr[j] += (mr[i]*re+mi[i]*im)*P;
	pi[j] -= (mr[i]*im-mi[i]*re)*P;
      }
    break;

  case RANKINE:
    cutoff = xopt*xopt;
    shape = 1.0/cutoff;
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = shape;
	pr[i] -= (mr[j]*re+mi[j]*im)*P;
	pi[i] += (mr[j]*im-mi[j]*re)*P;
	pr[j] += (mr[i]*re+mi[i]*im)*P;
	pi[j] -= (mr[i]*im-mi[i]*re)*P;
      }
    break;

  case SCULLY:
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = 1.0+shape/cutoff;
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = 1.0/(shape+ad)*scale;
	pr[i] -= (mr[j]*re+mi[j]*im)*P;
	pi[i] += (mr[j]*im-mi[j]*re)*P;
	pr[j] += (mr[i]*re+mi[i]*im)*P;
	pi[j] -= (mr[i]*im-mi[i]*re)*P;
      }
    break;

  case OSEEN:
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = -1.0/expm1(-shape*cutoff);
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad != 0.0) {
	  if (ad >= cutoff)
	    P = 1.0/ad;
	  else
	    P = -expm1(-shape*ad)/ad*scale;
	  pr[i] -= (mr[j]*re+mi[j]*im)*P;
	  pi[i] += (mr[j]*im-mi[j]*re)*P;
	  pr[j] += (mr[i]*re+mi[i]*im)*P;
	  pi[j] -= (mr[i]*im-mi[i]*re)*P;
	}
      }
    break;
  }
}
/*------------------------------------------------------------------------*/
void rinv(int N,
	  const double *zr,const double *zi,
	  const double *mr,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont)
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++)
	if (er[i] != zr[j] || ei[i] != zi[j]) {
	  const double re = er[i]-zr[j],im = ei[i]-zi[j];
	  const double P = 1.0/(re*re+im*im);
	  qr[i] -= mr[j]*P*re;
	  qi[i] += mr[j]*P*im;
	}
    break;

  case RANKINE:
    cutoff = xopt*xopt;
    shape = 1.0/cutoff;
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = shape;
	qr[i] -= mr[j]*P*re;
	qi[i] += mr[j]*P*im;
      }
    break;

  case SCULLY:
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = 1.0+shape/cutoff;
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = 1.0/(shape+ad)*scale;
	qr[i] -= mr[j]*P*re;
	qi[i] += mr[j]*P*im;
      }
    break;

  case OSEEN:
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = -1.0/expm1(-shape*cutoff);
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad != 0.0) {
	  if (ad >= cutoff)
	    P = 1.0/ad;
	  else
	    P = -expm1(-shape*ad)/ad*scale;
	  qr[i] -= mr[j]*P*re;
	  qi[i] += mr[j]*P*im;
	}
      }
    break;
  }
}
/*------------------------------------------------------------------------*/
void zinv(int N,
	  const double *zr,const double *zi,
	  const double *mr,const double *mi,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont)
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++)
	if (er[i] != zr[j] || ei[i] != zi[j]) {
	  const double re = er[i]-zr[j],im = ei[i]-zi[j];
	  const double P = 1.0/(re*re+im*im);
	  qr[i] -= (mr[j]*re+mi[j]*im)*P;
	  qi[i] += (mr[j]*im-mi[j]*re)*P;
	}
    break;

  case RANKINE:
    cutoff = xopt*xopt;
    shape = 1.0/cutoff;
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = shape;
	qr[i] -= (mr[j]*re+mi[j]*im)*P;
	qi[i] += (mr[j]*im-mi[j]*re)*P;
      }
    break;

  case SCULLY:
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = 1.0+shape/cutoff;
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = 1.0/(shape+ad)*scale;
	qr[i] -= (mr[j]*re+mi[j]*im)*P;
	qi[i] += (mr[j]*im-mi[j]*re)*P;
      }
    break;

  case OSEEN:
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = -1.0/expm1(-shape*cutoff);
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad != 0.0) {
	  if (ad >= cutoff)
	    P = 1.0/ad;
	  else
	    P = -expm1(-shape*ad)/ad*scale;
	  qr[i] -= (mr[j]*re+mi[j]*im)*P;
	  qi[i] += (mr[j]*im-mi[j]*re)*P;
	}
      }
    break;
  }
}
/*------------------------------------------------------------------------*/


