/* direct_cuda.cpp 
 * Provides interface function between cpu-code and gpu-code
 */

/* M. Holm, from direct.c */

#include <cstdlib> 

#include "expint.h"
#include "direct.h"
#include "cudaeval.h" // includes "cudaevalkernels.h"


/*------------------------------------------------------------------------*/

void MPexp_evals_direct(const MPexp *This, int N, int NE,
                double *pr,double *pi,
                double *qr,double *qi,
                const double *er,const double *ei,
                const double *zr,const double *zi,
                const double *mr,const double *mi,
                int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Evaluation of the potential. Function is critical to performance. */
{
  cuda_evals_direct(This, N, NE, pr, pi, qr, qi, er, ei, zr, zi, mr, mi, pot,
    smooth, xopt, cutoff, cont);
}

/*------------------------------------------------------------------------*/

void MPexp_smooth_(int pot,SMOOTHER smooth,double xopt,bool cont,
                   double *cutoff,double *shape,double *scale)
/* Precomputes the three parameters (cutoff,shape,scale) given
   (pot,smooth,xopt,cutoff,cont). Used in cudaeval.cu. */
{
  *scale = 1.0;

  if (pot == 0)
    switch (smooth) {
    case DIRAC:
      break;

    case RANKINE:
      *cutoff = xopt*xopt;
      *shape = 0.5*log(*cutoff)-0.5;
      break;

    case SCULLY:
      *cutoff = *cutoff**cutoff;
      *shape = xopt*xopt;
      if (cont)
        *scale = log(*cutoff)/log(*shape+*cutoff);
      break;

    case OSEEN:
    *cutoff = *cutoff**cutoff;
    *shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      *scale = log(*cutoff)/(log(*cutoff)+expint(*shape**cutoff));
      break;
    }
  else
    switch (smooth) {
    case DIRAC:
      break;

    case RANKINE:
      *cutoff = xopt*xopt;
      *shape = 1.0/(*cutoff);
      break;

    case SCULLY:
      *cutoff = *cutoff**cutoff;
      *shape = xopt*xopt;
      if (cont)
        *scale = 1.0+*shape/(*cutoff);
      break;

    case OSEEN:
      *cutoff = *cutoff**cutoff;
      *shape = 1.2564312086261696770/(xopt*xopt);
      if (cont)
        *scale = -1.0/expm1(-*shape**cutoff);
      break;
    }
}

#define BEGIN(begin,i) (begin == -1 ? i+1 : begin)
	  


/*------------------------------------------------------------------------*/
void directInteract(int N,
                    const double *zr,const double *zi,
		    const double *mr,const double *mi,
		    int NE,
		    const double *er,const double *ei,
                    double *pr,double *pi,
		    double *qr,double *qi,int pot,
		    SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Direct evaluation of the potential from N pointmasses (mr,mi) (with
   mi possibly NULL) at the positions (zr,zi). The evaluation is done
   at the points (zr,zi) if (pr,pi) is not NULL and/or at the points
   (er,ei) if (qr,qi) is not NULL. The result is stored in the vectors
   (pr,pi) and (qr,qi), respectively, which must be allocated and
   cleared prior to call. Logarithmic potential for pot == 0, harmonic
   potential otherwise. The smoothing is done using smoother smooth
   with parameters (xopt,cutoff,cont). */
{

  allpair_direct_cuda(N,zr,zi,mr,mi,NE,er,ei,pr,pi,qr,qi,pot,smooth,
        xopt,cutoff,cont);

}
/*------------------------------------------------------------------------*/
