
/* fmm.c */

/* M. Holm 2011-02-02 */
/* S. Engblom 2010-02-18 */

#include "fmm.h"
#include "fmmshift.h"
#include "expint.h"
#include "direct.h"
#include "tl.h"

#include <pthread.h>
#include <time.h>
#include <sys/time.h>
int tasks_submitted_counter;
int tasks_complete_counter;
pthread_mutex_t task_countup_mutex;
struct timeval tasks_start, tasks_complete;
struct timeval gpu_start, gpu_complete;

/* #defines: (preferably handled via make)
   ---------------------------------------

   FMM_SHRINK    Postsplit shrinking of boxes on/off: this option is
                 kept here for the purpose of visual
                 demonstrations. Also good to have this option if
                 time-dependent FMM is implemented.

   It is recommended to #define all these.

*/


#ifndef NUMTHREADS
#define NUMTHREADS 1
#endif


#if NUMTHREADS == 1
#undef TASKLOG
#endif

/* externed from fmmshift. Needed to allocate thread workspaces. */
// ** consider moving thread workspace allocation to fmmshift?
extern int pshift;

double theta;

// high-level routines
void MPexp_setup(MPexp *this,int N,int NE,
                 const double *zr,const double *zi,
                 const double *mr,const double *mi,
                 const double *er,const double *ei,
                 double *restrict pr,double *restrict pi,
                 double *restrict qr,double *restrict qi,
                 SMOOTHER smooth,double xopt,double cutoff,bool cont,
                 int pot,double tol,int Ndirect,double eta,bool meshonly);
void MPexp_free(MPexp *this);
void MPexp_eval2(const MPexp *this,
                 double *restrict pr,double *restrict pi,
                 const double *zr,const double *zi,
                 const double *mr,const double *mi,
                 int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont);
void MPexp_eval(const MPexp *this,
                double *restrict qr,double *restrict qi,
                const double *er,const double *ei,
                const double *zr,const double *zi,
                const double *mr,const double *mi,
                int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont);

// helper functions
void MPexp_minmax_(double *xmin,double *xmax,double *ymin,double *ymax,
                   int N,const double *x,const double *y);
void MPexp_box_(MPexp *this,int N,int NE,
                const double *zr,const double *zi,
                const double *er,const double *ei);
void MPexp_partition_(int begin,int *im,int end,double *z0,
                      double *min, double *max,
                      int *ix,const double *z,int Nmax);
void MPexp_partition_potentials_(MPexp *This,
                          int N,const double *zr,const double *zi,
                          int NE,const double *er,const double *ei,                          
                          int num_seq_levels,
                          int Ndirect,double cutoff,
                          int *ix,int *ixptr,
                          int *jx,int *jxptr);                      
void MPexp_connect_(MPexp *This,mpSparse *C,
                        int *levm2p,int *maxm2p,double cutoff);            
void MPexp_split(int begin,int *im,int end,double z0,
                  double *min,double *max,
                  int *ix,const double *z);
void MPexp_sort_(MPexp *this,
                 int N,const double *zr,const double *zi,
                 int NE,const double *er,const double *ei,
                 int *maxm2p,int *levm2p,double cutoff,int Ndirect,double eta, int num_seq_levels);

// low-level routines
void mpexp_init(mpexp *this,int p,
                const double zr,const double zi,
                const double mr,const double mi,
                int pot);
void mpexp_initp(mpexp *this,int p,
                 const double zr,const double zi,
                 const double mr,const double mi,
                 int pot);
bool mpexp_theta(const mpexp *this,const mpexp *that, double cutoff);
int mpexp_theta2(const mpexp *this,const mpexp *that, double cutoff);
void mpexp_eval(const mpexp *this,int p,
                double *vr,double *vi,
                const double *zr,const double *zi,
                const int *ix,int begin,int end);
void mpexp_evalmp(const mpexp *this,int p,int pot,
                  double *vr,double *vi,
                  const double *zr,const double *zi,
                  const int *ix,int begin,int end);

// Task parallelism routines
void task_down(tl_thread_info_t tinfo, void *data, void **ptr); 

void addTask_new_down(task_down_args *args);

void task_partition(tl_thread_info_t tinfo, void *data, void **ptr); 

void addTask_new_partition(task_partition_args *args);

void task_init(tl_thread_info_t tinfo, void *data, void **ptr); 

void addTask_new_init(task_init_args *args);

//
typedef struct {
  int index, Nf, pot, jstart, jend;
  const MPexp *This;
  double *pr, *pi, *qr, *qi;
  const double *zr, *er;
  const double *zi, *ei;
  const double *mr;
  const double *mi;
  SMOOTHER smooth;
  double xopt;
  double cutoff, shape, scale;
  bool cont;
} task_fareval_args;

void task_fareval(tl_thread_info_t tinfo, void *data, void **ptr); 
void addTask_fareval(task_fareval_args *args);
void task_fareval2(tl_thread_info_t tinfo, void *data, void **ptr); 	  
void addTask_fareval2(task_fareval_args *args);



/*------------------------------------------------------------------------*/
void fmm2dInteract(int N,
                   const double *zr,const double *zi,
                   const double *restrict mr,const double *restrict mi,
                   int NE,
                   const double *er,const double *ei,
                   double *restrict pr,double *restrict pi,
                   double *qr,double *qi,
                   int pot,double tol,int Ndirect,double theta_,double eta,
                   SMOOTHER smooth,double xopt,double cutoff,bool cont, double* timesarray)
/* Evaluation by the fast multipole method to accuracy TOL of the
   potential from N pointmasses (mr,mi) (with mi possibly NULL) at the
   positions (zr,zi). The evaluation is done at the points (zr,zi) if
   (pr,pi) is not NULL and/or at the points (er,ei) if (qr,qi) is not
   NULL (it is required that NE is zeroe if (qr,qi) are NULL). The
   result is added to the vectors (pr,pi) and (qr,qi), respectively,
   which must be allocated prior to call. Logarithmic potential for
   pot == 0, harmonic potential otherwise. The smoothing is done using
   smoother smooth with parameters (xopt,cutoff,cont). */
{    
  theta = theta_;
  // early return: cannot allocate anything
  if (N <= 0) return;

  if(NUMTHREADS > 1)
  {
    tl_init(NUMTHREADS);
  }

  // multipole structure
  MPexp MP;
  // allocate and setup representation
  // direct (local) evaluation is done here as well
  MPexp_setup(&MP,N,NE,zr,zi,mr,mi,er,ei,pr,pi,qr,qi,\
  smooth,xopt,cutoff,cont,pot,tol,Ndirect,eta,false);
  
  StartTime;
#ifdef TASKLOG
  unsigned long long start, stop;
  start = tl_log_time();
#endif
  // evaluation of far-field at potentials (zr,zi)
  if (pr != NULL) MPexp_eval2(&MP,pr,pi,zr,zi,mr,mi,
                              pot,smooth,xopt,cutoff,cont);
#ifdef TASKLOG
  stop = tl_log_time();
  tl_log("MPexp_eval2", start, stop);
#endif

#ifdef TASKLOG
  start = tl_log_time();
#endif
  // evaluation of far-field at coordinates (er,ei)
  if (qr != NULL) MPexp_eval(&MP,qr,qi,er,ei,zr,zi,mr,mi,
                             pot,smooth,xopt,cutoff,cont);
#ifdef TASKLOG
  stop = tl_log_time();
  tl_log("MPexp_eval", start, stop);
#endif

  StopTime;
  times.L2P = GetTime;

  // deallocate
  MPexp_free(&MP);

  if(NUMTHREADS > 1)
  {
     tl_destroy();
  }

  if(timesarray != NULL)
  {
     double * t = (double *) &times;
     for(int i = 0; i < sizeof(times)/sizeof(double); i++)
     {
        timesarray[i] = t[i];
     }
  }
}
/*------------------------------------------------------------------------*/
void fmm2dGetMesh(int nlhs,mxArray *plhs[],
                  int N,
                  const double *zr,const double *zi,
                  const double *mr,const double *mi,
                  int NE,
                  const double *er,const double *ei,
                  int pot,double tol,int Ndirect,double eta,double theta_)
/* Returns the multipole mesh in a cell vector of complex
   mxArrays. Additionally, all sparse connection matrices are
   optionally returned. For debugging and pedagogical purposes. */
{
  if (N <= 0) return;

  if(NUMTHREADS > 1)
    tl_init(NUMTHREADS);

  theta = theta_;

  // setup multipole tree
  MPexp MP;

  MPexp_setup(&MP,N,NE,zr,zi,mr,mi,er,ei,NULL,NULL,NULL,NULL,(SMOOTHER)0,0,0,\
  0,pot,tol,Ndirect,eta,true);

  if(NUMTHREADS > 1)
     tl_destroy();

  // create cell vector and copy the FMM mesh
  plhs[0] = mxCreateCellMatrix(1,MP.nlevel+1);
  for (int i = 0; i <= MP.nlevel; i++) {
    mxArray *cell = mxCreateDoubleMatrix(2,MP.lptr[i+1]-MP.lptr[i],
                                         mxCOMPLEX);
    double *mshr = mxGetPr(cell), *mshi = mxGetPi(cell);
    mpexp *offset = &MP.root[MP.lptr[i]];

    for (int j = 0; j < MP.lptr[i+1]-MP.lptr[i]; j++) {
      mshr[2*j] = creal(offset[j].z0);
      mshi[2*j] = cimag(offset[j].z0);
      mshr[2*j+1] = creal(offset[j].d0);
      mshi[2*j+1] = cimag(offset[j].d0);
    }
    mxSetCell(plhs[0],i,cell);
  }

  // additionally, the connection matrices
  if (nlhs > 1) {
    plhs[1] = mxCreateCellMatrix(1,MP.nlevel+1);
    for (int i = 0; i <= MP.nlevel; i++) {
      const mpSparse *S = &MP.connect[i];
      const int N = MP.lptr[i+1]-MP.lptr[i];
      mxArray *cell = mxCreateSparse(N,N,S->jcptr[N],mxREAL);
      mwIndex *jc = mxGetJc(cell),*ir = mxGetIr(cell);
      double *pr = mxGetPr(cell);
      // mwIndex not defined? use '-DmwIndex=int' in the mex-command!

          // note: interaction through parents as in the C99-code is not
      // implemented here (very little efficiency gain but more
      // complicated memory access pattern)
      jc[0] = 0;
      for (int j = 0,k; j < N; j++) {
        jc[j+1] = S->jcptr[j+1];

        // near connections
        if (i < MP.nlevel)
          for (k = jc[j]; k < S->kcptr[j]; k++) {
            ir[k] = S->ir[k];
            pr[k] = 1.0;
          }
        else {
          // final level contains more information:
          for (k = jc[j]; k < S->kcptr[j+N]; k++) {
            ir[k] = S->ir[k];
            pr[k] = 1.0;
          }
          for ( ; k < S->kcptr[j]; k++) {
            ir[k] = abs(S->ir[k]); // clear sign
            pr[k] = 2.0; // near-far connection
          }
        }
        // far connections
        for ( ; k < S->jcptr[j+1]; k++) {                       
          ir[k] = S->ir[k];
          pr[k] = -1.0;
        }
      }
      mxSetCell(plhs[1],i,cell);
    }
  }

  // deallocate and return
  MPexp_free(&MP);
}
/*------------------------------------------------------------------------*/
static void task_countup() /* catches "task complete" signal */
{
  pthread_mutex_lock(&task_countup_mutex);
  tasks_complete_counter++;
  if(tasks_complete_counter == tasks_submitted_counter)
  {
    gettimeofday(&tasks_complete,NULL);
  }
  pthread_mutex_unlock(&task_countup_mutex);
}


/*------------------------------------------------------------------------*/
void MPexp_setup(MPexp *this, int N, int NE,
                 const double *zr,const double *zi,
                 const double *mr,const double *mi,
                 const double *er,const double *ei,
                 double *restrict pr,double *restrict pi,
                 double *restrict qr,double *restrict qi,
                 SMOOTHER smooth,double xopt,double cutoff,bool cont,
                 int pot,double tol,int Ndirect,double eta,bool meshonly)
/* Setups and allocates a multipole tree for N potentials at (zr,zi)
   to be evaluated at the NE points (er,ei) to tolerance tol. The
   discretization is determined so that the number of potentials in
   each box at the finest level is <= Ndirect. If the boolean variable
   meshonly is true, then only the multipole tree and the connectivity
   information is determined. */
{
#ifdef TASKLOG
  unsigned long long log_start, log_stop;
#endif
  
  StartTime;
#ifdef TASKLOG
  log_start = tl_log_time();
#endif

  // size of the domain
  MPexp_box_(this,N,NE,zr,zi,er,ei);
  
  // number of multipole terms
  this->pcoeff = (int)floor(log2(4.0*tol)/log2(theta)-1.0)+(pot != 0);
  /* the magical constants 4 is a (not overly optimistic) estimate
     obtained from numerous experiments with the model error ~
     constant * theta^(p+1) and theta = 0.5 */
  if (this->pcoeff < 1) this->pcoeff = 1;

  // number of levels
  this->nlevel = ceil(0.5*log2(5.0/8.0*(double)N/Ndirect));
  if (this->nlevel < 0) this->nlevel = 0;
  /* the magical constant 5/8 comes from solving for the number of
     levels the equation E(N/Nf) = Ndirect, where E is expectation, N
     is considered a uniform random number on [1,M] (with M tending to
     infinity) and Nf is defined below */

  /* used frequently below: total number of coefficients, number of
     boxes at the finest level and total number of boxes */
  const int P = this->pcoeff+1;
  const int Nf = 1 << (this->nlevel << 1);
  const int Nt = (Nf-1)/3+Nf;

  // pointer to levels
  this->lptr = MALLOC((this->nlevel+2)*sizeof(int));
  this->lptr[0] = 0;
  for (int level = 0,nb = 1; level <= this->nlevel; level++, nb <<= 2)
    this->lptr[level+1] = this->lptr[level]+nb;

  // multipole tree
  this->root = MALLOC(Nt*sizeof(mpexp));

  // multipole coefficients
  dcmplx *coeffs1 = CALLOC(P*Nt,sizeof(dcmplx));
  dcmplx *coeffs2 = CALLOC(P*Nt,sizeof(dcmplx));
  
  for (int j = 0; j < Nt; j++) {
    this->root[j].coeff1 = coeffs1; coeffs1 += P;
    this->root[j].coeff2 = coeffs2; coeffs2 += P;
  }

  // permutation of input
  this->ixptr = MALLOC((Nf+1)*sizeof(int));
  this->ix = MALLOC(N*sizeof(int));
  if (NE) {
    this->jxptr = MALLOC((Nf+1)*sizeof(int));
    this->jx = MALLOC(NE*sizeof(int));
  }
  else
    this->jxptr = this->jx = NULL;

  // connections between boxes at each level
  this->connect = MALLOC((this->nlevel+1)*sizeof(mpSparse));
#ifdef TASKLOG
  log_stop = tl_log_time();
  tl_log("Alloc", log_start, log_stop);
#endif

  StopTime;  // Toss this time, it is so very small
//  PrintTime("Alloc: %f\n");


#ifdef TASKLOG
  log_start = tl_log_time();
#endif
  /* setup the multipole mesh, sort all potentials and evaluation
     points and determine all connections */
  
  int num_sequential;
  int num_sequential_sort;
  if(NUMTHREADS > 1)
  {
#ifndef NUMTASKS
#define NUMTASKS 3
#endif
    num_sequential = NUMTASKS;
    num_sequential_sort = (NUMTHREADS>4)?4:3;//this->nlevel+1;//
  }else
  {
    num_sequential = 1000;
    num_sequential_sort = this->nlevel+1;
  }
   

  int num_seq_levels = (num_sequential<this->nlevel)?num_sequential:this->nlevel;  // this->nlevel;//  
  num_sequential_sort = (num_sequential_sort<this->nlevel+1)?num_sequential_sort:this->nlevel+1;
  int maxm2p,levm2p;
  MPexp_sort_(this,N,zr,zi,NE,er,ei,&maxm2p,&levm2p,cutoff,Ndirect,eta,num_sequential_sort);
#ifdef TASKLOG
  log_stop = tl_log_time();
  tl_log("MPexp_sort", log_start, log_stop);
#endif

  if (meshonly) {
      return;
  }

  // ** Could start direct evaluation here already. 

  /* Start of the actual multipole algorithm */

  /* Step #1: Initialization at the finest level. */
  StartTime;
#ifdef TASKLOG
  log_start = tl_log_time();
#endif
  {
    mpexp *offset = &this->root[this->lptr[this->nlevel]];
    mpSparse *C = &this->connect[this->nlevel];

#if NUMTHREADS > 1000
    int num_init_tasks = NUMTHREADS<<2; // 4x extra for load-balancing reasons
    int tasksize = (Nf+num_init_tasks-1)/num_init_tasks; // to round up
    task_init_args args[num_init_tasks];
    for (int i = 0; i < num_init_tasks; i++)
    {
       args[i].index = i;
       args[i].jbegin = i*tasksize;
       args[i].jend = ((i+1)*tasksize<Nf)?(i+1)*tasksize:Nf; // run to at most j=Nf: last task is smallest
       args[i].pot = pot;
       args[i].This = this;
       args[i].C = C;
       args[i].offset = offset;
       args[i].zr = zr;
       args[i].zi = zi;
       args[i].mr = mr;
       args[i].mi = mi;
       args[i].Nf = Nf;
       addTask_new_init(&args[i]); 
    }
    tl_barrier();
#else   
    if (mi == NULL)
      for (int j = 0; j < Nf; j++) {
        // initialize multipole expansion in box j
        for (int i = this->ixptr[j]; i < this->ixptr[j+1]; i++)
          mpexp_init(&offset[j],this->pcoeff,
                     zr[this->ix[i]],zi[this->ix[i]],
                     mr[this->ix[i]],0.0,pot);

        // 'less near' connections: shift into polynomial directly      
        for (int jj = C->kcptr[j+Nf]; jj < C->kcptr[j]; jj++) {
          const int k = C->ir[jj];
          if(k > 0)  
            for (int i = this->ixptr[k]; i < this->ixptr[k+1]; i++)
              mpexp_initp(&offset[j],this->pcoeff,
                          zr[this->ix[i]],zi[this->ix[i]],
                          mr[this->ix[i]],0.0,pot);
        }
      }
    else
      for (int j = 0; j < Nf; j++) {
                for (int i = this->ixptr[j]; i < this->ixptr[j+1]; i++)
                        mpexp_init(&offset[j],this->pcoeff,
                         zr[this->ix[i]],zi[this->ix[i]],
                         mr[this->ix[i]],mi[this->ix[i]],pot);

                for (int jj = C->kcptr[j+Nf]; jj < C->kcptr[j]; jj++) {
                        const int k = C->ir[jj];
                        if (k > 0) 
                        for (int i = this->ixptr[k]; i < this->ixptr[k+1]; i++)
                                mpexp_initp(&offset[j],this->pcoeff,
                                                zr[this->ix[i]],zi[this->ix[i]],
                                                mr[this->ix[i]],mi[this->ix[i]],pot);
        }
      }
#endif
  }

#ifdef TASKLOG
  log_stop = tl_log_time();
  tl_log("P2M", log_start, log_stop);
#endif

  StopTime;
  times.P2M = GetTime;

  /* Step #2: Upward pass of the fast multipole algorithm. After the
     finest level has been initialized, this loop performs the
     'upward' pass where all multipole contributions are shifted and
     added onto all coarser levels. */
  StartTime;
#ifdef TASKLOG
  log_start = tl_log_time();
#endif
  shift_alloc(this->pcoeff,pot,maxm2p); // prepare to shift
  for (int l = this->nlevel,nb = Nf; l > 0; l--, nb >>= 2) {
    mpexp *parent = &this->root[this->lptr[l-1]];
    mpexp *child = &this->root[this->lptr[l]];

    // loop over all children at this level
    for (int j = 0; j < nb; j += 4) {
      // shift and add result from 4 children at once
      shift_m2m(child[j].z0-parent->z0,child[j+1].z0-parent->z0,
                child[j+2].z0-parent->z0,child[j+3].z0-parent->z0,
                parent->coeff1,child[j].coeff1);
      parent++;
    }
  }
#ifdef TASKLOG
  log_stop = tl_log_time();
  tl_log("M2M", log_start, log_stop);
#endif

  StopTime;
  times.M2M = GetTime;

  /* Step #3: Downward pass of the fast multipole method. This shifts
     expansions from the interaction list of each box to the center of
     the box and adds up the result. Before continuing to finer
     levels, each parent shifts and adds its expansion to each of its
     children. */
  StartTime;
#ifdef TASKLOG
  log_start = tl_log_time();
#endif
  int l, nb;
  //for (l = levm2p-1, nb = 1 << 2*levm2p; l < num_seq_levels; l++, nb <<= 2)
  for (l = levm2p-1, nb = 1 << 2*levm2p; l < this->nlevel; l++, nb <<= 2)
  {
    mpexp *parent = &this->root[this->lptr[l]];
    mpexp *child = &this->root[this->lptr[l+1]];
    mpSparse *C = &this->connect[l+1];
  
    /* tricky part: loop over all boxes (parents) at this level and
     * symmetrically shift results to and from the interaction list of
     * each child */
    for (int j = 0; j < nb; ) {
      // contribution from parent to 4 children at once
      shift_p2p(parent->z0-child[j].z0, parent->z0-child[j+1].z0,
              parent->z0-child[j+2].z0, parent->z0-child[j+3].z0,
              child[j].coeff2, parent->coeff2, NULL);
      parent++;
      
      // loop through the interaction list of each child
      for (int c = 0; c < 4; c++, j++){
        shift_m2ps(child[j].z0, &child[0].z0, sizeof(mpexp),
                C->ir, C->kcptr[j], C->jcptr[j+1],
                child[j].coeff2, child[0].coeff1, NULL, NULL);
      }              
    }
  }
#ifdef TASKLOG
  log_stop = tl_log_time();
  tl_log("M,L2L serial", log_start, log_stop);
#endif
  StopTime;
  if(NUMTHREADS == 1)
     times.ML2L = GetTime;    

  gettimeofday(&tasks_start,NULL);

  pthread_mutex_init(&task_countup_mutex, NULL);
  tasks_complete_counter = 0; // initialize task counters
  tasks_submitted_counter = 0;

  if(l < this->nlevel) {            
    //  then create tasks for each subtree                           
    task_down_args args[nb/4];
         
    for (int j = 0; j < nb/4; j ++) {
      args[j].nlevel = this->nlevel;
      args[j].startlevel = l;
      args[j].connect = this->connect;
      args[j].root = this->root;
      args[j].lptr = this->lptr;        
      args[j].j = j;            
      addTask_new_down(&args[j]);
    }            
  }
  
// If only using CPU, don't want to overlap direct with FMM
#ifdef CPUONLY
  if(NUMTHREADS > 1)
  {
     tl_barrier();
     gettimeofday(&tasks_complete, NULL);
  }
#endif
#ifdef TASKLOG
  log_start = tl_log_time();
#endif
  StartTime;     
  gettimeofday(&gpu_start,NULL);                
  /* direct evaluation done while tasks are executing */
  MPexp_evals_direct(this,N, NE, pr,pi,qr,qi,er,ei,zr,zi,mr,mi,
            pot,smooth,xopt,cutoff,cont);

#ifdef TASKLOG
  log_stop = tl_log_time();
  tl_log("P2P", log_start, log_stop);
#endif

  gettimeofday(&gpu_complete,NULL);                
  
  StopTime;
  times.P2P = GetTime;

  double gpu_time = ((gpu_complete.tv_sec-gpu_start.tv_sec)+ \
		(gpu_complete.tv_usec-gpu_start.tv_usec)/1000000.0) ;

  StartTime;       
#ifdef TASKLOG
  log_start = tl_log_time();
#endif
   
  if(NUMTHREADS > 1)
  {
     tl_barrier();
  }
     
#ifdef TASKLOG
  log_stop = tl_log_time();
  tl_log("tl_barrier", log_start, log_stop);
#endif

  StopTime;
  shift_free();

  double task_time = (tasks_complete.tv_sec-tasks_start.tv_sec) + \
    (tasks_complete.tv_usec-tasks_start.tv_usec)/1000000.0;

  // if nlevels was so low that no tasks were created..
  if(tasks_submitted_counter == 0)
     task_time = 0;

  double load_imbalance = gpu_time - task_time;

#ifdef CPUONLY // load imbalance is meaningless
  load_imbalance = 0;
#endif

  if(NUMTHREADS > 1)
    times.ML2L = task_time;
  times.P2P = gpu_time;
  times.loadbalance = load_imbalance;
}
/*------------------------------------------------------------------------*/

void task_init(tl_thread_info_t tinfo, void *data, void **ptr)
{
  task_init_args *args = (task_init_args *) data; 

  int Nf = args->Nf;
  int pot = args->pot;
  MPexp *this = args->This;
  const mpSparse *C = args->C;
  mpexp *offset = args->offset;
  const double *restrict zr = args->zr; 
  const double *restrict zi = args->zi;
  const double *restrict mr = args->mr;
  const double *restrict mi = args->mi;
     
    if (mi == NULL)
      for (int j = args->jbegin; j < args->jend; j++) {
        // initialize multipole expansion in box j
        for (int i = this->ixptr[j]; i < this->ixptr[j+1]; i++)
          mpexp_init(&offset[j],this->pcoeff,
                     zr[this->ix[i]],zi[this->ix[i]],
                     mr[this->ix[i]],0.0,pot);

        // 'less near' connections: shift into polynomial directly      
        for (int jj = C->kcptr[j+Nf]; jj < C->kcptr[j]; jj++) {
          const int k = C->ir[jj];
          if(k > 0)  
            for (int i = this->ixptr[k]; i < this->ixptr[k+1]; i++)
              mpexp_initp(&offset[j],this->pcoeff,
                          zr[this->ix[i]],zi[this->ix[i]],
                          mr[this->ix[i]],0.0,pot);
        }
      }
    else
      for (int j = args->jbegin; j < args->jend; j++) {
                for (int i = this->ixptr[j]; i < this->ixptr[j+1]; i++)
                        mpexp_init(&offset[j],this->pcoeff,
                         zr[this->ix[i]],zi[this->ix[i]],
                         mr[this->ix[i]],mi[this->ix[i]],pot);

                for (int jj = C->kcptr[j+Nf]; jj < C->kcptr[j]; jj++) {
                        const int k = C->ir[jj];
                        if (k > 0) 
                        for (int i = this->ixptr[k]; i < this->ixptr[k+1]; i++)
                                mpexp_initp(&offset[j],this->pcoeff,
                                                zr[this->ix[i]],zi[this->ix[i]],
                                                mr[this->ix[i]],mi[this->ix[i]],pot);
        }
      }
}

/*------------------------------------------------------------------------*/
void addTask_new_init(task_init_args *args)
{
  /* no dependencies */
  char taskname[40];
  sprintf(taskname, "P2M task %d", args->index);
  tl_add_task(task_init, args, sizeof(*args), taskname);
}

/*------------------------------------------------------------------------*/
void addTask_new_down(task_down_args *args){
  /* Creates a task for downward pass */
  
 tasks_submitted_counter++;
 
 /* no dependencies */
  char taskname[40];
  sprintf(taskname, "M,L2L %d", args->j);
  tl_add_task(task_down, args, sizeof(*args), taskname);
}

/*------------------------------------------------------------------------*/
// Should do an entire downward pass for the task subtree 
void task_down(tl_thread_info_t tinfo, void *data, void **ptr) {
  
  task_down_args *args = (task_down_args *) data;  
  
  int taskindex = args->j; // index of "task parent"
  
  mpexp *parent;
  mpexp *child;
  const mpSparse *connect = args->connect;
  
  const mpexp *root = args->root;
  int l, subl, nb, tasknb;
  
  // workspace allocation for HORNERSHIFT:
  dcmplx *rpow = tl_get_workspace(tinfo, (pshift+1)*sizeof(dcmplx));
  rpow[0] = 1.0; 
  dcmplx *wksp = tl_get_workspace(tinfo, 2*(pshift+1)*sizeof(dcmplx));
  
  // for each level
  for (l = args->startlevel, tasknb = 4; l < args->nlevel; \
          l++, tasknb <<=2){
//    printf("task %d: level %d, subl=%d, tasknb = %d, nb=%d\n", taskindex, l, subl,tasknb, nb);
    // base pointers to parent and child for this level
    parent = (mpexp *)&root[args->lptr[l]];
    child = (mpexp *)&root[args->lptr[l+1]];
    
    const mpSparse *C = &connect[l+1];
    
    // shift to task section (4 elements per task per level)
    int offset = taskindex*tasknb;
   
    parent += offset>>2;
          
    // for each box on level that is a descendant of task parent
    for(int j=offset; j<offset+tasknb; ) {      
//      printf("%d: p2p with j = %d\n",taskindex,j);
      shift_p2p(parent->z0-child[j].z0, parent->z0-child[j+1].z0,
              parent->z0-child[j+2].z0, parent->z0-child[j+3].z0,
              child[j].coeff2, parent->coeff2, wksp);
      
      parent++;
      
      for (int c = 0; c < 4; c++, j++) {        
//        printf("%d: m2ps with j = %d\n",taskindex,j);
        shift_m2ps(child[j].z0, &child[0].z0, sizeof(mpexp),
                C->ir, C->kcptr[j], C->jcptr[j+1],
                child[j].coeff2, child[0].coeff1,
                rpow, wksp);
        
      }
    }
  }

  //pthread_kill(pthread_self(),SIGUSR1);  // raise signal USR1: I'm done with a task.
  task_countup();
}


/*------------------------------------------------------------------------*/
void MPexp_free(MPexp *this)
/* Deallocates the whole tree. */
{
  for (int l = this->nlevel; l >= 0; l--)
    FREE(this->connect[l].jcptr);
  FREE(this->connect);
  FREE(this->jx);
  FREE(this->jxptr);
  FREE(this->ix);
  FREE(this->ixptr);
  FREE(this->root->coeff2);  
  FREE(this->root->coeff1); 
  FREE(this->root);
  FREE(this->lptr);
}
/*------------------------------------------------------------------------*/
void MPexp_minmax_(double *xmin,double *xmax,double *ymin,double *ymax,
                   int N,const double *x,const double *y)
/* Finds min/max of two vectors. Note: it is assumed that suitable
   initial values in xmin/xmax are already set. */
{
  // loop unrolled to get rid of a few N/2 comparisions
  for (int i = 0; i < N-1; i += 2) {
    double min,max;

    if (x[i] < x[i+1]) {
      min = x[i];
      max = x[i+1];
    }
    else {
      min = x[i+1];
      max = x[i];
    }
    if (min < *xmin) *xmin = min;
    if (max > *xmax) *xmax = max;

    if (y[i] < y[i+1]) {
      min = y[i];
      max = y[i+1];
    }
    else {
      min = y[i+1];
      max = y[i];
    }
    if (min < *ymin) *ymin = min;
    if (max > *ymax) *ymax = max;
  }
  if (N&1) {
    if (x[N-1] < *xmin)
      *xmin = x[N-1];
    else if (x[N-1] > *xmax)
      *xmax = x[N-1];

    if (y[N-1] < *ymin)
      *ymin = y[N-1];
    else if (y[N-1] > *ymax)
      *ymax = y[N-1];
  }
}
/*------------------------------------------------------------------------*/
void MPexp_box_(MPexp *this,int N,int NE,
                const double *zr,const double *zi,
                const double *er,const double *ei)
/* Determines and sets the dimensions of the computational domain. */
{
  this->xmin = zr[0];
  this->xmax = zr[0];
  this->ymin = zi[0];
  this->ymax = zi[0];
  MPexp_minmax_(&this->xmin,&this->xmax,
                &this->ymin,&this->ymax,
                N-1,zr+1,zi+1);

  // bounding box might become larger:
  if (NE) 
    MPexp_minmax_(&this->xmin,&this->xmax,
                  &this->ymin,&this->ymax,
                  NE,er,ei);
}
/*------------------------------------------------------------------------*/
#define SWAPI(X,Y) { int tmpi = X; X = Y; Y = tmpi; }

void MPexp_partition_(int begin,int *im,int end,
                      double *min, double *max,double *z0,
                      int *ix,const double *z,int Nmax)
/* Partitions the set of points z[ix[begin]] through z[ix[end]]
  (exclusive) in two sets indicated by [begin,im) and [im,end) such
  that all points are <= or >= than z0 (which is both input and
  output). The division is done such that Nlarge <= Nmax, where Nlarge
  is the larger number of particles in the two lists. It is assumed
  that this split is possible and hence that Nmax is not too small.
 
  The routine is a workhorse and performance is important; for each
  box which is not a leaf it will be called three times. */
{
  int il,ir,left = begin,right = end;
  int Nsmall,Nlarge;
  // first split according to input z0 (geometric midpoint)
  const double z0_ = *z0; // compiler: this is really a constant!
  il = left-1; ir = right;
  do il++; while (il < right && z[ix[il]] < z0_);
  do ir--; while (left <= ir && z0_ < z[ix[ir]]);
  if (il < ir) {
    SWAPI(ix[il],ix[ir])
    /* after the first swap, there is no need to check for index out
       of bounds */
    for ( ; ; ) {
      do il++; while (z[ix[il]] < z0_);
      do ir--; while (z0_ < z[ix[ir]]);
      if (il >= ir) break;
      SWAPI(ix[il],ix[ir])
    }
  }

  // proceed with largest interval [left,il) or [il,right)
  Nsmall = il-begin;
  Nlarge = end-il;
  if (Nsmall > Nlarge) {
    Nlarge = Nsmall;
    *im = right = il;
  }
  else
    *im = left = il;
  const int im0 = *im; // saved for later

  while (Nlarge > Nmax) {
    if (right-left >= 3) {
      /* "median of 3": find pivot and create boundaries such that
	 left <= left+1 <= right-1 */
      const int mid = (left+right)>>1;
      SWAPI(ix[left+1],ix[mid])
      if (z[ix[left]] > z[ix[right-1]])
	SWAPI(ix[left],ix[right-1])
      if (z[ix[left+1]] > z[ix[right-1]])
        SWAPI(ix[left+1],ix[right-1])
      else if (z[ix[left]] > z[ix[left+1]])
	SWAPI(ix[left],ix[left+1])
      const double z1_ = *z0 = z[ix[left+1]];

      for (il = left+1,ir = right-1; ; ) {
	do il++; while (z[ix[il]] < z1_);
	do ir--; while (z1_ < z[ix[ir]]);
	if (il >= ir) break;

	SWAPI(ix[il],ix[ir])
      }
      SWAPI(ix[left+1],ix[ir]) // ir is the correct place for the pivot

      // the pivot belongs to the shortest interval:
      Nsmall = ir-begin; // [begin,ir)
      Nlarge = end-ir-1; // [ir+1,end)
      if (Nsmall > Nlarge) {
	Nlarge = Nsmall;
	*im = right = ir;
      }
      else
	/* note: if Nsmall == Nlarge (can only happen if right-left is
	   odd), then the right array will be shorter than the left by
	   one */
	*im = left = ir+1;
    }
    else {
      // code above does not work properly for less than three elements
      if (z[ix[left]] > z[ix[right-1]])
	SWAPI(ix[left],ix[right-1])
      *im = right-1;
      *z0 = z[ix[right-1]];
      break;
      /*** prefer a direct sort when right-left <= 7 or so? */ 
    }
  }

  // determine min and max
  if (min != NULL) {
    *min = *z0;
    for (int i = begin; i < *im; i++)
      if (z[ix[i]] < *min) *min = z[ix[i]];
  }
  if (max != NULL) {
    *max = *z0;
    for (int i = *im; i < end; i++)
      if (*max < z[ix[i]]) *max = z[ix[i]];
  }
}
/*------------------------------------------------------------------------*/
void MPexp_split_(int begin,int *im,int end,double z0,
                  double *min, double *max,
                  int *ix,const double *z)
/* Split as above, but the coordinate z0 around which to split is
   already known. It is also assumed that min/max, whenever nonzero,
   are set to suitable initial guesses of the true min/max (typically
   the ones you get from MPexp_partition_() above). */
{
  int il = begin-1,ir = end;
  do il++; while (il < end && z[ix[il]] < z0);
  do ir--; while (begin <= ir && z0 < z[ix[ir]]);
  if (il < ir) {
    SWAPI(ix[il],ix[ir])
    for ( ; ; ) {
      do il++; while (z[ix[il]] < z0);
      do ir--; while (z0 < z[ix[ir]]);
      if (il >= ir) break;
      SWAPI(ix[il],ix[ir])
    }
  }
  *im = il;

  if (min != NULL) {
    for (int i = begin; i < *im; i++)
      if (z[ix[i]] < *min) *min = z[ix[i]];
  }
  if (max != NULL) {
    for (int i = *im; i < end; i++)
      if (*max < z[ix[i]]) *max = z[ix[i]];
  }
}


/*------------------------------------------------------------------------*/
void addTask_new_partition(task_partition_args *args){
  /* Creates a task for partitioning */
 
  int l = args->startlevel;
  const mpexp *root = args->root;
  const mpexp *parent = &root[args->lptr[l]];
  const mpexp *child = &root[args->lptr[l+1]];
  const mpSparse *connect = args->connect;
  const mpSparse *C = &connect[l+1];
  const int *ix = C->ir;
  int j = args->j; // index of "task parent"
 
 /* no dependencies */
  char taskname[40];
  sprintf(taskname, "partition %d", j);
  tl_add_task(task_partition, args, sizeof(*args), taskname);
}

/*------------------------------------------------------------------------*/
// Do a partition for the task subtree 
void task_partition(tl_thread_info_t tinfo, void *data, void **ptr) {
  
  task_partition_args *args = (task_partition_args *) data;  
  
  int taskindex = args->j; // index of "task parent"
  
  mpexp *parent;
  mpexp *child;
  const mpSparse *connect = args->connect;
  
  const mpexp *root = args->root;
  int l, tasknb;
  
  int *ix = args->ix;
  int *jx = args->jx;      
  int *ixptr = args->ixptr;
  int *jxptr = args->jxptr;      
  
  int Ndirect = args->Ndirect;
  int Nmax = args->Nmax;
  int Nmax0 = args->Nmax0;
  
  int NE = args->NE;
  const double * zr = args->zr;
  const double * er = args->er;
  const double * zi = args->zi;  
  const double * ei = args->ei; 
  
  // do a partition 
  
  int cstride, pstride;
  
  // ** TODO: optimize this ugly statement
for (l = args->startlevel,Nmax = args->Nmax, cstride = args->cstride, pstride = args->pstride, tasknb = 1; \
       l <= args->nlevel; l++, cstride >>= 2, pstride >>= 2, tasknb <<= 2) {  
    mpexp *parent = (mpexp *)&root[args->lptr[l-1]];
    mpexp *child = (mpexp*)&root[args->lptr[l]];
    
    /* divide each box on level l into 4 children such that about
       1/4th of the potentials are located in each child box. */
        
    /* determine the maximum number of coordinates after each split
       (the map is x --> ceil(0.5*x)) */
    if ((Nmax0 = (Nmax >> 1)+(Nmax&1)) <= Ndirect)
      Nmax0 = Nmax = Ndirect;
    else if ((Nmax = (Nmax0 >> 1)+(Nmax0&1)) <= Ndirect)
      Nmax = Ndirect;
    /* the intent with using Ndirect as a cutoff is to save some work
       at the final level whenever the number of particles is already
       sufficiently low */

    int pindex; // index of parent box    

    // loop over all parent boxes that descend from this thread's parent index
    for (int i = taskindex*tasknb, j = (taskindex*tasknb)<<2; i < (taskindex+1)*tasknb; i++, j += 4) {  
    
      dcmplx z,d;
      dcmplx z0,d0,z1,d1;
      dcmplx z00,d00,z01,d01,z10,d10,z11,d11;
      double split, min, max;

      pindex = i*pstride; // index of parent box 
    
      // box to be split
      z = parent[i].z0;
      d = parent[i].d0;
      /* partition the coordinates [ixptr[i],ixptr[i+1]) in box i
         into 4 sections indexed by [ixptr[4*i+n],ixptr[4*i+n+1])
         for n = 0..3 */

      /* cut across the longest side if it is a certain factor longer,
         otherwise take a y-cut followed by two x-cuts; this is the
         Lebesgue ordering */
      if (creal(d) > meshratio*cimag(d)) {
        split = creal(z);

        MPexp_partition_(ixptr[pindex],&ixptr[pindex+2*cstride],ixptr[pindex+1*pstride],
                         &min, &max,&split,ix,zr,Nmax0 );
        
/*        COMPLEXASSIGN(z0,1.0,1.0);
        COMPLEXASSIGN(z0,0.5*(creal(z)-creal(d)+split),cimag(z));
        d0 = z0-(z-d);
        COMPLEXASSIGN(z1,0.5*(creal(z)+creal(d)+split),cimag(z));
        d1 = (z+d)-z1;*/
        if (NE)
          MPexp_split_(jxptr[pindex],&jxptr[pindex+2*cstride],jxptr[pindex+1*pstride],split,&min,&max,jx,er );

#ifdef FMM_SHRINK
	z0 = 0.5*(min+split)+I*cimag(z);
	d0 = z0-(min+I*cimag(z-d));
	z1 = 0.5*(split+max)+I*cimag(z);
	d1 = (max+I*cimag(z+d))-z1;
#else
	z0 = 0.5*(creal(z-d)+split)+I*cimag(z);
	d0 = z0-(z-d);
	z1 = 0.5*(creal(z+d)+split)+I*cimag(z);
	d1 = (z+d)-z1;
#endif



      }
      else {
        split = cimag(z);

        MPexp_partition_(ixptr[pindex],&ixptr[pindex+2*cstride],ixptr[pindex+1*pstride],
                         &min, &max,
                         &split,ix,zi,Nmax0 );

/*        COMPLEXASSIGN(z0,creal(z),0.5*(cimag(z)-cimag(d)+split));
        d0 = z0-(z-d);
        COMPLEXASSIGN(z1,creal(z),0.5*(cimag(z)+cimag(d)+split));
        d1 = (z+d)-z1;*/
        if (NE)
          MPexp_split_(jxptr[pindex],&jxptr[pindex+2*cstride],jxptr[pindex+1*pstride],split,&min,&max,jx,ei );
#ifdef FMM_SHRINK
	z0 = creal(z)+I*0.5*(min+split);
	d0 = z0-(creal(z-d)+I*min);
	z1 = creal(z)+I*0.5*(split+max);
	d1 = (creal(z+d)+I*max)-z1;
#else
	z0 = creal(z)+I*0.5*(cimag(z-d)+split);
	d0 = z0-(z-d);
	z1 = creal(z)+I*0.5*(cimag(z+d)+split);
	d1 = (z+d)-z1;
#endif
      }     
      // same procedure for the resulting two boxes...
      if (meshratio*creal(d0) > cimag(d0)) {
        split = creal(z0);

        MPexp_partition_(ixptr[pindex],&ixptr[pindex+1*cstride],ixptr[pindex+2*cstride],
                         &min, &max,
                         &split,ix,zr,Nmax );

        /*COMPLEXASSIGN(z00,0.5*(creal(z0)-creal(d0)+split),cimag(z0));
        d00 = z00-(z0-d0);
        COMPLEXASSIGN(z01,0.5*(creal(z0)+creal(d0)+split),cimag(z0));
        d01 = (z0+d0)-z01;*/
        if (NE)
          MPexp_split_(jxptr[pindex],&jxptr[pindex+1*cstride],jxptr[pindex+2*cstride],split,&min,&max,jx,er );
#ifdef FMM_SHRINK
	z00 = 0.5*(min+split)+I*cimag(z0);
	d00 = z00-(min+I*cimag(z0-d0));
	z01 = 0.5*(split+max)+I*cimag(z0);
	d01 = (max+I*cimag(z0+d0))-z01;
#else
	z00 = 0.5*(creal(z0-d0)+split)+I*cimag(z0);
	d00 = z00-(z0-d0);
	z01 = 0.5*(creal(z0+d0)+split)+I*cimag(z0);
	d01 = (z0+d0)-z01;
#endif

      }
      else {
        split = cimag(z0);

        MPexp_partition_(ixptr[pindex],&ixptr[pindex+1*cstride],ixptr[pindex+2*cstride],
                         &min, &max,
                         &split,ix,zi,Nmax );

       /* COMPLEXASSIGN(z00,creal(z0),0.5*(cimag(z0)-cimag(d0)+split));
        d00 = z00-(z0-d0);
        COMPLEXASSIGN(z01,creal(z0),0.5*(cimag(z0)+cimag(d0)+split));
        d01 = (z0+d0)-z01;*/
        if (NE)
          MPexp_split_(jxptr[pindex],&jxptr[pindex+1*cstride],jxptr[pindex+2*cstride],split,&min,&max,jx, ei );
#ifdef FMM_SHRINK
	z00 = creal(z0)+I*0.5*(min+split);
	d00 = z00-(creal(z0-d0)+I*min);
	z01 = creal(z0)+I*0.5*(split+max);
	d01 = (creal(z0+d0)+I*max)-z01;
#else
	z00 = creal(z0)+I*0.5*(cimag(z0-d0)+split);
	d00 = z00-(z0-d0);
	z01 = creal(z0)+I*0.5*(cimag(z0+d0)+split);
	d01 = (z0+d0)-z01;
#endif

      }
      // and again...
      if (meshratio*creal(d1) > cimag(d1)) {
        split = creal(z1);

        MPexp_partition_(ixptr[pindex+2*cstride],&ixptr[pindex+3*cstride],ixptr[pindex+1*pstride],
                         &min, &max,
                         &split,ix,zr,Nmax );

       /* COMPLEXASSIGN(z10,0.5*(creal(z1)-creal(d1)+split),cimag(z1));
        d10 = z10-(z1-d1);
        COMPLEXASSIGN(z11,0.5*(creal(z1)+creal(d1)+split),cimag(z1));
        d11 = (z1+d1)-z11;*/
        if (NE)
          MPexp_split_(jxptr[pindex+2*cstride],&jxptr[pindex+3*cstride],jxptr[pindex+1*pstride],split,&min,&max,jx,er );
#ifdef FMM_SHRINK
	z10 = 0.5*(min+split)+I*cimag(z1);
	d10 = z10-(min+I*cimag(z1-d1));
	z11 = 0.5*(split+max)+I*cimag(z1);
	d11 = (max+I*cimag(z1+d1))-z11;
#else
	z10 = 0.5*(creal(z1-d1)+split)+I*cimag(z1);
	d10 = z10-(z1-d1);
	z11 = 0.5*(creal(z1+d1)+split)+I*cimag(z1);
	d11 = (z1+d1)-z11;
#endif
      }
      else {
        split = cimag(z1);

        MPexp_partition_(ixptr[pindex+2*cstride],&ixptr[pindex+3*cstride],ixptr[pindex+1*pstride],
                         &min, &max,
                         &split,ix,zi,Nmax );
/*
        COMPLEXASSIGN(z10,creal(z1),0.5*(cimag(z1)-cimag(d1)+split));
        d10 = z10-(z1-d1);
        COMPLEXASSIGN(z11,creal(z1),0.5*(cimag(z1)+cimag(d1)+split));
        d11 = (z1+d1)-z11;*/
        if (NE)
          MPexp_split_(jxptr[pindex+2*cstride],&jxptr[pindex+3*cstride],jxptr[pindex+1*pstride],split,&min,&max,jx,ei );
#ifdef FMM_SHRINK
	z10 = creal(z1)+I*0.5*(min+split);
	d10 = z10-(creal(z1-d1)+I*min);
	z11 = creal(z1)+I*0.5*(split+max);
	d11 = (creal(z1+d1)+I*max)-z11;	
#else
	z10 = creal(z1)+I*0.5*(cimag(z1-d1)+split);
	d10 = z10-(z1-d1);
	z11 = creal(z1)+I*0.5*(cimag(z1+d1)+split);
	d11 = (z1+d1)-z11;
#endif
      }

      // produced boxes
      child[j].z0 = z00;
      child[j].d0 = d00;
      child[j+1].z0 = z01;
      child[j+1].d0 = d01;
      child[j+2].z0 = z10;
      child[j+2].d0 = d10;
      child[j+3].z0 = z11;
      child[j+3].d0 = d11;      
    }
  }
  

}
/*------------------------------------------------------------------------*/
void MPexp_sort_(MPexp *This,
                 int N,const double *zr,const double *zi,
                 int NE,const double *er,const double *ei,
                 int *maxm2p,int *levm2p,double cutoff,int Ndirect,double eta,int num_seq_levels)
/* Sorts the N potentials (zr,zi) in the multipole tree this which is
   simultaneously constructed and, for NE > 0, also assigns each
   evaluation point (er,ei) a multipole box. After each level has been
   thus constructed the connectivity information is also
   determined. The integer maxm2p is set to the maximum number of
   cluster to cluster interactions in the whole tree (the maximum
   number of -1's in any column of all the connectivity matrices) and
   levm2p is the first level where a cluster to cluster interaction
   occurs. This function is critical to performance. */
{
 const int Nf = This->lptr[This->nlevel+1]-This->lptr[This->nlevel];
  int *ix,*ixptr;
  int *jx,*jxptr;
#ifdef TASKLOG
  unsigned long long log_start, log_stop;
#endif
  mpSparse *C = This->connect;

  // (ixptr,ix) to be built
  ix = This->ix;
  ixptr = This->ixptr;
//  for (int ii = 1; ii <= Nf; ii++) ixptr[ii] = -666;  // ** unnecessary for-loop for clarity
  ixptr[0] = 0;
  ixptr[Nf] = N; 
  
    
  // ditto (jxptr,jx)
  if (NE) {
    jx = This->jx;
    jxptr = This->jxptr;
    jxptr[0] = 0;
    jxptr[Nf] = NE;
  }
  else{
      jx = jxptr = NULL;
  }
  
  /* connectivity: each matrix is strictly lower triangular since a
     box is always in near connection with itself and since
     connectivity is a symmetric relation */
  C[0].jcptr = (int *)mxMalloc(5*sizeof(int));
  C[0].jcptr[0] = 0;
  // convenient in fmm2dGetMesh():
  C[0].kcptr = &C[0].jcptr[2];
  C[0].ir = &C[0].kcptr[2];
  C[0].ir[0] = 0;
  
  
  C[0].jcptr[1] = 1; // direct interaction to self is included
  C[0].kcptr[0] = 1;
  C[0].kcptr[1] = 1; 

  *maxm2p = 0;
  *levm2p = This->nlevel+1;
  StartTime;
#ifdef TASKLOG
  log_start = tl_log_time();
#endif
  MPexp_partition_potentials_(This,N,zr,zi,NE,er,ei,num_seq_levels,
                       Ndirect,cutoff,ix,ixptr,
                       jx,jxptr);
#ifdef TASKLOG
  log_stop = tl_log_time();
  tl_log("MPexp_partition_potentials", log_start, log_stop);
#endif
  StopTime;                        
  times.partition = GetTime;
  
  
  StartTime;  
#ifdef TASKLOG
  log_start = tl_log_time();
#endif
  MPexp_connect_(This,C,levm2p,maxm2p,cutoff);                       
#ifdef TASKLOG
  log_stop = tl_log_time();
  tl_log("MPexp_connect", log_start, log_stop);
#endif
  StopTime;                        
  times.connectivity = GetTime;

}

/*-----------------------------------------------------------------------*/
void MPexp_partition_potentials_(MPexp *This,
                          int N,const double *zr,const double *zi,
                          int NE,const double *er,const double *ei,                          
                          int num_seq_levels,
                          int Ndirect,double cutoff,
                          int *ix,int *ixptr,
                          int *jx,int *jxptr)
/* Partitioning of potentials and evaluation points. This
   function creates the multipole mesh. */
{
  const int Nf = This->lptr[This->nlevel+1]-This->lptr[This->nlevel];
  const int Nt = (Nf-1)/3+Nf;

  // (ixptr,ix) and (jxptr,jx) to be built
  for (int i = 0; i < N; i++) ix[i] = i;
  if(NE) for (int i = 0; i < NE; i++) jx[i] = i;
 
  
  // root box
  This->root[0].z0 = 0.5*(This->xmin+This->xmax)+I*0.5*(This->ymin+This->ymax);
  This->root[0].d0 = 0.5*(This->xmax-This->xmin)+I*0.5*(This->ymax-This->ymin);

  // loop over levels
  int l, Nmax, Nmax0, nb, cstride, pstride;

#ifdef TASKLOG
  unsigned long long log_start, log_stop;
#endif

#ifdef TASKLOG
    log_start = tl_log_time();
#endif
  num_seq_levels = This->nlevel + 1; // uncomment to turn off parallel partitioning

  for (l = 1,Nmax = N, nb=1, cstride = Nf>>2, pstride = Nf; \
       l < num_seq_levels; l++, nb <<= 2, cstride >>= 2, pstride >>= 2) {  

    mpexp *parent = &This->root[This->lptr[l-1]];
    mpexp *child = &This->root[This->lptr[l]];

    /* divide each box on level l into 4 children such that about
       1/4th of the potentials are located in each child box. */
        
    /* determine the maximum number of coordinates after each split
       (the map is x --> ceil(0.5*x)) */
    if ((Nmax0 = (Nmax >> 1)+(Nmax&1)) <= Ndirect)
      Nmax0 = Nmax = Ndirect;
    else if ((Nmax = (Nmax0 >> 1)+(Nmax0&1)) <= Ndirect)
      Nmax = Ndirect;
      
    /* the intent with using Ndirect as a cutoff is to save some work
       at the final level whenever the number of particles is already
       sufficiently low */

    int pindex; // index of parent box

    // loop over all parent boxes
    for (int i = 0, j = 0; i < nb; i++, j += 4) {    
      dcmplx z,d;
      dcmplx z0,d0,z1,d1;
      dcmplx z00,d00,z01,d01,z10,d10,z11,d11;
      double split, min, max;

      pindex = i*pstride; // index of parent box 
    
      // box to be split
      z = parent[i].z0;
      d = parent[i].d0;
      /* partition the coordinates [ixptr[i],ixptr[i+1]) in box i
         into 4 sections indexed by [ixptr[4*i+n],ixptr[4*i+n+1])
         for n = 0..3 */

      /* cut across the longest side if it is a certain factor longer,
         otherwise take a y-cut followed by two x-cuts; this is the
         Lebesgue ordering */
      if (creal(d) > meshratio*cimag(d)) {
        split = creal(z);

        MPexp_partition_(ixptr[pindex],&ixptr[pindex+2*cstride],ixptr[pindex+1*pstride],
		       &min,&max,
                         &split,ix,zr,Nmax0 );
        
        COMPLEXASSIGN(z0,1.0,1.0);
        COMPLEXASSIGN(z0,0.5*(creal(z)-creal(d)+split),cimag(z));
        d0 = z0-(z-d);
        COMPLEXASSIGN(z1,0.5*(creal(z)+creal(d)+split),cimag(z));
        d1 = (z+d)-z1;
        if (NE)
          MPexp_split_(jxptr[pindex],&jxptr[pindex+2*cstride],jxptr[pindex+1*pstride],split,&min,&max,jx,er );

      }
      else {
        split = cimag(z);

        MPexp_partition_(ixptr[pindex],&ixptr[pindex+2*cstride],ixptr[pindex+1*pstride],
		       &min,&max,
                         &split,ix,zi,Nmax0 );

        COMPLEXASSIGN(z0,creal(z),0.5*(cimag(z)-cimag(d)+split));
        d0 = z0-(z-d);
        COMPLEXASSIGN(z1,creal(z),0.5*(cimag(z)+cimag(d)+split));
        d1 = (z+d)-z1;
        if (NE)
          MPexp_split_(jxptr[pindex],&jxptr[pindex+2*cstride],jxptr[pindex+1*pstride],split,&min,&max,jx,ei );

      }
      
      // same procedure for the resulting two boxes...   
      if (meshratio*creal(d0) > cimag(d0)) {
        split = creal(z0);

        MPexp_partition_(ixptr[pindex],&ixptr[pindex+1*cstride],ixptr[pindex+2*cstride],
		       &min,&max,
                         &split,ix,zr,Nmax );

        COMPLEXASSIGN(z00,0.5*(creal(z0)-creal(d0)+split),cimag(z0));
        d00 = z00-(z0-d0);
        COMPLEXASSIGN(z01,0.5*(creal(z0)+creal(d0)+split),cimag(z0));
        d01 = (z0+d0)-z01;
        if (NE)
          MPexp_split_(jxptr[pindex],&jxptr[pindex+1*cstride],jxptr[pindex+2*cstride],split,&min,&max,jx,er );

      }
      else {
        split = cimag(z0);

        MPexp_partition_(ixptr[pindex],&ixptr[pindex+1*cstride],ixptr[pindex+2*cstride],
		       &min,&max,
                         &split,ix,zi,Nmax );

        COMPLEXASSIGN(z00,creal(z0),0.5*(cimag(z0)-cimag(d0)+split));
        d00 = z00-(z0-d0);
        COMPLEXASSIGN(z01,creal(z0),0.5*(cimag(z0)+cimag(d0)+split));
        d01 = (z0+d0)-z01;
        if (NE)
          MPexp_split_(jxptr[pindex],&jxptr[pindex+1*cstride],jxptr[pindex+2*cstride],split,&min,&max,jx, ei );

      }

      // and again...
      if (meshratio*creal(d1) > cimag(d1)) {
        split = creal(z1);

        MPexp_partition_(ixptr[pindex+2*cstride],&ixptr[pindex+3*cstride],ixptr[pindex+1*pstride],
		       &min,&max,
                         &split,ix,zr,Nmax );

        COMPLEXASSIGN(z10,0.5*(creal(z1)-creal(d1)+split),cimag(z1));
        d10 = z10-(z1-d1);
        COMPLEXASSIGN(z11,0.5*(creal(z1)+creal(d1)+split),cimag(z1));
        d11 = (z1+d1)-z11;
        if (NE)
          MPexp_split_(jxptr[pindex+2*cstride],&jxptr[pindex+3*cstride],jxptr[pindex+1*pstride],split,&min,&max,jx,er );

      }
      else {
        split = cimag(z1);

        MPexp_partition_(ixptr[pindex+2*cstride],&ixptr[pindex+3*cstride],ixptr[pindex+1*pstride],
		       &min,&max,
                         &split,ix,zi,Nmax );

        COMPLEXASSIGN(z10,creal(z1),0.5*(cimag(z1)-cimag(d1)+split));
        d10 = z10-(z1-d1);
        COMPLEXASSIGN(z11,creal(z1),0.5*(cimag(z1)+cimag(d1)+split));
        d11 = (z1+d1)-z11;
        if (NE)
          MPexp_split_(jxptr[pindex+2*cstride],&jxptr[pindex+3*cstride],jxptr[pindex+1*pstride],split,&min,&max,jx,ei );

      }

      // produced boxes
      child[j].z0 = z00;
      child[j].d0 = d00;
      child[j+1].z0 = z01;
      child[j+1].d0 = d01;
      child[j+2].z0 = z10;
      child[j+2].d0 = d10;
      child[j+3].z0 = z11;
      child[j+3].d0 = d11;      
    }
  }
#ifdef TASKLOG
    log_stop = tl_log_time();
    tl_log("sequential partition", log_start, log_stop);
#endif

  if(NUMTHREADS > 1)
  {   
    if(l <= This->nlevel) {            
      //  then create tasks for each subtree                           
      task_partition_args args[nb];
         
      // ** TODO: remove redundant cruft from this struct.
      for (int j = 0; j < nb; j ++) {
        args[j].nlevel = This->nlevel;
        args[j].startlevel = l;
        args[j].connect = This->connect;
        args[j].root = This->root;
        args[j].lptr = This->lptr;        
        args[j].j = j;            
        args[j].ix = ix;
        args[j].jx = jx;      
        args[j].ixptr = ixptr;
        args[j].jxptr = jxptr;  
        args[j].N = N;
        args[j].NE = NE;
        args[j].cstride = cstride;
        args[j].pstride = pstride;      
        args[j].Nmax = Nmax;
        args[j].Nmax0 = Nmax0;  
        args[j].Ndirect = Ndirect; 
        args[j].zr = zr;
        args[j].zi = zi;
        args[j].er = er;
        args[j].ei = ei;      
        addTask_new_partition(&args[j]);
      }            

      // wait for tasks to complete...
      tl_barrier();
    }
  }
  
       
}

/*------------------------------------------------------------------------*/

void MPexp_connect_(MPexp *This,mpSparse *C,
                        int *levm2p,int *maxm2p,double cutoff)
/* Once the mesh has been obtained, this function determines the type
   of connection between all multipole boxes. Executed on the CPU. */
{
  // loop over all levels (nc = 1 since the root box is strongly
  // connected to itself)
  for (int l = 1,nc = 1; l <= This->nlevel; l++) {
    const int N0 = This->lptr[l]-This->lptr[l-1];
    const int N1 = This->lptr[l+1]-This->lptr[l];
    mpexp *child = &This->root[This->lptr[l]];

    /* allocate connectivity matrix using that all near connections at
       the previous level implies 4 connections for each of 4 children
       at this level */
    
    This->jsize = (2+(l == This->nlevel))*N1+1+(4*4*nc);
    C[l].jcptr = (int *)mxMalloc((This->jsize)*sizeof(int));
    C[l].jcptr[0] = 0;
    C[l].kcptr = &C[l].jcptr[N1+1];
    C[l].ir = &C[l].kcptr[(1+(l == This->nlevel))*N1];

    // setup connections
    nc = 0;  // counter for near connections
    for (int i = 0, j = 0; i < N0; i++)  // parent i
      for (int c = 0; c < 4; c++, j++) { // children j
        // forward- and backward counters in each column
        int fwd = C[l].jcptr[j];
        int bwd = fwd+4*(C[l-1].kcptr[i]-C[l-1].jcptr[i]);
        C[l].jcptr[j+1] = bwd;

        // parents (near) connections (includes the parent itself)
        for (int ii = C[l-1].jcptr[i]; ii < C[l-1].kcptr[i]; ii++) {
          /* the children of the near connections of box i become
             either near connections or far connections of box j */
          int k = C[l-1].ir[ii] << 2; // 0-child of connection
          // (note: minor differences to the C99-code in that
          // interactions through parents are not supported here)

          for (int d = 0; d < 4; d++, k++)
            // connection according to theta criterion
            if (!mpexp_theta(&child[j],&child[k],cutoff))        
              // near connection
              C[l].ir[fwd++] = k;
            else
              // far connection
              C[l].ir[--bwd] = k;
        }
        C[l].kcptr[j] = bwd;
        /* this splits the column into near connections
           [jcptr[j],kcptr[j]) and far connections
           [kcptr[j],jcptr[j+1]) */
        nc += bwd-C[l].jcptr[j];
        if (C[l].jcptr[j+1]-bwd > *maxm2p)
          *maxm2p = C[l].jcptr[j+1]-bwd;
        if (l < *levm2p && *maxm2p > 0)
          *levm2p = l;

        // last level: sort the near connections one step further
        if (l == This->nlevel) {
          int dir1,dir2; // 1: cluster to particle, -1: particle to cluster
          fwd = C[l].jcptr[j]-1;
          bwd = C[l].kcptr[j];
          // since the format uses the sign of the index to denote the
          // type of the near connection, box 0 must be skipped
          if (j == 0)
            bwd--; // skip box 0
          else {
            // again, skip box 0 (must be first since column is sorted)
            if (C[l].ir[fwd+1] == 0)
              fwd++;
            for ( ; ; ) {
              do
                fwd++;
              while (fwd < bwd &&
                     !(dir1 = mpexp_theta2(&child[j],&child[C[l].ir[fwd]],
                                           cutoff)));

              do {
                bwd--;
                if (fwd <= bwd &&
                    (dir2 = mpexp_theta2(&child[j],&child[C[l].ir[bwd]],
                                         cutoff)))
                  C[l].ir[bwd] *= dir2;
                else
                  break;
              } while (1);

              if (fwd >= bwd) break;
              SWAPI(C[l].ir[fwd],C[l].ir[bwd])
              C[l].ir[bwd] *= dir1;
            }
          }
          C[l].kcptr[j+N1] = bwd+1;
          /* this splits the near connections into 'truly near'
             [jcptr[j],kcptr[j+N]) and 'less near'
             [kcptr[j+N],kcptr[j]) connections */
        }
      }    
  }
}

/*------------------------------------------------------------------------*/

void task_fareval(tl_thread_info_t tinfo, void *data, void **ptr)
{
  
  // load data from struct
  task_fareval_args *args = (task_fareval_args *) data;
  const MPexp *This = args->This;
  double *pr = args->pr;
  double *pi = args->pi;
  double *qr = args->qr;
  double *qi = args->qi;
  const double *zr = args->zr;
  const double *zi = args->zi;
  const double *er = args->er;
  const double *ei = args->ei;
  const double *mr = args->mr;
  const double *mi = args->mi;
  int pot = args->pot;
  SMOOTHER smooth = args->smooth;
  double xopt = args->xopt;
  double cutoff = args->cutoff;
  bool cont = args->cont;

  const int Nf = args->Nf;
  const mpexp *restrict offset = &This->root[This->lptr[This->nlevel]];
  const int *restrict ix = This->ix,*restrict ixptr = This->ixptr;
  const int *restrict jx = This->jx,*restrict jxptr = This->jxptr;
  const mpSparse *C = &This->connect[This->nlevel];
  const int *restrict jcptr = C->jcptr,*restrict j2cptr = &C->kcptr[Nf],
    *restrict kcptr = C->kcptr,*restrict ir = C->ir;
    
  // get start and end indexes
  int jstart = args->jstart;
  int jend = args->jend;
 
  for (int j = jstart; j < jend; j++) {
    // far-field: contribution from distant potentials
    mpexp_eval(&offset[j],This->pcoeff,qr,qi,er,ei,jx,jxptr[j],jxptr[j+1]);

    // cluster to particle interactions
    for (int jj = j2cptr[j] ; jj < kcptr[j]; jj++) {
      const int k = ir[jj];
      if (k < 0)
             mpexp_evalmp(&offset[-k],This->pcoeff,pot,qr,qi,er,ei,
                     jx,jxptr[j],jxptr[j+1]);
          //if (k > 0)
            // multipole expansion in box j evaluated in all points in box k
          //  mpexp_evalmp(&offset[j],this->pcoeff,pot,qr,qi,er,ei,
           //              jx,jxptr[k],jxptr[k+1]); 
    }
  } /* end j-loop */
}
 
/*------------------------------------------------------------------------*/
void addTask_fareval(task_fareval_args *args)
{
/* no dependencies */
  char taskname[40];
  sprintf(taskname, "L2P task %d", args->index);
  tl_add_task(task_fareval, args, sizeof(*args), taskname);
}

/*------------------------------------------------------------------------*/
void task_fareval2(tl_thread_info_t tinfo, void *data, void **ptr)
{
  
  // load data from struct
  task_fareval_args *args = (task_fareval_args *) data;
  const MPexp *This = args->This;
  double *pr = args->pr;
  double *pi = args->pi;
  const double *zr = args->zr;
  const double *zi = args->zi;
  const double *mr = args->mr;
  const double *mi = args->mi;
  int pot = args->pot;
  SMOOTHER smooth = args->smooth;
  double xopt = args->xopt;
  double cutoff = args->cutoff;
  bool cont = args->cont;
    
  // get start and end indexes
  int jstart = args->jstart;
  int jend = args->jend;
  
  const int Nf = args->Nf;
  const mpexp *restrict offset = &This->root[This->lptr[This->nlevel]];
  const int *restrict ix = This->ix,*restrict ixptr = This->ixptr;
  const mpSparse *C = &This->connect[This->nlevel];
  const int *restrict jcptr = C->jcptr,*restrict j2cptr = &C->kcptr[Nf],
    *restrict kcptr = C->kcptr,*restrict ir = C->ir;
    
   
  for (int j = jstart; j < jend; j++) 
  {
   // far-field: contribution from distant potentials
    mpexp_eval(&offset[j],This->pcoeff,pr,pi,zr,zi,ix,ixptr[j],ixptr[j+1]);
    // cluster to particle interactions
    for (int jj = j2cptr[j] ; jj < kcptr[j]; jj++) {
      const int k = ir[jj];
      if (k < 0)
        mpexp_evalmp(&offset[-k],This->pcoeff,pot,pr,pi,zr,zi,
                     ix,ixptr[j],ixptr[j+1]);
//      if (k > 0)
        // multipole expansion in box j evaluated at all points in box k
//        mpexp_evalmp(&offset[j],This->pcoeff,pot,pr,pi,zr,zi,
 //                    ix,ixptr[k],ixptr[k+1]); 

    }                          
  } /* end j-loop */

}	  

/*------------------------------------------------------------------------*/
void addTask_fareval2(task_fareval_args *args)
{
/* no dependencies */
  char taskname[40];
  sprintf(taskname, "L2P task %d", args->index);
  tl_add_task(task_fareval2, args, sizeof(*args), taskname);
}

/*------------------------------------------------------------------------*/
void MPexp_eval2(const MPexp *this,
                 double *restrict pr,  double *restrict pi,
                 const double *zr,const double *zi,
                 const double *mr,const double *mi,
                 int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Evaluation of the potential at the potentials themselves. Function
   is critical to performance. */
{
  // hopefully, some optimizing compiler can use this:
  const int Nf = this->lptr[this->nlevel+1]-this->lptr[this->nlevel];
  const mpexp *restrict offset = &this->root[this->lptr[this->nlevel]];
  const int *restrict ix = this->ix,*restrict ixptr = this->ixptr;
  const mpSparse *C = &this->connect[this->nlevel];
  const int *restrict jcptr = C->jcptr,*restrict j2cptr = &C->kcptr[Nf],
    *restrict kcptr = C->kcptr,*restrict ir = C->ir;

  if(NUMTHREADS > 1 && NUMTHREADS <= Nf)
  {
     int boxespertask = (Nf/NUMTHREADS<1024)?Nf/NUMTHREADS:1024;
     int numtasks = Nf/boxespertask;
     task_fareval_args args[numtasks];
     // loop to create tasks
     for (int i = 0; i < numtasks; i++)
     {
        args[i].index = i;
        args[i].This = this; args[i].pr = pr; args[i].pi = pi;
        args[i].zr = zr; args[i].zi = zi; args[i].mr = mr; args[i].mi = mi;
        args[i].pot = pot; args[i].smooth = smooth; args[i].xopt = xopt;
        args[i].cutoff = cutoff; args[i].cont = cont;
        args[i].Nf = Nf; 
        //args[i].shape = shape; args[i].scale = scale;
        args[i].jstart = (Nf/(numtasks-1))*i;
        args[i].jend = ((Nf/(numtasks-1))*(i+1)>Nf)?Nf:(Nf/(numtasks-1))*(i+1); // the last task may be very small
        addTask_fareval2(&args[i]);
     }
     tl_barrier();
  }else{
      for (int j = 0; j < Nf; j++) {
        // far-field: contribution from distant potentials
        mpexp_eval(&offset[j],this->pcoeff,pr,pi,zr,zi,ix,ixptr[j],ixptr[j+1]);
        // cluster to particle interactions
        for (int jj = j2cptr[j] ; jj < kcptr[j]; jj++) {
          const int k = ir[jj];
      if (k < 0)
        mpexp_evalmp(&offset[-k],this->pcoeff,pot,pr,pi,zr,zi,
                     ix,ixptr[j],ixptr[j+1]);
//      if (k > 0)
        // multipole expansion in box j evaluated at all points in box k
//        mpexp_evalmp(&offset[j],This->pcoeff,pot,pr,pi,zr,zi,
 //                    ix,ixptr[k],ixptr[k+1]); 
    
        }                          
      } /* end j-loop */
  }
}
/*------------------------------------------------------------------------*/
void MPexp_eval(const MPexp *this,
                double *restrict qr, double *restrict qi,
                const double *er,const double *ei,
                const double *zr,const double *zi,
                const double *mr,const double *mi,
                int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Evaluation of the potential at the points (er,ei). Function is
   critical to performance. */
{
  const int Nf = this->lptr[this->nlevel+1]-this->lptr[this->nlevel];
  const mpexp *restrict offset = &this->root[this->lptr[this->nlevel]];
  const int *restrict ix = this->ix,*restrict ixptr = this->ixptr;
  const int *restrict jx = this->jx,*restrict jxptr = this->jxptr;
  const mpSparse *C = &this->connect[this->nlevel];
  const int *restrict jcptr = C->jcptr,*restrict j2cptr = &C->kcptr[Nf],
    *restrict kcptr = C->kcptr,*restrict ir = C->ir;
  
  if(NUMTHREADS > 1 && NUMTHREADS <= Nf)
  {
     int boxespertask = (Nf/NUMTHREADS<64)?Nf/NUMTHREADS:64;
     int numtasks = Nf/boxespertask;
     task_fareval_args args[numtasks];
     // loop to create tasks
     for (int i = 0; i < numtasks; i++)
     {
        args[i].index = i;
        args[i].This = this; 
        args[i].qr = qr; args[i].qi = qi;
        args[i].er = er; args[i].ei = ei;
        args[i].zr = zr; args[i].zi = zi; args[i].mr = mr; args[i].mi = mi;
        args[i].pot = pot; args[i].smooth = smooth; args[i].xopt = xopt;
        args[i].cutoff = cutoff; args[i].cont = cont;
        args[i].Nf = Nf; 
        //args[i].shape = shape; args[i].scale = scale;
        args[i].jstart = (Nf/(numtasks-1))*i;
        args[i].jend = ((Nf/(numtasks-1))*(i+1)>Nf)?Nf:(Nf/(numtasks-1))*(i+1); // the last task may be very small
        addTask_fareval(&args[i]);
     }
     tl_barrier();
  }else{
     for (int j = 0; j < Nf; j++) {
        // far-field: contribution from distant potentials
        mpexp_eval(&offset[j],this->pcoeff,qr,qi,er,ei,jx,jxptr[j],jxptr[j+1]);
    
        // cluster to particle interactions
        for (int jj = j2cptr[j] ; jj < kcptr[j]; jj++) {
          const int k = ir[jj];
          if (k < 0)
             mpexp_evalmp(&offset[-k],this->pcoeff,pot,qr,qi,er,ei,
                     jx,jxptr[j],jxptr[j+1]);
          //if (k > 0)
            // multipole expansion in box j evaluated in all points in box k
          //  mpexp_evalmp(&offset[j],this->pcoeff,pot,qr,qi,er,ei,
           //              jx,jxptr[k],jxptr[k+1]); 
        }
      } /* end j-loop */
  }
}
/*------------------------------------------------------------------------*/
void mpexp_init(mpexp *this,int p,
                const double zr,const double zi,
                const double mr,const double mi,int pot)
/* Initializes a multipole expansion at the lowest level with a single
   pointmass (mr,mi) at (zr,zi). */
{
  /* We have that the potential is given by (pot = 0)
     mp(z) = -m*log(z-zi)
           = -m*log(z-z0-(zi-z0))
           = -m*log((z-z0)*(1-(zi-z0)/(z-z0)))
           = -m*[log(z-z0)+log(1-(zi-z0)/(z-z0))]
           = -m*[log(z-z0)-sum_{j >= 1} (zi-z0)^j/j*(z-z0)^{-j},
     or othwerwise by (pot = 1)
     mp(z) = -m*1/(z-zi)
           = -m*1/(z-z0-(zi-z0))
           = -m/(z-z0)*1/(1-(zi-z0)/(z-z0))
           = -m*sum_{j >= 1} (zi-z0)^{j-1}*(z-z0)^{-j}.
  */

  const double re = zr-creal(this->z0),im = zi-cimag(this->z0);
  dcmplx *coeff = this->coeff1;

  if (pot == 0) {
    double zre = mr,zim = mi;
    coeff[0] -= (mr+I*mi);
    for (int j = 1; j <= p; j++) {
      double zre0 = zre*re-zim*im;
      zim = zim*re+zre*im;
      zre = zre0;
      coeff[j] += (zre/j+I*zim/j);
    }
  }
  else {
    double zre = mr,zim = mi;
    for (int j = 1; j <= p; j++) {
      coeff[j] -= (zre+I*zim);
      double zre0 = zre*re-zim*im;
      zim = zim*re+zre*im;
      zre = zre0;
    }
  }
}
/*------------------------------------------------------------------------*/
void mpexp_initp(mpexp *this,int p,
                 const double zr,const double zi,
                 const double mr,const double mi,int pot)
/* Initializes a multipole expansion at the lowest level with a single
   pointmass (mr,mi) at (zr,zi). The pointmass is assumed to be
   sufficiently far away from the center of the box that a polynomial
   expansion may be used directly. */
{
  /* We have that the potential is given by (pot = 0)
     mp(z) = -m*log(z-zi)
           = -m*log(z0-zi+(z-z0))
           = -m*[log(z0-zi)-sum_{j >= 1} (z-z0)^j/j*(zi-z0)^{-j},
     or othwerwise by (pot = 1)
     mp(z) = -m*1/(z-zi)
           = -m*1/(z0-zi+(z-z0))
           = m*sum_{j >= 1} (z-z0)^{j-1}*(zi-z0)^{-j}.
  */
  const dcmplx r = 1.0/((zr+I*zi)-this->z0);
  const double re = creal(r),im = cimag(r);
  dcmplx *coeff = this->coeff2;

  if (pot == 0) {
    double zre = mr*re-mi*im,zim = mi*re+mr*im;
    {
      const dcmplx logr = clog(-r);
      coeff[0] += (mr*creal(logr)-mi*cimag(logr)+
                   I*(mi*creal(logr)+mr*cimag(logr)));
    }
    for (int j = 1; j <= p; j++) {
      coeff[j] += (zre/j+I*zim/j);
      double zre0 = zre*re-zim*im;
      zim = zim*re+zre*im;
      zre = zre0;
    }
  }
  else {
    double zre = mr*re-mi*im,zim = mi*re+mr*im;
    for (int j = 0; j <= p; j++) {
      coeff[j] += (zre+I*zim);
      double zre0 = zre*re-zim*im;
      zim = zim*re+zre*im;
      zre = zre0;
    }
  }
}
/*------------------------------------------------------------------------*/
bool mpexp_theta(const mpexp *this,const mpexp *that, double cutoff)
/* Theta criterion for accurate cluster to cluster interaction between
   two boxes this and that.

   This routine is a workhorse and needs to be fast.

   The maximum relative error in using a multipole expansion from a
   box 'this' to evaluate the field in a box 'that' is error <=
   constant * (R/|R+minimum distance between the boxes|)^(p+1) for
   points |z| > R. Using that the minimum distance is >= d-R-r (where
   d is the center-to-center distance and where r is the radius of the
   smaller box), we are led to the requirement that R/(d-r) <= theta
   and this is what is implemented below (see [4] for the full
   analysis).

   This also means that theta = 1/2 is equivalent to the boxes being
   classically well-separated as in [1,2]. With a general theta, any
   of the two sets may be expanded by a factor of 1/theta and
   arbitrarily rotated about its center point without touching the
   other set.

   However, in that case one should note that the theta used in the
   classical uniform multipole method [2] is actually 1/(2*sqrt(2)-1)
   = 0.5469181 and not 1/2 -- perhaps for convenience, given the
   quadratic boxes.

   Also, in [3] a different and less stringent criterion is used; R <=
   c*d. For uniform boxes, their c = 1/2 would mean theta = sqrt(2)/3
   = 0.4714045 here. Note that trivially, R <= theta/(1+theta)*d
   implies R+theta*r <= theta*d, but the former criterion is less
   sharp when the two boxes have different sizes.

   A slightly sharper result stems from squaring and using Young's
   weighted inequality; then R^2+theta*r^2 <= theta^2/(1+theta)*d^2
   implies R+theta*r <= theta*d. This is used below as a first check
   in order to avoid computing the square-root.

   Also, note that the GKZ-criterion in [3] yields an error
   proportional to (c/(1-2*c))^(p+1) and not to c^(p+1) as they
   claim. Hence c = theta/(1+2*theta) would be appropriate -- this is
   even less sharp compared to the trivial interpretation mentioned
   above. Unfortunately, following their method of proof, the
   currently implemented criterion yields an error proportional to
   (theta/(1-theta))^(p+1), which would suggest that the choice theta
   = 1/2 should be avoided.

   Notes: the GKZ-criterion is found in [3], Eq. (8.31) at p. 379 and
   the error estimate is derived at p. 386. The first estimate [3] in
   (8.40) is incorrect (change (1+2*theta) to (1-2*theta) and the
   second is unsharp (replace (1-2*theta) with 1/(1+2*theta)). In the
   proof that follows, the problematic factor (1-2*theta) is
   inadvertently omitted.

   References:
     [1] J. Carrier, L. Greengard, and V. Rokhlin: "A fast adaptive
         multipole algorithm for particle simulations". SIAM
         J. Sci. Stat. Comput., 9(4):669--686, 1988.
     [2] L. Greengard and V. Rokhlin: "A fast algorithm for particle
         simulations". J.Comput.Phys., 73:325--348, 1987.
     [3] M. Griebel, S. Knapek, and G. Zumbusch: "Numerical Simulation
         in Molecular Dynamics", volume 5 of Texts in Computational
         Science and Engineering. Springer Verlag, Berlin, 2007.
     [4] S. Engblom: "On well-separated sets and fast mutlipole
         methods", 2010. Available at http://arxiv.org/abs/1006.2269.
*/
{
  const double theta2_ = theta*theta;
  const double thetac_ = theta2_/(1.0+theta);
  // avoid cabs():
  const double thisrad = creal(this->d0)*creal(this->d0)+
    cimag(this->d0)*cimag(this->d0);
  const double thatrad = creal(that->d0)*creal(that->d0)+
    cimag(that->d0)*cimag(that->d0);
  const dcmplx d = this->z0-that->z0;
  const double d2 = creal(d)*creal(d)+cimag(d)*cimag(d);

  /* note: criterion uses strict inequality to get rid of the singular
     case when both boxes are the same points (this can happen!) */
  if (thisrad >= thatrad) {
    // sufficient condition (return probability is about 80%)
    if (thisrad+theta*thatrad < thetac_*d2)
      return true;
    else
      return thisrad+theta2_*thatrad+
        2.0*theta*sqrt(thisrad*thatrad) < theta2_*d2;
  }
  else {
    if (thatrad+theta*thisrad < thetac_*d2)
      return true;
    else
      return thatrad+theta2_*thisrad+
        2.0*theta*sqrt(thatrad*thisrad) < theta2_*d2;
  }
}
/*------------------------------------------------------------------------*/
int mpexp_theta2(const mpexp *this,const mpexp *that, double cutoff)
/* 2nd theta criterion: if mpexp_theta() above is false, is it due to
   only one of the boxes? Then the multipoles of the larger box can be
   shifted into a polynomial expansion in the smaller, and the
   multipole expansion in the smaller can be evaluated at each point
   in the larger box. Returns -1 if this is the smaller box (cluster
   to particle) and 1 otherwise (particle to cluster). */
{
  const double theta2_ = theta*theta;
  const double thetac_ = theta2_/(1.0+theta);
  const double thisrad = creal(this->d0)*creal(this->d0)+
    cimag(this->d0)*cimag(this->d0);
  const double thatrad = creal(that->d0)*creal(that->d0)+
    cimag(that->d0)*cimag(that->d0);
  const dcmplx d = this->z0-that->z0;
  const double d2 = creal(d)*creal(d)+cimag(d)*cimag(d);

  // derived as a kind of adjoint to mpexp_theta() above
  if (thatrad >= thisrad) {
    if (thisrad+theta*thatrad < thetac_*d2)
      return 1;
    else
      return thisrad+theta2_*thatrad+
        2.0*theta*sqrt(thisrad*thatrad) < theta2_*d2;
  }
  else {
    if (thatrad+theta*thisrad < thetac_*d2)
      return -1;
    else
      return -(thatrad+theta2_*thisrad+
               2.0*theta*sqrt(thatrad*thisrad) < theta2_*d2);
  }
}
/*------------------------------------------------------------------------*/
void mpexp_eval(const mpexp *this,int p,
                double *vr,double *vi,
                const double *zr,const double *zi,
                const int *ix,int begin,int end)
/* Evaluation of near field at coordinates (zr,zi). */
{
  const dcmplx *coeff = this->coeff2;

  if (vi != NULL)
    for (int i = begin; i < end; i++) {
      const double re = zr[ix[i]]-creal(this->z0),im = zi[ix[i]]-cimag(this->z0);
      double pre = creal(coeff[p]),pim = cimag(coeff[p]);
      
      // Horner's method
      for (int j = p-1; j >= 0; j--) {
        double pre0;
        pre0 = re*pre-im*pim+creal(coeff[j]);
        pim = re*pim+im*pre+cimag(coeff[j]);
        pre = pre0;
      }

      // add the result
      vr[ix[i]] += pre;
      vi[ix[i]] += pim;
    }
  else
    for (int i = begin; i < end; i++) {
      const double re = zr[ix[i]]-creal(this->z0),im = zi[ix[i]]-cimag(this->z0);
      double pre = creal(coeff[p]),pim = cimag(coeff[p]);
      for (int j = p-1; j >= 0; j--) {
        double pre0;
        pre0 = re*pre-im*pim+creal(coeff[j]);
        pim = re*pim+im*pre+cimag(coeff[j]);
        pre = pre0;
      }
      vr[ix[i]] += pre;
    }
}
/*------------------------------------------------------------------------*/
void mpexp_evalmp(const mpexp *this,int p,int pot,
                  double *vr,double *vi,
                  const double *zr,const double *zi,
                  const int *ix,int begin,int end)
/* Evaluation of multipole expansion at coordinates (zr,zi). */
{
  const dcmplx *coeff = this->coeff1;

  if (vi != NULL) {
    for (int i = begin; i < end; i++) {
      const dcmplx r = 1.0/((zr[ix[i]]+I*zi[ix[i]])-this->z0);
      const double re = creal(r),im = cimag(r);
      double pre = creal(coeff[p]),pim = cimag(coeff[p]);

      // Horner's method for inverse polynomial
      for (int j = p-1; j > 0; j--) {
        double pre0;
        pre0 = re*pre-im*pim+creal(coeff[j]);
        pim = re*pim+im*pre+cimag(coeff[j]);
        pre = pre0;
      }
      {
        double pre0;
        pre0 = re*pre-im*pim;
        pim = re*pim+im*pre;
        pre = pre0;
        if (pot == 0) {
          const dcmplx logr = clog(r);
          pre -= creal(coeff[0])*creal(logr)-cimag(coeff[0])*cimag(logr);
          pim -= creal(coeff[0])*cimag(logr)+creal(coeff[0])*cimag(logr);
        }
      }

      vr[ix[i]] += pre;
      vi[ix[i]] += pim;
    }
  }
  else {
    for (int i = begin; i < end; i++) {
      const dcmplx r = 1.0/((zr[ix[i]]+I*zi[ix[i]])-this->z0);
      const double re = creal(r),im = cimag(r);
      double pre = creal(coeff[p]),pim = cimag(coeff[p]);

      // Horner's method for inverse polynomial
      for (int j = p-1; j > 0; j--) {
        double pre0;
        pre0 = re*pre-im*pim+creal(coeff[j]);
        pim = re*pim+im*pre+cimag(coeff[j]);
        pre = pre0;
      }
      {
        double pre0;
        pre0 = re*pre-im*pim;
        pim = re*pim+im*pre;
        pre = pre0;
        if (pot == 0) {
          const dcmplx logr = clog(r);
          pre -= creal(coeff[0])*creal(logr)-cimag(coeff[0])*cimag(logr);
          pim -= creal(coeff[0])*cimag(logr)+creal(coeff[0])*cimag(logr);
        }
      }

      vr[ix[i]] += pre;
    }
  }
}
