/* fmm2d.c -- Fast multipole method for 2-D potentials in free space */

/* M. Holm    2012-04-11 (Added nlevels parameter) */
/* S. Engblom 2009-05-08 (Revision) */
/* S. Engblom 2007-07-08 (Major revision) */
/* S. Engblom 2006-10-11 (Port to Mex and a major revision) */
/* S. Engblom 2005-01-05 */

#include <string.h>

#include "mex.h"
#include "matrix.h"

#include "expint.h"
#include "fmm.h"
#include "direct.h"

// forward declarations
//typedef enum {NDIRECT = 0,SMOOTH = 1,XOPT = 2,ETA = 3,TOL = 4,CUTOFF = 5,NLEVELS = 6,OUT = 7,POT = 8,CONT = 9,ERR_PROPERTY = -1} PROPERTY;
typedef enum {NLEVELS = 0,NDIRECT = 2,ETA = 3,CONT = 4,TOL = 5,OUT = 6,CUTOFF = 8,POT = 9,SMOOTH = 10,THETA = 12,XOPT = 13,ERR_PROPERTY = -1} PROPERTY;
typedef void (*getvalFun)(void *,const mxArray *);

void getXOPT(void *,const mxArray *);
void getCONT(void *,const mxArray *);
void getTOL(void *,const mxArray *);
void getSMOOTH(void *,const mxArray *);
void getCUTOFF(void *,const mxArray *);
void getNDIRECT(void *,const mxArray *);
void getNLEVELS(void *,const mxArray *);
void getETA(void *,const mxArray *);
void getTHETA(void *,const mxArray *);
void getPOT(void *,const mxArray *);
void getOUT(void *,const mxArray *);

PROPERTY getprop(const char *str);
SMOOTHER getsmooth(const char *str);

#define ISDOUBLEMATRIX(A) (mxIsDouble(A) && !mxIsSparse(A) && \
			   mxGetNumberOfDimensions(A) == 2)
#define ISREALSCALAR(A) (!mxIsCell(A) && !mxIsStruct(A) && \
			 !mxIsComplex(A) && mxGetNumberOfElements(A) == 1)
#define ISLOGICALSCALAR(A) (mxIsLogical(A) && \
			    mxGetNumberOfElements(A) == 1)

/*------------------------------------------------------------------------*/
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
  // main input
  const double *zr,*zi,*mr,*mi,*er,*ei;
  int N,NE = 0;


  // values of properties with defaults
  double xopt = 1e-3,cutoff = 2e-3,tol = 1e-6,eta = 1.0,theta = 0.5;
  int pot = 0,out = 0,Ndirect = 5, Nlevels = -1;
  SMOOTHER smooth = DIRAC;
  bool cont = true;

  // output
  double *pr = NULL,*pi = NULL,*qr = NULL,*qi = NULL, *times = NULL;
  int eval_syntax = 2 < nrhs && ISDOUBLEMATRIX(prhs[2]);
  mxComplexity kind;

  // parsing functions
//  static const getvalFun getval[] = {&getNDIRECT,&getSMOOTH,&getXOPT,&getETA,&getTHETA,&getTOL,&getCUTOFF,                                  &getNLEVELS,&getOUT,&getPOT,&getCONT};

//  void *val[] = {&Ndirect,&smooth,&xopt,&eta,&theta,&tol,&cutoff,&Nlevels,&out,&pot,&cont};
  static const getvalFun getval[] = {&getNLEVELS,NULL,&getNDIRECT,&getETA,&getCONT,&getTOL,&getOUT,NULL,&getCUTOFF,&getPOT,&getSMOOTH,NULL,&getTHETA,&getXOPT};

  void *val[] = {&Nlevels,NULL,&Ndirect,&eta,&cont,&tol,&out,NULL,&cutoff,&pot,&smooth,NULL,&theta,&xopt};

  // hidden output: to facilitate testing of expint()
  if (nrhs == 1) {
    double *p = mxGetPr(plhs[0] = mxDuplicateArray(prhs[0]));
    for (int i = 0; i < mxGetNumberOfElements(prhs[0]); i++)
      p[i] = expint(p[i]);
    return;
  }

  // syntax
  if (3 < nlhs || nrhs < 2 || (nrhs-eval_syntax)&1)
    mexErrMsgTxt("Expecting one to three outputs and two or three inputs "
		 "followed by property/value-pairs.");

  // inputs Z and M
  if (!ISDOUBLEMATRIX(prhs[0]) || !ISDOUBLEMATRIX(prhs[1]) ||
      mxGetM(prhs[0]) != mxGetM(prhs[1]) || 
      mxGetN(prhs[0]) != mxGetN(prhs[1]))
    mexErrMsgTxt("First two arguments must be double matrices of " 
		 "matching sizes.");
  N = mxGetNumberOfElements(prhs[0]);
  if (N != 0 && !mxIsComplex(prhs[0]))
    mexErrMsgTxt("First argument must be complex.");

  zr = mxGetPr(prhs[0]); zi = mxGetPi(prhs[0]);
  mr = mxGetPr(prhs[1]); mi = mxGetPi(prhs[1]);

  // input ZE -- or just (er,ei) for short
  if (eval_syntax) {
    NE = mxGetNumberOfElements(prhs[2]);
    if (NE != 0 && !mxIsComplex(prhs[2]))
      mexErrMsgTxt("Evaluation points must be complex.");
    er = mxGetPr(prhs[2]); ei = mxGetPi(prhs[2]);
  }

  // parse property/value-pairs in an automatic fashion
  for (int i = 2+eval_syntax; i < nrhs-1; i += 2) {
    PROPERTY prop;
    char str[10];

    // determine property
    if (!mxIsChar(prhs[i]))
      mexErrMsgTxt("Expecting a property name.");
    if (mxGetString(prhs[i],str,10) != 0 ||
	(prop = getprop(str)) == ERR_PROPERTY)
      mexErrMsgTxt("Unknown property.");

    // get its value
    getval[prop](val[prop],prhs[i+1]);
  }

  // remaining checks
  if (nlhs > 1 && !eval_syntax && out != 1 && out != 2)
    mexErrMsgTxt("Three input matrices required for two outputs.");

  if (cutoff <= xopt)
    mexErrMsgTxt("Parameter 'cutoff' must be larger than 'xopt'.");

  // complexity
  kind = pot == 1 || mi != NULL ? mxCOMPLEX : mxREAL;
  
  // if Nlevels is set, override Ndirect.  Otherwise, do nothing.
  if(Nlevels != -1)
  {
    Ndirect = N/(1<<(2*Nlevels));
    if(Ndirect < 1)
       Ndirect = 1;
  }

  // allocate and produce output
  if (out == 1) {
    // special syntax: return FMM-mesh only
    fmm2dGetMesh(nlhs,plhs,
		 N,zr,zi,mr,mi,NE,er,ei,
		 pot,tol,Ndirect,eta,theta);
    return;
  }
  else if (nlhs == 2 && out != 2) {
    // evaluate both at potentials Z and at coordinates ZE
    plhs[0] = mxCreateDoubleMatrix(mxGetM(prhs[0]),mxGetN(prhs[0]),kind);
    pr = mxGetPr(plhs[0]); pi = mxGetPi(plhs[0]);

    plhs[1] = mxCreateDoubleMatrix(mxGetM(prhs[2]),mxGetN(prhs[2]),kind);
    qr = mxGetPr(plhs[1]); qi = mxGetPi(plhs[1]);
  }
  else if (nlhs == 3 && out == 2) {
    // evaluate both at potentials Z and at coordinates ZE
    plhs[0] = mxCreateDoubleMatrix(mxGetM(prhs[0]),mxGetN(prhs[0]),kind);
    pr = mxGetPr(plhs[0]); pi = mxGetPi(plhs[0]);

    plhs[1] = mxCreateDoubleMatrix(mxGetM(prhs[2]),mxGetN(prhs[2]),kind);
    qr = mxGetPr(plhs[1]); qi = mxGetPi(plhs[1]);
   // plhs[2] = mxCreateDoubleScalar(0);    
    plhs[2] = mxCreateDoubleMatrix(8,1,mxREAL);
    times = mxGetPr(plhs[3]);
  }
  else if (!eval_syntax || zr == er && zi == ei && N == NE) {
    // only at potentials Z
    NE = 0;   
    plhs[0] = mxCreateDoubleMatrix(mxGetM(prhs[0]),mxGetN(prhs[0]),kind);
    pr = mxGetPr(plhs[0]); pi = mxGetPi(plhs[0]);
    if (out == 2 && nlhs == 2)
    {
      // plhs[1] = mxCreateDoubleScalar(0);
       //times = mxGetPr(plhs[1]);        
       plhs[1] = mxCreateDoubleMatrix(8,1,mxREAL);
       times = mxGetPr(plhs[1]);
    }
  }
  else {
    // at coordinates ZE
    plhs[0] = mxCreateDoubleMatrix(mxGetM(prhs[2]),mxGetN(prhs[2]),kind);
    qr = mxGetPr(plhs[0]); qi = mxGetPi(plhs[0]);
    if (out == 2 && nlhs == 2)
    {
      plhs[1] = mxCreateDoubleScalar(0);
      //times = mxGetPr(plhs[1]);      
       plhs[2] = mxCreateDoubleMatrix(8,1,mxREAL);
       times = mxGetPr(plhs[2]);
    }
  }

  // evaluate
  if (tol == 0.0)
    directInteract(N,zr,zi,mr,mi,NE,er,ei,pr,pi,qr,qi,
		   pot,smooth,xopt,cutoff,cont);
  else
    fmm2dInteract(N,zr,zi,mr,mi,NE,er,ei,pr,pi,qr,qi,
		  pot,tol,Ndirect,theta,eta,smooth,xopt,cutoff,cont,times);

}
/*------------------------------------------------------------------------*/
PROPERTY getprop(const char *str)
/* Small perfect hash for the supported properties. Note that str[0]
   and str[1] must be readable for this to work. Returns ERR_PROPERTY
   on failure. */
{
  static const char *tab[] = {"nlevels","","ndirect","eta","cont","tol","out","","cutoff","pot","smooth","","theta","xopt"};
  const unsigned hash = ((str[0] * 1)+(str[1] * 9)+(str[2] * 8)) % 14;

  return str[0] != '\0' && strcmp(str,tab[hash]) == 0 ? 
    hash : ERR_PROPERTY;
}
/*------------------------------------------------------------------------*/
void getXOPT(void *xopt_,const mxArray *rhs)
{
  double *xopt = (double *)xopt_;

  if (!ISREALSCALAR(rhs))
      mexErrMsgTxt("Expecting a real and scalar option 'xopt'.");
  *xopt = *mxGetPr(rhs);
  if (*xopt <= 0)
    mexErrMsgTxt("Property 'xopt' must be strictly positive.");
}
/*------------------------------------------------------------------------*/
void getCONT(void *cont_,const mxArray *rhs)
{
  bool *cont = (bool *)cont_;

  if (!ISLOGICALSCALAR(rhs))
    mexErrMsgTxt("Expecting a logical scalar option 'cont'.");
  *cont = mxIsLogicalScalarTrue(rhs);
}
/*------------------------------------------------------------------------*/
void getTOL(void *tol_,const mxArray *rhs)
{
  double *tol = (double *)tol_;

  if (!ISREALSCALAR(rhs))
      mexErrMsgTxt("Expecting a real and scalar option 'tol'.");
  *tol = *mxGetPr(rhs);
  if (*tol < 0)
    mexErrMsgTxt("Property 'tol' must not be negative.");
}
/*------------------------------------------------------------------------*/
void getSMOOTH(void *smooth_,const mxArray *rhs)
{
  SMOOTHER *smooth = (SMOOTHER *)smooth_;
  char str[10];

  if (!mxIsChar(rhs))
    mexErrMsgTxt("Expecting a character array as option 'smooth'.");
  if (mxGetString(rhs,str,10) != 0 || 
      (*smooth = getsmooth(str)) == ERR_SMOOTHER)
    mexErrMsgTxt("Unknown smoother.");
}
/*------------------------------------------------------------------------*/
void getCUTOFF(void *cutoff_,const mxArray *rhs)
{
  double *cutoff = (double *)cutoff_;

  if (!ISREALSCALAR(rhs))
      mexErrMsgTxt("Expecting a real and scalar option 'cutoff'.");
  *cutoff = *mxGetPr(rhs);
  if (*cutoff <= 0)
    mexErrMsgTxt("Property 'cutoff' must be strictly positive.");
}
/*------------------------------------------------------------------------*/
void getNLEVELS(void *Nlevels_,const mxArray *rhs)
{
  int *Nlevels = (int *)Nlevels_;

  if (!ISREALSCALAR(rhs))
    mexErrMsgTxt("Expecting a real and scalar option 'nlevels'.");
  *Nlevels = *mxGetPr(rhs);
  if (*Nlevels <= 0)
    mexErrMsgTxt("Property 'nlevels' must be strictly positive.");
}
/*------------------------------------------------------------------------*/
void getNDIRECT(void *Ndirect_,const mxArray *rhs)
{
  int *Ndirect = (int *)Ndirect_;

  if (!ISREALSCALAR(rhs))
    mexErrMsgTxt("Expecting a real and scalar option 'ndirect'.");
  *Ndirect = *mxGetPr(rhs);
  if (*Ndirect <= 0)
    mexErrMsgTxt("Property 'ndirect' must be strictly positive.");
}
/*------------------------------------------------------------------------*/
void getETA(void *eta_,const mxArray *rhs)
{
  double *eta = (double *)eta_;

  if (!ISREALSCALAR(rhs))
      mexErrMsgTxt("Expecting a real and scalar option 'eta'.");
  *eta = *mxGetPr(rhs);
  if (*eta < 0 || 1 < *eta)
    mexErrMsgTxt("Property 'eta' must be in [0,1].");
}
/*------------------------------------------------------------------------*/
void getTHETA(void *theta_,const mxArray *rhs)
{
  double *theta = (double *)theta_;

  if (!ISREALSCALAR(rhs))
      mexErrMsgTxt("Expecting a real and scalar option 'theta'.");
  *theta = *mxGetPr(rhs);
  if (*theta < 0 || 1 < *theta)
    mexErrMsgTxt("Property 'theta' must be in [0,1].");
}
/*------------------------------------------------------------------------*/
void getPOT(void *pot_,const mxArray *rhs)
{
  int *pot = (int *)pot_;

  if (!ISREALSCALAR(rhs))
    mexErrMsgTxt("Expecting a real and scalar option 'pot'.");
  *pot = *mxGetPr(rhs);
  if (*pot != 0 && *pot != 1)
    mexErrMsgTxt("Property 'pot' must be 0 or 1.");
}
/*------------------------------------------------------------------------*/
void getOUT(void *out_,const mxArray *rhs)
{
  int *out = (int *)out_,err = 0;
  char str[10];

  if (!mxIsChar(rhs))
    mexErrMsgTxt("Expecting a character array as option 'out'.");
  if (mxGetString(rhs,str,12) != 0)
    err = 1;
  else if (strcmp(str,"sol") == 0)
    *out = 0;
  else if (strcmp(str,"mesh") == 0)
    *out = 1;
  else if (strcmp(str,"times") == 0)
    *out = 2;
  else
    err = 1;

  if (err)
    mexErrMsgTxt("Property 'out' must be 'sol' or 'mesh' or 'times'.");
}
/*------------------------------------------------------------------------*/
SMOOTHER getsmooth(const char *str)
/* Small perfect hash for the supported smoothers. Note that str[4]
   must be readable for this to work. Returns ERR_SMOOTHER on
   failure. */
{
  static const char *tab[] = {"scully","rankine","oseen","dirac"};
  const unsigned hash = ((unsigned)str[4])%4;

  return strcmp(str,tab[hash]) == 0 ? hash : ERR_SMOOTHER;
}
/*------------------------------------------------------------------------*/
