/* fmmshift.c */

/* S. Engblom 2010-02-18 */

#include <string.h>
#include "fmmshift.h"

/*------------------------------------------------------------------------*/

/* First version of file fmmshift.c: shift-operators implemented
   through Horner-like recursions. */

// workspace arrays
static dcmplx *wksp=NULL,*rpow=NULL;

// scalars defining shift-type and sizes
int pshift,potshift;

/*------------------------------------------------------------------------*/
void shift_alloc(int p,int pot,int maxm2p)
/* Allocates and initializes static data given a polynomial order p,
   type of potential pot and maximum number of cluster to cluster
   interactions maxm2p. */
{
  // workspace arrays
  if(wksp != NULL)
    FREE(wksp);
  wksp = MALLOC(2*(p+1)*sizeof(dcmplx));

  // powers of distances (used by m2ps)
  if(rpow != NULL)
    FREE(rpow);
  rpow = MALLOC((p+1)*sizeof(dcmplx));
  rpow[0] = 1.0;

  pshift = p;
  potshift = pot;
}
/*------------------------------------------------------------------------*/
void shift_free(void)
/* Deallocates data. */
{
  FREE(rpow);
  rpow = NULL;
  FREE(wksp);
  wksp = NULL;
  pshift = 0;
}
/*------------------------------------------------------------------------*/
void shift_m2m(dcmplx r0,dcmplx r1,dcmplx r2,dcmplx r3,
	       dcmplx *this,const dcmplx *that)
/* Multipole to multipole shift from 4 children to 1 parent; this +=
   shift(that[0...p]) {child 0} + shift(that[p+1...2*p]) {child 1} +
   ...  and so on. */
{
  dcmplx rpow[4],mul;
  rpow[0] = r0;
  rpow[1] = r1;
  rpow[2] = r2;
  rpow[3] = r3;

  for(int i = 0; i < 4; i++, that += pshift+1) {
    memcpy(wksp,that,(pshift+1)*sizeof(dcmplx));

    for(int k = pshift; k >= 2; k--)
      for(int j = k; j <= pshift; j++)
        wksp[j] += rpow[i]*wksp[j-1];

    if (potshift == 0) {
      mul = rpow[i];
      this[0] += that[0];
      for (int k = 1; k <= pshift; k++) {
	this[k] += wksp[k]-(that[0]/k)*mul;
	mul *= rpow[i];
      }
    }
    else
      for(int k = 0; k <= pshift; k++)
        this[k] += wksp[k];
  }
}
/*------------------------------------------------------------------------*/
void shift_m2ps(dcmplx z0,const void *zi,size_t siz,
		const int *ix,int begin,int end,
		dcmplx *this2, const dcmplx *that1,
        dcmplx *rpow_, dcmplx *wksp1_)
/* Symmetric multipole to polynomial shift; this2 +=
   shift(that1[ix[i]]) followed by that2[ix[i]] += shift(this1) for
   index i in [begin,end). Center coordinates of the boxes are given
   in z0 and zi (in which each entry is separated by siz bytes).

   A somewhat singular feature is that a negative index in ix,
   pointing to a box at a previous level, is detected and implies that
   the 3 following indices are ignored. This ensures that connections
   between levels are properly handled and that the connectivity
   indices are convenient to assemble. */
{ 
  
  for (int i = begin; i < end; i++) {
    const dcmplx r0 = 1.0/(*(dcmplx *)(zi+ix[i]*siz)-z0); 
    const double re = creal(r0),im = cimag(r0);
    double *ri;
    if(rpow_ != NULL)
      ri = (double *)rpow_;
    else
      ri = (double *)rpow;
    
    double *wksp1;
    if(wksp1_ != NULL){
      wksp1 = (double *)wksp1_; 
    }else{
      wksp1 = (double *)&wksp[0] ;
    }
    const dcmplx *that1i = &that1[ix[i]*(pshift+1)];
    double mre = 1.0,mim = 0.0,pre,pim;

    // prescale
    for (int j = 1; j < pshift; ) {
      ri[2*j] = mre*re-mim*im;
      ri[2*j+1] = mre*im+mim*re;
      mre = ri[2*j];
      mim = ri[2*j+1];
      pre = creal(that1i[j])*mre-cimag(that1i[j])*mim;
      pim = creal(that1i[j])*mim+cimag(that1i[j])*mre;
      wksp1[2*j-2] = -pre;
      wksp1[2*j-1] = -pim;
      j++;
 
      ri[2*j] = mre*re-mim*im;
      ri[2*j+1] = mre*im+mim*re;
      mre = ri[2*j];
      mim = ri[2*j+1];
      pre = creal(that1i[j])*mre-cimag(that1i[j])*mim;
      pim = creal(that1i[j])*mim+cimag(that1i[j])*mre;
      wksp1[2*j-2] = pre;
      wksp1[2*j-1] = pim;
      j++;
    }
    if (pshift&1) {
      ri[2*pshift] = mre*re-mim*im;
      ri[2*pshift+1] = mre*im+mim*re;
      mre = ri[2*pshift];
      mim = ri[2*pshift+1];
      pre = creal(that1i[pshift])*mre-cimag(that1i[pshift])*mim;
      pim = creal(that1i[pshift])*mim+cimag(that1i[pshift])*mre;
      wksp1[2*pshift-2] = -pre;
      wksp1[2*pshift-1] = -pim;
    }
    wksp1[2*pshift] = wksp1[2*pshift+1] = 0.0;;

    // (1) "upward" phase
    for (int k = 2; k <= pshift; k++)
      for (int j = pshift-k; j < pshift; j++) {
	wksp1[2*j] += wksp1[2*(j+1)];
	wksp1[2*j+1] += wksp1[2*(j+1)+1];
      }

    // (2) "downward" phase
    for (int k = pshift; k > 0; k--)
      for (int j = k; j <= pshift; j++) {
	wksp1[2*j] += wksp1[2*(j-1)];
	wksp1[2*j+1] += wksp1[2*(j-1)+1];
      }

    // postscale
    if (potshift == 0) {      
      const double red0 = creal(that1[ix[i]*(pshift+1)]),
	imd0 = cimag(that1[ix[i]*(pshift+1)]);

      {
	const dcmplx logr = clog(ri[2]+I*ri[3]);
	this2[0] += wksp1[0]+I*wksp1[1]
	  -(red0*creal(logr)-imd0*cimag(logr)+
	    I*(imd0*creal(logr)+red0*cimag(logr)));
      }
      for (int j = 1; j < pshift; ) {
	pre = (wksp1[2*j]-red0/j)*ri[2*j]-
	  (wksp1[2*j+1]-imd0/j)*ri[2*j+1];
	pim = (wksp1[2*j]-red0/j)*ri[2*j+1]+
	  (wksp1[2*j+1]-imd0/j)*ri[2*j];
	this2[j] += pre+I*pim;	j++;

	pre = (wksp1[2*j]-red0/j)*ri[2*j]-
	  (wksp1[2*j+1]-imd0/j)*ri[2*j+1];
	pim = (wksp1[2*j]-red0/j)*ri[2*j+1]+
	  (wksp1[2*j+1]-imd0/j)*ri[2*j];
	this2[j] += pre+I*pim;	j++;
      }
      if (pshift&1) {
	pre = (wksp1[2*pshift]-red0/pshift)*ri[2*pshift]-
	  (wksp1[2*pshift+1]-imd0/pshift)*ri[2*pshift+1];
	pim = (wksp1[2*pshift]-red0/pshift)*ri[2*pshift+1]+
	  (wksp1[2*pshift+1]-imd0/pshift)*ri[2*pshift];
	this2[pshift] += pre+I*pim;      }
    }
    else { 
      for (int j = 0; j < pshift; ) {
	pre = wksp1[2*j]*ri[2*j]-wksp1[2*j+1]*ri[2*j+1];
	pim = wksp1[2*j]*ri[2*j+1]+wksp1[2*j+1]*ri[2*j];
	this2[j] += pre+I*pim;
    j++;

	pre = wksp1[2*j]*ri[2*j]-wksp1[2*j+1]*ri[2*j+1];
	pim = wksp1[2*j]*ri[2*j+1]+wksp1[2*j+1]*ri[2*j];
	this2[j] += pre+I*pim;	
	j++;
      }
      if (~pshift&1) {
	pre = wksp1[2*pshift]*ri[2*pshift]-wksp1[2*pshift+1]*ri[2*pshift+1];
	pim = wksp1[2*pshift]*ri[2*pshift+1]+wksp1[2*pshift+1]*ri[2*pshift];
	this2[pshift] += pre+I*pim;
      }
    }

    // in case the shift was managed via parents
    if (ix[i] < 0) i += 3;
  }
}
/*------------------------------------------------------------------------*/
void shift_p2p(dcmplx r0,dcmplx r1,dcmplx r2,dcmplx r3,
	       dcmplx *this,const dcmplx *that,
         dcmplx *wksp_)
/* Polynomial to polynomial shift from 1 parent to 4 children;
   this[0..p] += shift(that) {child 0}, this[p+1..2*p] += shift(that)
   {child 1}, and so on. */
{
  dcmplx * wkspace;
  if(wksp_ != NULL)
    wkspace = wksp_;
  else
    wkspace = wksp;
  dcmplx rpow[4];
  rpow[0] = r0;
  rpow[1] = r1;
  rpow[2] = r2;
  rpow[3] = r3;

  for(int i = 0; i < 4; i++, this += pshift+1) {
    memcpy(wkspace,that,(pshift+1)*sizeof(dcmplx));

    for (int k = 1; k <= pshift; k++)
      for(int j = pshift-k; j < pshift; j++)
	wkspace[j] -= rpow[i]*wkspace[j+1];

    for(int k = 0; k <= pshift; k++)
      this[k] += wkspace[k];
  }
}
/*------------------------------------------------------------------------*/

