function make(pustr,varargin)
%MAKE Makefile for FMM2D.
%   MAKE('CPU' | 'GPU', ARG1,...) compiles FMM2D for CPU/GPU with the
%   compiler arguments ARG1, ... using the MEX-command. With no
%   arguments, default platform-dependent define arguments are used
%   (passed to MEX through the -D switch).

% M. Holm 2012-10-25 (removed PRINTTIMES define)
% S. Engblom 2012-08-23 (GPU/CPU-syntax)
% S. Engblom 2012-06-26 (attempt at mexmaci)
% M. Holm 2012-02-07 (removed deprecated defines)
% M. Holm 2012-01-23 (CUDADIRECT)
% S. Engblom 2011-07-15 (FMM_HORNERSHIFT)

% determine platform
if nargin == 0
  fprintf(1,'Defaulting to GPU-make.\n');
  cpu = false;
elseif strcmpi(pustr,'CPU')
  cpu = true;
elseif strcmpi(pustr,'GPU')
  cpu = false;
else
  error('First argument must be a valid processing unit (CPU/GPU).');
end

% Public defines:
%
% -use '-DmwIndex=int' for platforms where mex is not automatically
% linked against the library defining the mwIndex type,
%
% For experimental use:
%
% '-DFMM_SHRINK' (recommended) sets Postsplit shrinking of boxes on/off,
%
% '-DFMM_MESHRATIO=<value>' (default 2.0) tunes the criterion for
% splitting long edges in the mesh.
%
% '-DNUMTHREADS=<value>' (default 4) number of CPU threads
%
% '-DNUMTASKS=<value>' (default 3) number of tasks = 4^NUMTASKS, sets available 
% paralellism in downward pass of FMM.
% Good choice is s.t. number of tasks = 16*NUMTHREADS.
%
% '-DTASKLOG' Will output a log of task execution to log_X.log, where X is a
% hexadecimal counter, reset whenever the .mex is reloaded. This feature is not 
% supported on OSX platforms.
%
% Make some effort in finding good optimization flags!

mx = mexext;
ver = version;

% possible defines
define = {'-DmwIndex=int' ...              % (1)
          '-DFMM_SHRINK' ...               % (2)
          '-DNUMTHREADS=4' ...             % (3)
          '-DNUMTASKS=3' ...               % (4) 
          '-DTASKLOG'};                    % (5)
% (edit defaults below)

% libdirect is set to '-ldirect_cuda' och '-ldirect_cpu' depending on 
% which hardware direct evaluation is going to be run
if cpu
  libdirect = '-ldirect_cpu';
  archdefine = '-DCPUONLY';
  archlibs = '';
else
  libdirect = '-ldirect_cuda';
  archdefine = '-DCUDA';
  archlibs = ' -L/usr/local/cuda/lib64 -lcudart -lrt';
end

% path, include and link path, source
s = pwd;
include = {['-I' s '/source'] ...
	   ['-I' s '/source/tl']};
clibs = ['CLIBS=-L' s '/source -L.'];
source = {[s '/source/fmm2d.c'] ...
          [s '/source/fmm.c'] ...
          [s '/source/fmmshift.c'] ...                    
          [s '/source/tl.cpp']};
% [s '/source/expint.c'] ...          
% [s '/source/direct.c'] ...          

% make libdirect using Makefile   
% set direct_cuda for GPU, direct_cpu for CPU, after defines above
if strcmp(libdirect, '-ldirect_cpu')   
  disp('Using CPU to calculate direct interactions')   
  unix(sprintf('rm libdirect_cpu.a; make direct_cpu MLDIR=%s DIRECTTHREADS=4',matlabroot));
elseif strcmp(libdirect, '-ldirect_cuda')
  disp('Using GPU (CUDA) to calculate direct interactions')
  unix(sprintf('rm libdirect_cuda.a; make direct_cuda MLDIR=%s',matlabroot));
end

if strcmp(mx,'mexa64')
  if ~strncmp(version,'7.8',3) && ~strncmp(version,'7.11',4) && ...
        ~strncmp(version,'7.12',4) && ~strncmp(version,'7.13',4) && ...
     ~strncmp(version,'8.1',3)
    warning(['Extension .' mexext ...
             ' tested with Matlab version(s) 7.8, 7.11, 7.12, 7.13, and 8.1 only.']);
  end

  cflags = ['CFLAGS=-fPIC -fno-omit-frame-pointer -std=c99 -O3 ' ...
            '-D_GNU_SOURCE -fexceptions'];
  clibs = [clibs ' -L/usr/lib64 -lmwblas'];
  
  if strncmp(version,'7.11',4)
    clibs = [clibs ' -L' matlabroot '/sys/os/glnxa64'];
  else 
    clibs = [clibs ' -L' matlabroot '/bin/glnxa64']; 
  end
  cxxlibs = ['CXXLIBS=' ' -L' matlabroot '/bin/glnxa64' ...
             ' -L' matlabroot '/sys/os/glnxa64' ...
             ' -lmx -lmex -lmat -lm -lmwblas -L/usr/lib64 -lstdc++ ' ...
             libdirect];
  cxxlibs = [cxxlibs archlibs];
  cxxflags = ['CXXFLAGS=-g -fPIC -pthread'];
  
  define = define([2:5]);
  define = [define archdefine];
  if nargin > 1, define = varargin; end
  fprintf(1,['MAKE ' sprintf('%s ',define{:}) '\n']);
  mex('-g',cflags,define{:},include{:},clibs, ...
      cxxflags, cxxlibs, ...
      '-largeArrayDims','-outdir',s, ...
      source{:});
  if strcmp(libdirect, '-ldirect_cpu')   
    unix('cp fmm2d.mexa64 fmm2dcpu.mexa64');
  elseif strcmp(libdirect, '-ldirect_cuda')
    unix('cp fmm2d.mexa64 fmm2dgpu.mexa64');
  end
elseif strcmp(mx,'mexmaci')
  error('Platform not supported.');
elseif strcmp(mx,'mexmaci64')
  error('Platform not supported.');
elseif strcmp(mx,'mexmac')
  error('Platform never tried.');
elseif strcmp(mx,'mexw32')
  error('Platform never tried.');
elseif strcmp(mx,'mexglx')
  error('Platform never tried.');
elseif strcmp(mx,'mexs64')
  error('Platform never tried.');
else
  error('New platform. No make.');
end
