/* tl.hcpp -- implementation file for SuperGlue library interface */

extern "C" {
#include "tl.h"
}
#include "tasklib.hpp"
#include "options/defaults.hpp"

#include <cassert>
#include <map>
#include <vector>

#define TASK_NAMES

#ifdef TASK_NAMES
#undef IF_DEBUG
#define IF_DEBUG(a) a
#else
#undef IF_DEBUG
#define IF_DEBUG(a)
#endif

#ifdef TASKLOG
template<typename Options>
struct Instrumentation {
    enum { NUM_PERF_COUNTERS = 2 };
    Time::TimeUnit start, stop;
    PerformanceCounter pc[NUM_PERF_COUNTERS];

    unsigned long long startc[NUM_PERF_COUNTERS], stopc[NUM_PERF_COUNTERS];

    void stopMeasure() {
        stop = Time::getTime();
        for (size_t i = 0; i < NUM_PERF_COUNTERS; ++i)
            pc[i].stop();
        for (size_t i = 0; i < NUM_PERF_COUNTERS; ++i)
            stopc[i] = pc[i].readCounter();
    }
    void startMeasure() {
        for (size_t i = 0; i < NUM_PERF_COUNTERS; ++i)
            startc[i] = pc[i].readCounter();
        for (size_t i = 0; i < NUM_PERF_COUNTERS; ++i)
            pc[i].start();
        start = Time::getTime();
    }

    void runTaskBefore(TaskBase<Options> *) {
        startMeasure();
    }
    void runTaskAfter(TaskBase<Options> *task) {
        stopMeasure();
        std::stringstream ss;
        ss << task->getName().c_str() << " ";
        for (size_t i = 0; i < NUM_PERF_COUNTERS; ++i)
          ss << (stopc[i]-startc[i]) << " ";
        Log<Options>::addEvent(ss.str().c_str(), start, stop);
    }
    void init() {
        pc[0].init(PERF_TYPE_HARDWARE,PERF_COUNT_HW_CACHE_MISSES);
        pc[1].init(PERF_TYPE_HW_CACHE, (PERF_COUNT_HW_CACHE_DTLB | (PERF_COUNT_HW_CACHE_OP_READ << 8) | (PERF_COUNT_HW_CACHE_RESULT_MISS << 16)));
    }
    static void destroy() {}
    static void taskNotRunDeps() {}
    static void taskNotRunLock() {}
    static void setMainThreadFlag() {}
};
#endif

struct Options : public DefaultOptions<Options> {
  enum { ThreadWorkspace_size = 128*1024 }; // 128 kB workspace per thread
  typedef void ThreadWorkspace;             // Each thread keeps workspace for tasks
  typedef void PassTaskExecutor;            // Send task executor along to task (needed to get threadworkspace)
  typedef void Logging;
  #ifdef TASKLOG
  typedef Instrumentation<Options> TaskExecutorInstrumentation;
 // typedef void Logging_Barrier;           // Log barrier activity
  #endif

#ifdef TASK_NAMES
  typedef void TaskName;                    // Each task has a name
#endif // TASK_NAMES
};

ThreadManager<Options> *tm;

Handle<Options> **tl_unique(Handle<Options> **start, Handle<Options> **end) {
  std::sort(start, end);
  return std::unique(start, end);
}

class FMMTask : public TaskDynamic<Options> {
protected:
  tl_task_function function;
  char *args;
#ifdef TASK_NAMES
  std::string name;
#endif // TASK_NAMES
public:
  FMMTask(tl_task_function function_,
          void *args_, int argsize,
          Handle<Options> **handles,
          int num_reads,
          int num_writes,
          int num_adds
#ifdef TASK_NAMES
          , const char *name_
#endif // TASK_NAMES
   ) : function(function_) {
    args = new char[argsize];
    memcpy(args, args_, argsize);
    int idx = 0;
    for (int i = 0; i < num_reads; ++i)
      registerAccess(ReadWriteAdd::read, handles[idx++]);

    Handle<Options> **writeStart = handles + num_reads;
    Handle<Options> **writeEnd = tl_unique(writeStart, writeStart+num_writes);
    for (Handle<Options> **i = writeStart; i != writeEnd; ++i)
      registerAccess(ReadWriteAdd::write, handles[idx++]);

    Handle<Options> **addStart = handles + num_reads + num_writes;
    Handle<Options> **addEnd = tl_unique(addStart, addStart+num_adds);
    for (Handle<Options> **i = addStart; i != addEnd; ++i)
      registerAccess(ReadWriteAdd::add, handles[idx++]);

#ifdef TASK_NAMES
    name = name_;
#endif // TASK_NAMES
  }
  void run(TaskExecutor<Options> *taskExecutor) {
    function((tl_thread_info_t) taskExecutor, args, 0);
  }
  ~FMMTask() {
    delete[] args;
  }
#ifdef TASK_NAMES
  std::string getName() { return name; }
#endif // TASK_NAMES
};

void tl_barrier() {
  tm->barrier();
}

void tl_add_task(tl_task_function function, void *args, int argsize,
                 const char * IF_DEBUG(name) ) {
  FMMTask *t = new FMMTask(function, args, argsize, NULL, 0, 0, 0
#ifdef TASK_NAMES
                     , name
#endif // TASK_NAMES
                     );
  tm->addTask(t);
}

void tl_add_task_dep(tl_task_function function, void *args, int argsize,
                     tl_handle_t *handles, 
                     int num_reads, int num_writes, int num_adds,
                     const char * IF_DEBUG(name) ) {
  FMMTask *t = new FMMTask(function, args, argsize,
                           (Handle<Options> **) handles, 
                           num_reads, num_writes, num_adds
#ifdef TASK_NAMES
                     , name
#endif // TASK_NAMES
                     );
  tm->addTask(t);
}


tl_handle_t *tl_create_handles(int num) {
  tl_handle_t *mem = new tl_handle_t[num];
  for (int i = 0; i < num; ++i)
    mem[i] = (tl_handle_t) new Handle<Options>();
  return mem;
}

void tl_destroy_handles(tl_handle_t *handles, int num) {
  for (int i = 0; i < num; ++i)
     delete (Handle<Options> *) handles[i];
}

void *tl_get_workspace(tl_thread_info_t threadinfo, int size) {
  return tm->getWorkspace((TaskExecutor<Options> *) threadinfo, size);
}

void tl_init(int n) {
  tm = new ThreadManager<Options>(n);
}

void tl_destroy() {

#ifdef TASKLOG
  static int counter=0;
  char logout[40];
  sprintf(logout, "log_%x.log", ++counter);
  printf("Writing to log: [%s]\n", logout);
  Log<Options>::dump(logout);
#endif
  Log<Options>::clear();

  delete tm;
}


unsigned long long tl_log_time() {
  return Time::getTime();
}

void tl_log(const char *name, unsigned long long start, unsigned long long stop) {
  Log<Options>::addEvent(name, start, stop);
}


