function [z,m,t] = vortexfix(z,m,t,tol,mtol)
%VORTEXFIX Merging and annihilation of vortices.
%   [Z,M,T] = VORTEXFIX(Z,M,T,tol,mtol) merges all the (complex)
%   coordinates Z that lie in a common box of dimensions tol-by-tol.
%
%   The merging preserves the first moment with respect to M as
%   follows: weighted (according to M) averages in both Z and T are
%   computed and the corresponding sum in M is evaluated. It is
%   assumed that the number of elements in the inputs (Z,M,T) agree
%   and that M is purely real.
%
%   Positive and negative vortices are first merged separately and
%   then resulting vortices in the same box of different signs are
%   either merged to the position of the stronger vortex, or if
%   abs(M1+M2) < mtol*abs(M1-M2), they are both annihilated. The
%   default value of mtol is eps(1000).
%
%   Example:
%     % 10000 random points
%     np = 10000;
%     z = 2*complex(rand(np,1),rand(np,1))-(1+1i);
%     m = 2*rand(np,1)-1; t = rand(np,1);
%
%     % only merge points outside a circle
%     ixok = abs(z) >= 0.5;
%     [zn,mn,tn] = vortexfix(z(ixok),m(ixok),t(ixok),1e-1);
%     zn = [z(~ixok); zn];
%     mn = [m(~ixok); mn];
%     tn = [t(~ixok); tn];
%
%     figure, plot(zn(mn > 0),'r.');
%     hold on, plot(zn(mn < 0),'b.');
%     axis equal, axis([-1 1 -1 1]);

% S. Engblom 2007-10-12 (Minor revision)
% S. Engblom 2007-09-30 (Revision)
% S. Engblom 2007-09-13

% check input
if ~isreal(m) || numel(z) ~= numel(m) || numel(m) ~= numel(t) || ...
  ~isscalar(tol) || tol <= 0
  error('Invalid input.');
end

% defaults
if nargin < 5, mtol = eps(1000); end

% column vectors
z = z(:); m = m(:); t = t(:);

% coordinates
x = real(z); y = imag(z);

% bounding box
xmin = min(x); hx = max(x)-xmin;
ymin = min(y); hy = max(y)-ymin;

% number of boxes
nx = max(1,ceil(hx/tol)); ny = max(1,ceil(hy/tol));

% a possibly slightly larger bounding box
hx = nx*tol; hy = ny*tol;

% change to local coordinates in [0,1]
x = (x-xmin)/hx; y = (y-ymin)/hy;

% linear index in an imagined nx-by-ny matrix
ix = 1+min(floor(x*nx),nx-1)+min(floor(y*ny),ny-1)*nx;

% separate positive and negative m's
i2 = m < 0;
i1 = find(~i2);
i2 = find(i2);

% count the number of coordinates in each box
boxcount1 = sparse(ix(i1),1,1,nx*ny,1);
boxcount2 = sparse(ix(i2),1,1,nx*ny,1);

% determine the boxes that should be merged
boxmerge1 = find(boxcount1 > 1);
boxmerge2 = find(boxcount2 > 1);

% find the corresponding coordinates (this is tricky!)
B1 = sparse(boxmerge1,1,1:numel(boxmerge1),nx*ny,1);
[ixmerge1,foo,ixs1] = find(B1(ix(i1)));
B2 = sparse(boxmerge2,1,1:numel(boxmerge2),nx*ny,1);
[ixmerge2,foo,ixs2] = find(B2(ix(i2)));

% merge z and t weighted by m and compute the corresponding sum in m
i1 = i1(ixmerge1);
ms = m(i1);
[foo,foo,Z1] = find(sparse(ixs1,1,z(i1).*ms));
[foo,foo,M1] = find(sparse(ixs1,1,ms));
[foo,foo,T1] = find(sparse(ixs1,1,t(i1).*ms));
Z1 = Z1./M1;
T1 = T1./M1;

i2 = i2(ixmerge2);
ms = m(i2);
[foo,foo,Z2] = find(sparse(ixs2,1,z(i2).*ms));
[foo,foo,M2] = find(sparse(ixs2,1,ms));
[foo,foo,T2] = find(sparse(ixs2,1,t(i2).*ms));
Z2 = Z2./M2;
T2 = T2./M2;

% assemble result from merging
z([i1; i2])  = [];  z = [z; Z1; Z2];
m([i1; i2])  = [];  m = [m; M1; M2];
t([i1; i2])  = [];  t = [t; T1; T2];
ix([i1; i2]) = []; ix = [ix; boxmerge1; boxmerge2];

% separate positive and negative m's
i2 = m < 0;
i1 = find(~i2);
i2 = find(i2);

% redo the counting
boxcount1 = sparse(ix(i1),1,i1,nx*ny,1);
boxcount2 = sparse(ix(i2),1,i2,nx*ny,1);

% find boxes to merge
boxmerge3 = find(boxcount1 > 0 & boxcount2 > 0);

% corresponding indices to vortices
ixmerge1 = boxcount1(boxmerge3);
ixmerge2 = boxcount2(boxmerge3);

% mark vortices to annihilate
ixrem = find(abs(m(ixmerge1)+m(ixmerge2)) < ...
             mtol*(m(ixmerge1)-m(ixmerge2)));
ixannihilate = [ixmerge1; ixmerge2];
ixmerge1(ixrem) = []; ixmerge2(ixrem) = [];

% cancel the smaller vortex
[foo,ixkeep] = max([m(ixmerge1) -m(ixmerge2)],[],2);
ixkeep = [1:size(ixmerge1,1)]'+(ixkeep(:)-1)*size(ixmerge2,1);
Z3 = [z(ixmerge1); z(ixmerge2)];
Z3 = Z3(ixkeep);
M3 = m(ixmerge1)+m(ixmerge2);
T3 = [t(ixmerge1); t(ixmerge2)];
T3 = T3(ixkeep);

% assemble result
z(ixannihilate) = []; z = [z; Z3];
m(ixannihilate) = []; m = [m; M3];
t(ixannihilate) = []; t = [t; T3];

